﻿using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.NetworkInformation;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Dandelion
{
	/// <summary>
	/// The controller of AOD(AirFusion Of Dandelion) Framework Library.
	/// </summary>
	public class AODController
	{
		private static Logger _logger;

		private static string _dandelionHost;
		private static ushort? _dandelionPort;
		private static bool _isSecure;

		/// <summary>
		/// Occurs when login.
		/// </summary>
		public static event AODCommon.LoginResultEventHandler LoginHandler;

		/// <summary>
		/// Occurs when logout.
		/// </summary>
		public static event AODCommon.LogoutResultEventHandler LogoutHandler;

		private bool _isInitialized;
		private static bool _isLogining;
		private static bool _isLogouting;

		private int _loopTimes;
		private bool _isLoopWithNetworkAvailabilityChangedThread;
		private Thread _checkInternetConnectivityThread;

		private static AODCommon.NotificationResultEventHandler _notificationResultEventHandler;

		private string _profile;
		private static string _instanceID;
		private static Dictionary<Common.AODControllerAction, string> _callDictionary;

		private readonly AsyncLock _lock = new AsyncLock();

		/// <summary>
		/// Initializes a new instance of the AODController class.
		/// </summary>
		/// <param name="dandelionHost"> The host name of the web server. </param>
		/// <param name="dandelionPort"> Default is null. The port of the web server. </param>
		/// <param name="isSecure"> Default is true. The connection is secure or not. </param>
		public AODController(string dandelionHost = "www.acerapis.com", ushort? dandelionPort = null, bool isSecure = true)
		{
			_dandelionHost = dandelionHost;
			_dandelionPort = dandelionPort;
			_isSecure = isSecure;
			LoginHandler = null;
			LogoutHandler = null;

			this._loopTimes = Common.CheckInternetLoopTimes;
			this._checkInternetConnectivityThread = null;

			this._isInitialized = false;
			_isLogining = false;
			_isLogouting = false;
			_instanceID = string.Empty;
			this._profile = string.Empty;
			_callDictionary = new Dictionary<Common.AODControllerAction, string>();

			_notificationResultEventHandler = null;

			//try
			//{
			//	if (Directory.Exists(Common.CommAppFolderPath))
			//	{
			//		Directory.Delete(Common.CommAppFolderPath, true);
			//	}
			//}
			//catch (Exception e)
			//{
			//	Console.WriteLine("[AODController][DeleteLogFolder] Exception: " + e.ToString());
			//}

			if (Common.CheckLogEnableStatus())
			{
				EnableDebugLog(true);
			}
			else
			{
				EnableDebugLog(false);
			}

			string tmpTime = DateTime.Now.ToString("MMdd HH-mm-ss.fff");
			string logFileName = tmpTime + "_AODController.txt";
			_logger = new Logger(Common.LogFolderPath + logFileName);
		}

		/// <summary>
		/// Enables the debug log.
		/// </summary>
		/// <param name="isEnable">The value is enabled/disabled the debug log.</param>
		/// <param name="fileLimit">The file limit.</param>
		private void EnableDebugLog(bool isEnable, int fileLimit = 200)
		{
			Common.EnableOutputToLogFile = isEnable;
			Common.LogFileLimit = fileLimit;

			if (isEnable)
			{
				//Fixed the null exception when enable the log
				if (_logger != null)
				{
					_logger.WriteNewLine("=====");
				}
			}
		}

		#region Controller API for Get Local Info

		/// <summary>
		/// Get the machine information, ex: OS version, IP address, and device model...
		/// </summary>
		internal Dictionary<string, string> GetSystemInfo()
		{
			_logger.WriteNewLine("GetSystemInfo ..... ");

			Dictionary<string, string> dictionary = new Dictionary<string, string>();

			DeviceInfo deviceInfo = new DeviceInfo();
			dictionary.Add("UserName", deviceInfo.GetUserName());
			dictionary.Add("OSType", deviceInfo.GetOSType());

			string securityIdentifier = deviceInfo._securityIdentifier;
			try
			{
				securityIdentifier = deviceInfo.GetSecurityIdentifier();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[AODController][GetSecurityIdentifier] Exception:" + e.ToString());
			}
			dictionary.Add("SecurityIdentifier", securityIdentifier);

			string version = deviceInfo._OSVersion;
			try
			{
				version = deviceInfo.GetOSVersion();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[AODController][GetOSVersion] Exception:" + e.ToString());
			}
			dictionary.Add("OSVersion", version);

			string brand = deviceInfo._deviceBrand;
			try
			{
				brand = deviceInfo.GetDeviceBrand();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[AODController][GetDeviceBrand] Exception:" + e.ToString());
			}
			dictionary.Add("DeviceBrand", brand);

			string model = deviceInfo._deviceModelName;
			try
			{
				model = deviceInfo.GetDeviceModel();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[AODController][GetDeviceModel] Exception:" + e.ToString());
			}
			dictionary.Add("DeviceModel", model);

			string uuid = deviceInfo._deviceUuid;
			try
			{
				uuid = deviceInfo.GetDeviceUUID();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[AODController][GetDeviceUUID] Exception:" + e.ToString());
			}
			dictionary.Add("DeviceUUID", uuid);

			string types = deviceInfo._deviceChassisTypes;
			try
			{
				types = deviceInfo.GetDeviceChassisTypes();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[AODController][GetDeviceChassisTypes] Exception:" + e.ToString());
			}
			dictionary.Add("DeviceChassisTypes", types);

			string mac = deviceInfo._wifiMacAddress;
			try
			{
				mac = deviceInfo.GetWifiMacAddress();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[AODController][GetWifiMacAddress] Exception:" + e.ToString());
			}
			dictionary.Add("WifiMacAddress", mac);

			string ip = deviceInfo.GetIPAddress();
			try
			{
				ip = deviceInfo.GetIPAddress();
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[AODController][GetIPAddress] Exception:" + e.ToString());
			}
			dictionary.Add("IPAddress", ip);

			return dictionary;
		}

		#endregion Controller API for Get Local Info

		#region Controller API for initialize

		/// <summary>
		/// Initializes the <see cref="AODController"/> object.
		/// </summary>
		/// <param name="instanceId"> The instance ID of local device. </param>
		/// <exception cref="ArgumentNullException"></exception>
		/// <exception cref="InvalidOperationException"></exception>
		public void Initialize(string instanceId)
		{
			_logger.WriteNewLine("Initialize >>>>> ");

			if (string.IsNullOrEmpty(instanceId))
			{
				throw new ArgumentNullException("instanceId");
			}

			if (this._isInitialized)
			{
				throw (new InvalidOperationException("This instance is already initialized."));
			}

			this._isInitialized = true;
			_instanceID = instanceId;

			// register network change event to ensure the initialize will success
			NetworkChange.NetworkAvailabilityChanged += this.NetworkAvailabilityChanged;
			NetworkChange.NetworkAddressChanged += this.NetworkAddressChanged;

			InitializeAsync().ConfigureAwait(false);
			_logger.WriteNewLine("Initialize <<<<< ");
		}

		/// <summary>
		/// The initialize asynchronously.
		/// </summary>
		private async Task InitializeAsync()
		{
			_logger.WriteNewLine("InitializeAsync ..... ");

			string tmpDeviceId = string.Empty;
			tmpDeviceId = await this.GetLocalDeviceIdAsync().ConfigureAwait(false);

			if (string.IsNullOrEmpty(tmpDeviceId) || tmpDeviceId.Equals("null"))
			{
				UpdateCallDictionary(true, Common.AODControllerAction.ActionRegisterDeviceID, tmpDeviceId);
				CreateThreadToRetry();
			}
		}

		#endregion Controller API for initialize

		#region Controller API for login/logout

		/// <summary>
		/// Gets the login status.
		/// </summary>
		/// <param name="result"> The out value of the login information. </param>
		public static void GetLoginStatus(out AODCommon.LoginStatusResult result)
		{
			_logger.WriteNewLine("CheckLoginStatus ..... ");

			result = new AODCommon.LoginStatusResult();

			RegistryUtils tmpReg = new RegistryUtils();
			string tmpKeyPath = Common.RegRootKeyName + "\\" + Common.RegFirstSubKeyName + "\\" +
																				Common.RegSecondSubKeyName;
			int type = (int)AODCommon.LoginType.None;
			string username = string.Empty;
			try
			{
				type = (int)tmpReg.GetRegistryValue(tmpKeyPath, Common.RegLoginTypeKeyName, "");
				username = (string)tmpReg.GetRegistryValue(tmpKeyPath, Common.RegUserNameKeyName, "");
			}
			catch (Exception e)
			{
				_logger.WriteNewLine("[CheckLoginStatus] GetRegistryValue Exception:" + e.ToString());
			}

			result.Type = (AODCommon.LoginType)type;
			result.Username = username;
		}

		/// <summary>
		/// Logout the account.
		/// </summary>
		/// <exception cref="ArgumentNullException"></exception>
		/// <exception cref="InvalidOperationException"></exception>
		public static void LogoutAccount()
		{
			_logger.WriteNewLine("LogoutAccount ..... ");

			if (LogoutHandler == null)
			{
				throw new ArgumentNullException("LogoutHandler");
			}

			if (string.IsNullOrEmpty(_instanceID))
			{
				throw new InvalidOperationException("Please initialize this instance by Initialize().");
			}

			AODCommon.LoginStatusResult res;
			GetLoginStatus(out res);
			if (res != null && string.IsNullOrEmpty(res.Username))
			{
				throw (new InvalidOperationException("This device has not logged yet."));
			}
		}

		/// <summary>
		/// Creates the web Auth to login.
		/// </summary>
		/// <param name="type"> The value of login mode. </param>
		/// <param name="locationX"> The value of form's location of X. </param>
		/// <param name="locationY"> The value of form's location of Y. </param>
		/// <param name="width"> The value of form's width. </param>
		/// <param name="height"> The value of form's height. </param>
		/// <param name="title"> The string of form's title name. </param>
		/// <exception cref="ArgumentException"></exception>
		/// <exception cref="ArgumentNullException"></exception>
		/// <exception cref="InvalidOperationException"></exception>
		public static void LoginAccount(AODCommon.LoginType type, int locationX = 0, int locationY = 0, int width = 500, int height = 500, string title = "SocialWebLogin")
		{
			_logger.WriteNewLine("LoginAccount ..... ");

			if (type == AODCommon.LoginType.None)
			{
				throw new ArgumentException("The specified type is invalid.");
			}

			if (LoginHandler == null)
			{
				throw new ArgumentNullException("LoginHandler");
			}

			if (string.IsNullOrEmpty(_instanceID))
			{
				throw new InvalidOperationException("Please initialize this instance by Initialize().");
			}

			AODCommon.LoginStatusResult res;
			GetLoginStatus(out res);
			if (res != null && !string.IsNullOrEmpty(res.Username))
			{
				throw (new InvalidOperationException("This device has already logged on."));
			}
		}

		#endregion Controller API for login/logout

		#region Controller API for notification

		/// <summary>
		/// Enable the notification capture.
		/// </summary>
		/// <param name="notificationResultEventHandler">The notification result event handler.</param>
		/// <exception cref="ArgumentNullException"></exception>
		/// <exception cref="InvalidOperationException"></exception>
		public static void EnableNotification(AODCommon.NotificationResultEventHandler notificationResultEventHandler = null)
		{
			_logger.WriteNewLine("EnableNotification ..... ");

			if (notificationResultEventHandler == null)
			{
				throw new ArgumentNullException("notificationResultEventHandler");
			}

			if (!NetworkInterface.GetIsNetworkAvailable())
			{
				throw new InvalidOperationException("This operation cannot be done since there's no available network.");
			}

			_notificationResultEventHandler = notificationResultEventHandler;

			AODCommon.LoginStatusResult res;
			GetLoginStatus(out res);
			if (res != null && string.IsNullOrEmpty(res.Username))
			{
				throw (new InvalidOperationException("This device has not logged yet."));
			}
		}

		/// <summary>
		/// Gets the notification icon asynchronously.
		/// </summary>
		/// <param name="iconID">The resource identifier.</param>
		/// <exception cref="System.ArgumentException">The 'resourceID' is invalide.</exception>
		/// <exception cref="System.InvalidOperationException">This operation cannot be done since there's no available network.</exception>
		/// <returns></returns>
		public async Task<string> GetNotificationIconAsync(string iconID)
		{
			_logger.WriteNewLine("GetNotificationIconAsync ..... ");

			if (string.IsNullOrEmpty(iconID))
			{
				throw new ArgumentException("The specified icon ID is invalid.", "iconID");
			}

			if (!NetworkInterface.GetIsNetworkAvailable())
			{
				throw new InvalidOperationException("This operation cannot be done since there's no available network.");
			}

			return string.Empty;
		}

		/// <summary>
		/// Sends the message asynchronously.
		/// </summary>
		/// <param name="instanceID">The instance identifier of target device.</param>
		/// <param name="message">The action body.</param>
		/// <exception cref="ArgumentException"></exception>
		/// <exception cref="InvalidOperationException"></exception>
		/// <returns></returns>
		public async Task<AODCommon.SendMessageTaskResult> SendMessageAsync(string instanceID, string message)
		{
			_logger.WriteNewLine("SendMessageAsync ..... ");

			if (string.IsNullOrEmpty(instanceID))
			{
				throw new ArgumentException("The specified instance ID is invalid.", "instanceID");
			}

			if (string.IsNullOrEmpty(message))
			{
				throw new ArgumentException("The specified message is invalid.", "message");
			}

			if (!NetworkInterface.GetIsNetworkAvailable())
			{
				throw new InvalidOperationException("This operation cannot be done since there's no available network.");
			}

			return null;
		}

		#endregion Controller API for notification

		#region Controller API for set/get profile

		/// <summary>
		/// Sets the device profile.
		/// </summary>
		/// <param name="profile"> The profile of local device. </param>
		/// <exception cref="ArgumentNullException"></exception>
		/// <exception cref="InvalidOperationException"></exception>
		public void SetDeviceProfile(string profile)
		{
			_logger.WriteNewLine("SetDeviceProfile >>>>> ");

			if (string.IsNullOrEmpty(profile))
			{
				throw new ArgumentNullException("profile");
			}

			this._profile = profile;

			if (string.IsNullOrEmpty(_instanceID))
			{
				throw new InvalidOperationException("Please initialize this instance by Initialize().");
			}

			SetDeviceProfileAsync().ConfigureAwait(false);
			_logger.WriteNewLine("SetDeviceProfile <<<<< ");
		}

		/// <summary>
		/// Gets the profile of the specified device asynchronously.
		/// </summary>
		/// <param name="instanceId">The instance ID of the device.</param>
		/// <exception cref="ArgumentNullException"></exception>
		/// <exception cref="InvalidOperationException"></exception>
		public async Task<AODCommon.GetDeviceProfileResult> GetDeviceProfileAsync(string instanceId)
		{
			_logger.WriteNewLine("GetDeviceProfileAsync >>>>> ");

			if (string.IsNullOrEmpty(instanceId))
			{
				throw new ArgumentNullException("instanceId");
			}

			if (!NetworkInterface.GetIsNetworkAvailable())
			{
				throw new InvalidOperationException("Network is not Available");
			}

			CreateThreadToRetry();

			ProcessWithGetDeviceProfile gateProfile = new ProcessWithGetDeviceProfile(_dandelionHost,
																					_dandelionPort, _isSecure);
			gateProfile.SetUpdateDeviceProfileFields(instanceId);
			_logger.WriteNewLine("GetDeviceProfileAsync <<<<< ");
			return await gateProfile.GetDeviceProfileProcAsync().ConfigureAwait(false);
		}

		/// <summary>
		/// Set device profile asynchronously.
		/// </summary>
		private async Task SetDeviceProfileAsync()
		{
			string tmpDeviceId = string.Empty;
			tmpDeviceId = await this.GetLocalDeviceIdAsync().ConfigureAwait(false);

			if (string.IsNullOrEmpty(tmpDeviceId) || tmpDeviceId.Equals("null"))
			{
				UpdateCallDictionary(true, Common.AODControllerAction.ActionRegisterDeviceID, tmpDeviceId);
				UpdateCallDictionary(true, Common.AODControllerAction.ActionUpdateProfile, this._profile);
			}
			else
			{
				ProcessWithUpdateDeviceProfile updateProfile = new ProcessWithUpdateDeviceProfile(_dandelionHost,
																					_dandelionPort, _isSecure);
				updateProfile.SetUpdateDeviceProfileFields(tmpDeviceId, _instanceID, this._profile);

				AODCommon.UpdateDeviceProfileResult res = null;
				res = await updateProfile.UpdateDeviceProfileProcAsync().ConfigureAwait(false);

				if (res != null)
				{
					if (res.IsSuccess)
					{
						return;
					}
				}

				UpdateCallDictionary(true, Common.AODControllerAction.ActionUpdateProfile, this._profile);
			}

			CreateThreadToRetry();
		}

		#endregion Controller API for set/get profile

		#region Controller API for pairing

		/// <summary>
		/// Starts to discover the nearby devices and sets as discoverable.
		/// </summary>
		/// <param name="holdTime"> The time span that is hold on the server, in seconds. </param>
		/// <param name="pairingEventHandler"> Default is null. Otherwise the near devices' profile callback. </param>
		/// <param name="isAdvertiser"> Default is true, the near device list will not provide. If is false, the list will provide. </param>
		/// <exception cref="AODException"></exception>
		/// <exception cref="ArgumentNullException"></exception>
		/// <exception cref="InvalidOperationException"></exception>
		public void StartPairing(long holdTime = 5, AODCommon.PairingEventHandler pairingEventHandler = null, bool isAdvertiser = true)
		{
			_logger.WriteNewLine("StartPairing >>>>> ");

			if (pairingEventHandler == null)
			{
				throw new ArgumentNullException("pairingEventHandler");
			}

			if (!NetworkInterface.GetIsNetworkAvailable())
			{
				throw new InvalidOperationException("Network is not Available");
			}

			if (string.IsNullOrEmpty(_instanceID))
			{
				throw new InvalidOperationException("Please initialize this instance by Initialize().");
			}

			StartPairingAsync(isAdvertiser, holdTime, pairingEventHandler).ConfigureAwait(false);
			_logger.WriteNewLine("StartPairing <<<<< ");
		}

		/// <summary>
		/// Start pairing asynchronously.
		/// </summary>
		private async Task StartPairingAsync(bool isAdvertiser, long holdTime, AODCommon.PairingEventHandler pairingEventHandler)
		{
			CreateThreadToRetry();

			string tmpDeviceId = string.Empty;
			tmpDeviceId = await this.GetLocalDeviceIdAsync().ConfigureAwait(false);

			ProcessWithPairingStart pairingStart = new ProcessWithPairingStart(_dandelionHost, _dandelionPort,
																										_isSecure);
			pairingStart.SetPairingStartFields(tmpDeviceId, _instanceID, holdTime, pairingEventHandler);
			pairingStart.PairingStartProcAsync(isAdvertiser).ConfigureAwait(false);
		}

		/// <summary>
		/// Sends pairing invitation to the specified device asynchronously.
		/// </summary>
		/// <param name="instanceId"> The instance ID of target device. </param>
		/// <exception cref="AODException"></exception>
		/// <exception cref="ArgumentNullException"></exception>
		/// <exception cref="InvalidOperationException"></exception>
		public async Task<AODCommon.PairingInvitationTaskResult> SendPairingInvitationAsync(string instanceId)
		{
			_logger.WriteNewLine("SendPairingInvitationAsync >>>>> ");

			if (string.IsNullOrEmpty(instanceId))
			{
				throw new ArgumentNullException("instanceId");
			}

			if (!NetworkInterface.GetIsNetworkAvailable())
			{
				throw new InvalidOperationException("Network is not Available");
			}

			CreateThreadToRetry();

			string tmpDeviceId = await this.GetLocalDeviceIdAsync().ConfigureAwait(false);
			if (string.IsNullOrEmpty(tmpDeviceId) && tmpDeviceId.Equals("null"))
			{
				AODCommon.PairingInvitationTaskResult task = new AODCommon.PairingInvitationTaskResult();
				task.IsSuccess = false;
				task.StatusCode = AODCommon.DandelionStatusCode.ControllerStatusCodeInternetFailure;
				task.Exception = new Exception("The SendPairingInvitationAsync fail, because the internet error.");
				return task;
			}

			ProcessWithPairingRequest pairingRequest = new ProcessWithPairingRequest(_dandelionHost,
																					_dandelionPort, _isSecure);
			pairingRequest.SetPairingRequestFields(tmpDeviceId, instanceId);
			_logger.WriteNewLine("SendPairingInvitationAsync <<<<< ");
			return await pairingRequest.PairingRequestProcAsync().ConfigureAwait(false);
		}

		#endregion Controller API for pairing

		#region Internal/Protected/Private Methods

		/// <summary>
		/// Gets the device ID of the local device asynchronously.
		/// </summary>
		private async Task<string> GetLocalDeviceIdAsync(bool isInitialize = false)
		{
			using (var releaser = await _lock.LockAsync())
			{
				RegistryUtils tmpReg = new RegistryUtils();
				string tmpKeyPath = Common.RegRootKeyName + "\\" + Common.RegFirstSubKeyName + "\\" +
																					Common.RegSecondSubKeyName;
				string tmpDeviceId = string.Empty;

				try
				{
					tmpDeviceId = (string)tmpReg.GetRegistryValue(tmpKeyPath, Common.RegDeviceIDKeyName, "");
				}
				catch (Exception e)
				{
					_logger.WriteNewLine("[AODFrameworkLibrary][GetLocalDeviceIdAsync] GetRegistryValue Exception: " + e.ToString());
				}

				if (isInitialize)
				{
					ProcessWithGetDeviceID getDeviceID = new ProcessWithGetDeviceID();
					string newDeviceId = await getDeviceID.GetDeviceIDProcAsync(_dandelionHost, _dandelionPort, _isSecure).ConfigureAwait(false);

					if (string.IsNullOrEmpty(tmpDeviceId) || tmpDeviceId.Equals("null") || !tmpDeviceId.Equals(newDeviceId))
					{
						tmpDeviceId = newDeviceId;
						if (!string.IsNullOrEmpty(tmpDeviceId) && !tmpDeviceId.Equals("null"))
						{
							try
							{
								tmpReg.SetRegistryValue(tmpKeyPath, Common.RegDeviceIDKeyName, tmpDeviceId, RegistryValueKind.String);
							}
							catch (Exception e)
							{
								_logger.WriteNewLine("[AODFrameworkLibrary][GetLocalDeviceIdAsync] SetRegistryValue Exception: " + e.ToString());
							}
						}
					}
				}
				else
				{
					if (string.IsNullOrEmpty(tmpDeviceId) || tmpDeviceId.Equals("null"))
					{
						ProcessWithGetDeviceID getDeviceID = new ProcessWithGetDeviceID();
						tmpDeviceId = await getDeviceID.GetDeviceIDProcAsync(_dandelionHost, _dandelionPort, _isSecure).ConfigureAwait(false);

						if (!string.IsNullOrEmpty(tmpDeviceId) && !tmpDeviceId.Equals("null"))
						{
							try
							{
								tmpReg.SetRegistryValue(tmpKeyPath, Common.RegDeviceIDKeyName, tmpDeviceId, RegistryValueKind.String);
							}
							catch (Exception e)
							{
								_logger.WriteNewLine("[AODFrameworkLibrary][GetLocalDeviceIdAsync] SetRegistryValue Exception: " + e.ToString());
							}
						}
					}
				}

				return tmpDeviceId;
			}
		}

		#endregion Internal/Protected/Private Methods

		/// <summary>
		/// Called when Network address changed.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
		private void NetworkAvailabilityChanged(object sender, NetworkAvailabilityEventArgs e)
		{
			_logger.WriteNewLine("NetworkAvailabilityChanged ...");
			if (!e.IsAvailable)
			{
				this._isLoopWithNetworkAvailabilityChangedThread = false;
				this._checkInternetConnectivityThread = null;
			}
		}

		/// <summary>
		/// Called when Network address changed.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
		private void NetworkAddressChanged(object sender, EventArgs e)
		{
			_logger.WriteNewLine("NetworkAddressChanged ...");
			CreateThreadToRetry();
		}

		private static Object LOCK = new Object();

		/// <summary>
		/// Updates API request into dictionary, to wait network is available.
		/// </summary>
		/// <param name="key"> The key means the action item. </param>
		/// <param name="value"> The value means the action value. </param>
		private static Common.AODControllerAction UpdateCallDictionary(bool isAdd, Common.AODControllerAction key, string value)
		{
			lock (LOCK)
			{
				try
				{
					if (_callDictionary != null)
					{
						if (isAdd)
						{
							if (_callDictionary.ContainsKey(key))
							{
								_callDictionary.Remove(key);
							}

							_callDictionary.Add(key, value);
						}
						else
						{
							if (key == Common.AODControllerAction.ActionNone)
							{
								if (_callDictionary.Count > 0)
								{
									KeyValuePair<Common.AODControllerAction, string> tmpKeyValue = new KeyValuePair<Common.AODControllerAction, string>();
									if (_callDictionary.ContainsKey(Common.AODControllerAction.ActionRegisterDeviceID))
									{
										for (int i = 0; i < _callDictionary.Count; i++)
										{
											if (_callDictionary.ElementAt(i).Key == Common.AODControllerAction.ActionRegisterDeviceID)
											{
												tmpKeyValue = _callDictionary.ElementAt(i);
												return tmpKeyValue.Key;
											}
										}
									}

									tmpKeyValue = _callDictionary.ElementAt(0);
									return tmpKeyValue.Key;
								}
							}
							else
							{
								_callDictionary.Remove(key);
							}
						}
					}
				}
				catch (Exception e)
				{
					_logger.WriteNewLine("[AODController][AddCallDictionary] Exception: " + e.ToString());
				}

				return Common.AODControllerAction.ActionNone;
			}
		}

		/// <summary>
		/// Check the internet conncetivity asynchronously.
		/// </summary>
		private async Task CheckInternetConnectivityAsync()
		{
			_logger.WriteNewLine("CheckInternetConnectivityAsync, start ..... ");

			Common.AODControllerAction tmpAction = UpdateCallDictionary(false, Common.AODControllerAction.ActionNone, "");

			if (tmpAction == Common.AODControllerAction.ActionRegisterDeviceID)
			{
				_logger.WriteNewLine("CheckInternetConnectivityAsync, get Local DeviceId ..... ");
				string tmpDeviceId = await this.GetLocalDeviceIdAsync(true).ConfigureAwait(false);
				if (!string.IsNullOrEmpty(tmpDeviceId))
				{
					if (!tmpDeviceId.Equals("null"))
					{
						UpdateCallDictionary(false, Common.AODControllerAction.ActionRegisterDeviceID, "");
					}
				}
			}
			else if (tmpAction == Common.AODControllerAction.ActionUpdateProfile)
			{
				string tmpDeviceId = await GetLocalDeviceIdAsync().ConfigureAwait(false);
				ProcessWithUpdateDeviceProfile updateProfile = new ProcessWithUpdateDeviceProfile(
																							_dandelionHost,
																							_dandelionPort,
																							_isSecure);
				updateProfile.SetUpdateDeviceProfileFields(tmpDeviceId, _instanceID, this._profile);

				_logger.WriteNewLine("CheckInternetConnectivityAsync, update profile ..... ");
				AODCommon.UpdateDeviceProfileResult res = await updateProfile.UpdateDeviceProfileProcAsync().ConfigureAwait(false);
				if (res.IsSuccess)
				{
					_logger.WriteNewLine("CheckInternetConnectivityAsync, update profile (Success)..... ");
					UpdateCallDictionary(false, Common.AODControllerAction.ActionUpdateProfile, "");
				}
				else if (res.StatusCode == (AODCommon.DandelionStatusCode)Common.HttpStatusCode.HttpStatusCodeNotFound)
				{
					_logger.WriteNewLine("CheckInternetConnectivityAsync, update profile (Not Found)..... ");
					UpdateCallDictionary(true, Common.AODControllerAction.ActionRegisterDeviceID, "");
				}
				else if (res.StatusCode == (AODCommon.DandelionStatusCode)Common.HttpStatusCode.HttpStatusCodeRequestTimeout
					|| res.StatusCode == (AODCommon.DandelionStatusCode)Common.HttpStatusCode.HttpStatusCodeConflict)
				{
					_logger.WriteNewLine("CheckInternetConnectivityAsync, update profile (RequestTimeout or Conflict)..... ");
				}
				else if ((AODCommon.DandelionStatusCode)Common.HttpStatusCode.HttpStatusCodeBadRequest <= res.StatusCode
					&& res.StatusCode < (AODCommon.DandelionStatusCode)Common.HttpStatusCode.HttpStatusCodeInternalServerError)
				{
					_logger.WriteNewLine("CheckInternetConnectivityAsync, update profile (4xx)..... ");
					this._isLoopWithNetworkAvailabilityChangedThread = false;
					this._checkInternetConnectivityThread = null;
				}
			}
			else
			{
				_logger.WriteNewLine("CheckInternetConnectivityAsync, the action is clean ..... ");
				this._isLoopWithNetworkAvailabilityChangedThread = false;
				this._checkInternetConnectivityThread = null;
			}

			_logger.WriteNewLine("CheckInternetConnectivityAsync, stop ..... ");
		}

		private Object checkLock = new Object();

		private void CreateThreadToRetry()
		{
			if (_callDictionary != null && _callDictionary.Count > 0)
			{
				lock (checkLock)
				{
					if (this._checkInternetConnectivityThread == null)
					{
						this._loopTimes = Common.CheckInternetLoopTimes;
						this._isLoopWithNetworkAvailabilityChangedThread = true;
						this._checkInternetConnectivityThread = new Thread(new ThreadStart(ThreadWithNetworkAvailabilityChanged));
						this._checkInternetConnectivityThread.Name = "Dandelion.ThreadWithNetworkAvailabilityChanged";
						this._checkInternetConnectivityThread.IsBackground = true;
						this._checkInternetConnectivityThread.Start();
					}
				}
			}
		}

		/// <summary>
		/// The thread of check network availability changed.
		/// </summary>
		private void ThreadWithNetworkAvailabilityChanged()
		{
			_logger.WriteNewLine("ThreadWithNetworkAvailabilityChanged, start ..... ");

			do
			{
				this._loopTimes--;
				CheckInternetConnectivityAsync().ConfigureAwait(false);
				Thread.Sleep(Common.SocketTimeout);
			} while (this._isLoopWithNetworkAvailabilityChangedThread && this._loopTimes > 0);

			this._isLoopWithNetworkAvailabilityChangedThread = false;
			this._checkInternetConnectivityThread = null;
			_logger.WriteNewLine("ThreadWithNetworkAvailabilityChanged, stop ..... ");
		}
	}

	internal class AsyncSemaphore
	{
		private readonly static Task s_completed = Task.FromResult(true);
		private readonly Queue<TaskCompletionSource<bool>> m_waiters = new Queue<TaskCompletionSource<bool>>();
		private int m_currentCount;

		public AsyncSemaphore(int initialCount)
		{
			if (initialCount < 0) throw new ArgumentOutOfRangeException("initialCount");
			m_currentCount = initialCount;
		}

		public Task WaitAsync()
		{
			lock (m_waiters)
			{
				if (m_currentCount > 0)
				{
					--m_currentCount;
					return s_completed;
				}
				else
				{
					var waiter = new TaskCompletionSource<bool>();
					m_waiters.Enqueue(waiter);
					return waiter.Task;
				}
			}
		}

		public void Release()
		{
			TaskCompletionSource<bool> toRelease = null;
			lock (m_waiters)
			{
				if (m_waiters.Count > 0)
					toRelease = m_waiters.Dequeue();
				else
					++m_currentCount;
			}
			if (toRelease != null)
				toRelease.SetResult(true);
		}
	}

	internal class AsyncLock
	{
		private readonly AsyncSemaphore m_semaphore;
		private readonly Task<Releaser> m_releaser;

		public AsyncLock()
		{
			m_semaphore = new AsyncSemaphore(1);
			m_releaser = Task.FromResult(new Releaser(this));
		}

		public Task<Releaser> LockAsync()
		{
			var wait = m_semaphore.WaitAsync();
			return wait.IsCompleted ?
			  m_releaser :
			  wait.ContinueWith((_, state) => new Releaser((AsyncLock)state),
				this, CancellationToken.None,
				TaskContinuationOptions.ExecuteSynchronously, TaskScheduler.Default);
		}

		public struct Releaser : IDisposable
		{
			private readonly AsyncLock m_toRelease;

			internal Releaser(AsyncLock toRelease)
			{
				m_toRelease = toRelease;
			}

			public void Dispose()
			{
				if (m_toRelease != null)
					m_toRelease.m_semaphore.Release();
			}
		}
	}
}
