﻿using System;
using System.ComponentModel;
using System.Net;

namespace AirFusion.Common.Converters
{
	/// <summary>
	/// Provides a type converter to convert <see cref="System.Net.IPAddress"/> objects to and from various other representations.
	/// </summary>
	public class IPAddressConverter : TypeConverter
	{
		/// <summary>
		/// Returns whether this converter can convert an object of the given type to the type of this converter, using the specified context.
		/// </summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext" /> that provides a format context.</param>
		/// <param name="sourceType">A <see cref="T:System.Type" /> that represents the type you want to convert from.</param>
		/// <returns>
		/// true if this converter can perform the conversion; otherwise, false.
		/// </returns>
		public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
		{
			return sourceType.Equals(typeof(IPAddress)) || sourceType.Equals(typeof(byte[])) || sourceType.Equals(typeof(string));
		}

		/// <summary>
		/// Returns whether this converter can convert the object to the specified type, using the specified context.
		/// </summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext" /> that provides a format context.</param>
		/// <param name="destinationType">A <see cref="T:System.Type" /> that represents the type you want to convert to.</param>
		/// <returns>
		/// true if this converter can perform the conversion; otherwise, false.
		/// </returns>
		public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
		{
			return destinationType.Equals(typeof(IPAddress)) || destinationType.Equals(typeof(string)) || destinationType.Equals(typeof(byte[]));
		}

		/// <summary>
		/// Returns whether changing a value on this object requires a call to <see cref="M:System.ComponentModel.TypeConverter.CreateInstance(System.Collections.IDictionary)" /> to create a new value, using the specified context.
		/// </summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext" /> that provides a format context.</param>
		/// <returns>
		/// true if changing a property on this object requires a call to <see cref="M:System.ComponentModel.TypeConverter.CreateInstance(System.Collections.IDictionary)" /> to create a new value; otherwise, false.
		/// </returns>
		public override bool GetCreateInstanceSupported(ITypeDescriptorContext context)
		{
			return false;
		}

		/// <summary>
		/// Converts the given object to the type of this converter, using the specified context and culture information.
		/// </summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext" /> that provides a format context.</param>
		/// <param name="culture">The <see cref="T:System.Globalization.CultureInfo" /> to use as the current culture.</param>
		/// <param name="value">The <see cref="T:System.Object" /> to convert.</param>
		/// <returns>
		/// An <see cref="T:System.Object" /> that represents the converted value.
		/// </returns>
		/// <exception cref="System.ArgumentOutOfRangeException" />
		/// <exception cref="System.NotSupportedException" />
		public override object ConvertFrom(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value)
		{
			if (value == null)
			{
				return null;
			}
			else if (value is IPAddress)
			{
				return value;
			}
			else if (value is byte[])
			{
				try
				{
					return new IPAddress(value as byte[]);
				}
				catch
				{
					return new ArgumentOutOfRangeException("The specified byte array to describe an IP address is out of range.");
				}
			}
			else if (value is string)
			{
				try
				{
					return IPAddress.Parse(value as string);
				}
				catch
				{
					return new ArgumentOutOfRangeException("The specified string is not a valid IP address.");
				}
			}
			else
			{
				throw new NotSupportedException("The type '" + value.GetType() + "' is not supported.");
			}
		}

		/// <summary>
		/// Converts the given value object to the specified type, using the specified context and culture information.
		/// </summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext" /> that provides a format context.</param>
		/// <param name="culture">A <see cref="T:System.Globalization.CultureInfo" />. If null is passed, the current culture is assumed.</param>
		/// <param name="value">The <see cref="T:System.Object" /> to convert.</param>
		/// <param name="destinationType">The <see cref="T:System.Type" /> to convert the <paramref name="value" /> parameter to.</param>
		/// <returns>
		/// An <see cref="T:System.Object" /> that represents the converted value.
		/// </returns>
		/// <exception cref="System.ArgumentException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.NotSupportedException" />
		public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType)
		{
			// check parameters
			if (!(value is IPAddress))
			{
				throw new ArgumentException("The value you specified is not an IPAddress value.");
			}
			if (destinationType == null)
			{
				throw new ArgumentNullException("destinationType");
			}

			if (destinationType.Equals(typeof(IPAddress)))
			{
				return value;
			}
			else if (destinationType.Equals(typeof(string)))
			{
				return value.ToString();
			}
			else if (destinationType.Equals(typeof(byte[])))
			{
				return (value as IPAddress).GetAddressBytes();
			}
			else
			{
				throw new NotSupportedException("The destination type '" + destinationType.Name + "' is not supported.");
			}
		}
	}
}
