﻿//   Class: IniTable
// Purpose: Access the INI file without Win32API.
//  Author: Ajax 2012-10 (Original Version)
//          Seoker 2013-04 (Second Version)
//  Describes:
//      1. Leading spacing or trailing spacing of section names will be removed.
//      2. Duplicate section names will be combined.
//      3. The second duplicate key will overwrites the existing key/value pair.
// Change Logs:
//      Date        Name           Description
//      2012-11-26  Seoker Wang    Modify some function. No need to write an empty keypair to keep the section name.
//	    2013-02-13  Seoker Wang    Add a method 'ExistSection'. Modify WriteToFile with a return value of type boolean.
//      2013-04-17  Seoker Wang    Modify IniTable signature and functions. Therefore, we could use iniTable["Main"]["OS"] to get the value.
//                                 Remove some functions that doesn't need anymore.
//      2013-04-18  Seoker Wang    Rewrite IniTable and related class. No need to create object first before create new section/key.
//                                 Keep empty lines and empty sections.
//      2013-04-19  Seoker Wang    Keep invalid lines, but won't read as key/value pair. Override GetEnumerator() to get valid sections only.
//      2013-05-24  Seoker Wang    Modify to support iniTable[section, key] format.
//      2013-06-10  Seoker Wang    Modify return type of Sections to be List<string> instead of ICollection<string>.
//      2014-09-17  Seoker Wang    Add descriptions and modify some modifier and return type.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.Common
{
	/// <summary>
	/// Provides the methods of INI file manipulation.
	/// </summary>
	public class IniTable : SectionKeyPairs
	{
		internal string _filePath = string.Empty;

		#region Properties

		/// <summary>
		/// Gets section names of current INI table.
		/// </summary>
		public virtual List<string> Sections
		{
			get
			{
				return (this.Where(section => IsValidSection(section)).Select(c => c.Key)).ToList();
			}
		}

		/// <summary>
		/// Gets all key names of current INI table.
		/// </summary>
		public override ICollection<string> Keys
		{
			get
			{
				List<string> keyList = new List<string>();
				foreach (var section in this)
				{
					foreach (var key in section.Value.Keys)
						keyList.Add(key);
				}

				return keyList;
			}
		}

		/// <summary>
		/// Override the original 'Values' property to avoid mistakes using.
		/// </summary>
		public override ICollection<KeyValuePairs> Values
		{
			get
			{
				throw new NotImplementedException("Please use 'KeyValues' property instead.");
			}
		}

		/// <summary>
		/// Gets all key/value pair of current ini table.
		/// </summary>
		public virtual ICollection<KeyValuePairs> KeyValues
		{
			get
			{
				List<KeyValuePairs> keyValueList = new List<KeyValuePairs>();
				foreach (var section in this)
				{
					keyValueList.Add(section.Value);
				}

				return keyValueList;
			}
		}

		/// <summary>
		/// Gets or sets the specified section. Reminds the section name will be trimmed.
		/// </summary>
		/// <param name="sectionName"></param>
		/// <returns></returns>
		public override KeyValuePairs this[string sectionName]
		{
			get
			{
				return base[sectionName.Trim()];
			}
			set
			{
				base[sectionName.Trim()] = value;
			}
		}

		/// <summary>
		/// Gets or sets the specified section name and key name.
		/// </summary>
		/// <param name="sectionName">The section name.</param>
		/// <param name="keyName">The key name.</param>
		/// <returns></returns>
		public string this[string sectionName, string keyName]
		{
			get
			{
				return this[sectionName][keyName];
			}
			set
			{
				this[sectionName][keyName] = value;
			}
		}

		#endregion Properties

		#region Methods

		/// <summary>
		/// Initializes the instance.
		/// </summary>
		public IniTable()
		{
		}

		/// <summary>
		/// Initializes the instance with the specified file.
		/// </summary>
		/// <param name="filePath">The full path of the INI file.</param>
		public IniTable(string filePath)
		{
			this._filePath = filePath;

			if (File.Exists(filePath))
			{
				this.ReadFromFile(filePath);
			}
		}

		/// <summary>
		/// Reads from file into the INI table.
		/// </summary>
		/// <param name="filePath">The full path of the INI file.</param>
		/// <returns>true if success; false otherwise.</returns>
		public virtual void ReadFromFile(string filePath)
		{
			this._filePath = filePath;
			try
			{
				using (StreamReader sr = new StreamReader(this._filePath))
				{
					string line = null;
					KeyValuePairs currentSection = null;
					int blankLines = 0;

					while ((line = sr.ReadLine()) != null)
					{
						int equalIndex = line.IndexOf('=');
						string key = null;
						if (equalIndex >= 0)
						{
							key = line.Substring(0, equalIndex).Trim();
						}

						// [section]
						if (line.StartsWith("[") && line.TrimEnd().EndsWith("]"))
						{
							line = line.Substring(1, line.Length - 2).Trim();
							if (line.Length > 0)
							{
								if (!this.ContainsSection(line))
								{
									base[line] = currentSection = new KeyValuePairs();
								}
								else
								{
									currentSection = this[line];
								}
								currentSection.KeepEmptySection = true; // always keep the empty section that already existing in the file.
							}
							else
							{
								currentSection = null;
							}
						}
						// key=value
						else if (!string.IsNullOrWhiteSpace(key) && currentSection != null)
						{
							string value = string.Empty;
							if (line.Length > equalIndex + 1) // might out of range when the line is like xxx=
							{
								value = line.Substring(equalIndex + 1).Trim();
							}

							if ((value.StartsWith("\"") && value.EndsWith("\"")) || (value.StartsWith("'") && value.EndsWith("'")))
							{
								value = value.Substring(1, value.Length - 2);
							}

							currentSection[key] = value;
						}
						else // comment, invalid lines, or invalid key/value pair => keep as other form
						{
							base[" !!OtherForm" + (++blankLines) + "!! "][string.Empty] = line; // normal section won't contain leading spacing or trailing spacing
						}
					}
				}
			}
			catch (Exception ex)
			{
				throw new Exception("Failed reading the file.", ex);
			}
		}

		/// <summary>
		/// Writes the INI table to the file.
		/// </summary>
		/// <param name="filePath">The full path of the INI file. If this value isn't specified, the original path will be used.</param>
		public virtual void WriteToFile(string filePath = null)
		{
			if (string.IsNullOrEmpty(filePath))
			{
				filePath = this._filePath;
			}

			string directoryName = Path.GetDirectoryName(filePath);
			try
			{
				// create directory
				if (!Directory.Exists(directoryName))
				{
					Directory.CreateDirectory(directoryName);
				}
			}
			catch (Exception ex)
			{
				throw new Exception("Failed creating folder '" + directoryName + "'.", ex);
			}

			try
			{
				using (StreamWriter sw = new StreamWriter(filePath))
				{
					List<string> sections = new List<string>();

					foreach (var section in this._dic)
					{
						if (section.Key.Trim().Length != section.Key.Length) // if there's leading spacing or trailing spacing => consider as non-section
						{
							sw.WriteLine(section.Value[string.Empty]);
						}
						else if (section.Value.Count > 0 || section.Value.KeepEmptySection) // [section], write only when there's at least one key/value pair or KeepEmptySection flag
						{
							sw.WriteLine("[{0}]", section.Key);
							foreach (var line in section.Value)
							{
								if (string.IsNullOrWhiteSpace(line.Key) || line.Value == null)
								{
									continue;
								}
								if (line.Value.Trim().Length < line.Value.Length) // there's beginning or ending spacing
								{
									sw.WriteLine("{0}=\"{1}\"", line.Key, line.Value);
								}
								else
								{
									sw.WriteLine("{0}={1}", line.Key, line.Value);
								}
							}
						}
					}
				}
			}
			catch (Exception ex)
			{
				throw new Exception("Failed writing the file.", ex);
			}
		}

		#region Override

		/// <summary>
		/// Determines whether the ini table contains the specified section.
		/// </summary>
		/// <param name="sectionName"></param>
		/// <returns></returns>
		public bool ContainsSection(string sectionName)
		{
			return (!string.IsNullOrWhiteSpace(sectionName)) ? base.ContainsKey(sectionName) : false;
		}

		/// <summary>
		/// Determines whether the ini table contains the specified key.
		/// </summary>
		/// <param name="keyName"></param>
		/// <returns></returns>
		public override bool ContainsKey(string keyName)
		{
			if (string.IsNullOrWhiteSpace(keyName))
			{
				return false;
			}

			foreach (var section in this)
			{
				if (section.Value.ContainsKey(keyName))
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Returns an enumerator that iterates through the ini table.
		/// </summary>
		/// <returns></returns>
		public IEnumerator<KeyValuePair<string, KeyValuePairs>> GetEnumerator()
		{
			return this._dic.Where(section => IsValidSection(section)).GetEnumerator();
		}

		#endregion Override

		#endregion Methods

		#region Static Method

		/// <summary>
		/// Determines whether the specified section is a valid section.
		/// </summary>
		/// <param name="keyValuePair">The key-value pair of (section, key-value pair).</param>
		/// <returns></returns>
		internal static bool IsValidSection(KeyValuePair<string, KeyValuePairs> keyValuePair)
		{
			return (keyValuePair.Key.Trim().Length == keyValuePair.Key.Length &&
					keyValuePair.Value != null &&
					(keyValuePair.Value.Count > 0 || keyValuePair.Value.KeepEmptySection == true));
		}

		/// <summary>
		/// Gets the specified value of specified section and key. If such section or key doesn't exist, return null string.
		/// </summary>
		/// <param name="filePath">The full path of the INI file.</param>
		/// <param name="sectionName">The section name.</param>
		/// <param name="keyName">The key name.</param>
		/// <returns></returns>
		public static string GetValue(string filePath, string sectionName, string keyName)
		{
			IniTable temp = new IniTable();
			temp.ReadFromFile(filePath);
			if (temp.ContainsSection(sectionName) && temp[sectionName].ContainsKey(keyName))
				return temp[sectionName][keyName];
			else
				return null;
		}

		#endregion Static Method
	}

	/// <summary>
	/// Provides the methods of section-key pair table.
	/// </summary>
	public class SectionKeyPairs : IDictionary<string, KeyValuePairs>
	{
		internal Dictionary<string, KeyValuePairs> _dic = null;

		#region Property Overriding

		/// <summary>
		/// Gets the total number of sections.
		/// </summary>
		public virtual int Count
		{
			get
			{
				// debug assertion
				Debug.Assert(this._dic != null);

				return this._dic.Count;
			}
		}

		/// <summary>
		/// Gets if the table is read-only.
		/// </summary>
		public virtual bool IsReadOnly
		{
			get { return false; }
		}

		/// <summary>
		/// Gets the collection of section names.
		/// </summary>
		public virtual ICollection<string> Keys
		{
			get
			{
				// debug assertion
				Debug.Assert(this._dic != null);

				return this._dic.Keys;
			}
		}

		/// <summary>
		/// Gets the collection of key-value pairs.
		/// </summary>
		public virtual ICollection<KeyValuePairs> Values
		{
			get
			{
				// debug assertion
				Debug.Assert(this._dic != null);

				return this._dic.Values;
			}
		}

		#endregion Property Overriding

		#region Method Overriding

		/// <summary>
		/// Initializes the instance.
		/// </summary>
		public SectionKeyPairs()
		{
			this._dic = new Dictionary<string, KeyValuePairs>(StringComparer.OrdinalIgnoreCase);
		}

		/// <summary>
		/// Gets or sets the specified section. Reminds if you pass section name with leading spacing or trailing spacing, the section will be invalid.
		/// </summary>
		/// <param name="sectionName"></param>
		/// <returns></returns>
		public virtual KeyValuePairs this[string sectionName]
		{
			get
			{
				// debug assertion
				Debug.Assert(this._dic != null);

				return (!string.IsNullOrWhiteSpace(sectionName) && this._dic.ContainsKey(sectionName)) ? this._dic[sectionName] : (this._dic[sectionName] = new KeyValuePairs());
			}
			set
			{
				// debug assertion
				Debug.Assert(this._dic != null);

				this._dic[sectionName] = value;
			}
		}

		/// <summary>
		/// Adds the specified section.
		/// </summary>
		/// <param name="sectionName">The section name.</param>
		/// <param name="keyValuePair">The key-value pair table if (key, value).</param>
		public virtual void Add(string sectionName, KeyValuePairs keyValuePair)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			this._dic.Add(sectionName, keyValuePair);
		}

		/// <summary>
		/// Adds the specified section.
		/// </summary>
		/// <param name="keyValuePair">The key-value pair of (section, key-value pair table)</param>
		public virtual void Add(KeyValuePair<string, KeyValuePairs> keyValuePair)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			this._dic.Add(keyValuePair.Key, keyValuePair.Value);
		}

		/// <summary>
		/// Clears the table.
		/// </summary>
		public virtual void Clear()
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			this._dic.Clear();
		}

		/// <summary>
		/// Determines if the table contains a specified section name.
		/// </summary>
		/// <param name="sectionName"></param>
		/// <returns></returns>
		public virtual bool Contains(string sectionName)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.ContainsKey(sectionName);
		}

		/// <summary>
		/// Determines if the table contains a specified key-value pair of (section, key-value pair).
		/// </summary>
		/// <param name="pair">The key-value pair table of (section, key-value pair).</param>
		/// <returns></returns>
		public virtual bool Contains(KeyValuePair<string, KeyValuePairs> pair)
		{
			foreach (var keyValuePair in this)
			{
				if (keyValuePair.Key == pair.Key && keyValuePair.Value == pair.Value)
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Determines if the table contains a specified section.
		/// </summary>
		/// <param name="sectionName">The section name.</param>
		/// <returns></returns>
		public virtual bool ContainsKey(string sectionName)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.ContainsKey(sectionName);
		}

		/// <summary>
		/// Not implemented.
		/// </summary>
		/// <param name="keyValuePairs"></param>
		/// <param name="index"></param>
		public virtual void CopyTo(KeyValuePair<string, KeyValuePairs>[] keyValuePairs, int index)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Determines whether the specified object is equal to the current object.
		/// </summary>
		/// <param name="obj">The object to compare with the current object.</param>
		/// <returns></returns>
		public override bool Equals(object obj)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			if (obj is SectionKeyPairs)
			{
				return this._dic.Equals((obj as SectionKeyPairs)._dic);
			}
			else if (obj is Dictionary<string, KeyValuePairs>)
			{
				return this._dic.Equals(obj as Dictionary<string, KeyValuePairs>);
			}
			else if (obj is Dictionary<string, Dictionary<string, string>>)
			{
				return this._dic.Equals(obj as Dictionary<string, Dictionary<string, string>>);
			}

			return false;
		}

		/// <summary>
		/// Returns an enumerator that iterates through the table.
		/// </summary>
		/// <returns></returns>
		IEnumerator<KeyValuePair<string, KeyValuePairs>> IEnumerable<KeyValuePair<string, KeyValuePairs>>.GetEnumerator()
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.GetEnumerator();
		}

		/// <summary>
		/// Returns an enumerator that iterates through the table.
		/// </summary>
		/// <returns></returns>
		IEnumerator IEnumerable.GetEnumerator()
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.GetEnumerator();
		}

		/// <summary>
		/// Returns the hash code of the instance.
		/// </summary>
		/// <returns></returns>
		public override int GetHashCode()
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.GetHashCode();
		}

		/// <summary>
		/// Removes the specified section.
		/// </summary>
		/// <param name="sectionName">The section name to remove.</param>
		/// <returns></returns>
		public virtual bool Remove(string sectionName)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.Remove(sectionName);
		}

		/// <summary>
		/// Removes the specified key-value pair of (section, key-value pair).
		/// </summary>
		/// <param name="keyValuePair">The key-value pair table of (key, value).</param>
		/// <returns></returns>
		public virtual bool Remove(KeyValuePair<string, KeyValuePairs> keyValuePair)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return (this._dic.ContainsKey(keyValuePair.Key) && this._dic[keyValuePair.Key] == keyValuePair.Value) ? this.Remove(keyValuePair.Key) : false;
		}

		/// <summary>
		/// Gets a string that represents the current object.
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.ToString();
		}

		/// <summary>
		/// Gets the key-value pair associated with the specified section.
		/// </summary>
		/// <param name="sectionName">The section name.</param>
		/// <param name="keyValuePairs">The key-value pair table.</param>
		/// <returns></returns>
		public virtual bool TryGetValue(string sectionName, out KeyValuePairs keyValuePairs)
		{
			keyValuePairs = null;

			if (this.ContainsKey(sectionName))
			{
				keyValuePairs = this[sectionName];
				return true;
			}

			return false;
		}

		#endregion Method Overriding

		#region Operator Overriding

		/*
		public static bool operator ==(SectionKeyPairs objA, object objB)
		{
			return objA.Equals(objB);
		}

		public static bool operator !=(SectionKeyPairs objA, object objB)
		{
			return !objA.Equals(objB);
		}
		*/

		#endregion Operator Overriding
	}

	/// <summary>
	/// Provides the methods of key-value pairs.
	/// </summary>
	public class KeyValuePairs : IDictionary<string, string>
	{
		internal Dictionary<string, string> _dic = null;

		internal bool KeepEmptySection { get; set; }

		#region Property Overriding

		/// <summary>
		/// Gets the total number of sections.
		/// </summary>
		public virtual int Count
		{
			get
			{
				// debug assertion
				Debug.Assert(this._dic != null);

				return this._dic.Count;
			}
		}

		/// <summary>
		/// Gets if the table is read-only.
		/// </summary>
		public virtual bool IsReadOnly
		{
			get { return false; }
		}

		/// <summary>
		/// Gets the collection of keys.
		/// </summary>
		public virtual ICollection<string> Keys
		{
			get
			{
				// debug assertion
				Debug.Assert(this._dic != null);

				return this._dic.Keys;
			}
		}

		/// <summary>
		/// Gets the collection of values.
		/// </summary>
		public virtual ICollection<string> Values
		{
			get
			{
				// debug assertion
				Debug.Assert(this._dic != null);

				return this._dic.Values;
			}
		}

		#endregion Property Overriding

		#region Method Overriding

		/// <summary>
		/// Initializes the instance.
		/// </summary>
		public KeyValuePairs()
		{
			this._dic = new Dictionary<string, string>(StringComparer.OrdinalIgnoreCase);
			this.KeepEmptySection = false;
		}

		/// <summary>
		/// Gets or sets the value of the specified key.
		/// </summary>
		/// <param name="keyName">The key name.</param>
		/// <returns></returns>
		public virtual string this[string keyName]
		{
			get
			{
				// debug assertion
				Debug.Assert(this._dic != null);

				keyName = keyName.Trim();
				return this._dic.ContainsKey(keyName) ? this._dic[keyName] : string.Empty;
			}
			set
			{
				// debug assertion
				Debug.Assert(this._dic != null);

				keyName = keyName.Trim();
				this._dic[keyName] = value;
			}
		}

		/// <summary>
		/// Adds the specified key-value pair.
		/// </summary>
		/// <param name="keyName"></param>
		/// <param name="value"></param>
		public virtual void Add(string keyName, string value)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			this._dic.Add(keyName, value);
		}

		/// <summary>
		/// Adds the specified key-value pair.
		/// </summary>
		/// <param name="keyValuePair"></param>
		public virtual void Add(KeyValuePair<string, string> keyValuePair)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			this._dic.Add(keyValuePair.Key, keyValuePair.Value);
		}

		/// <summary>
		/// Clears the table.
		/// </summary>
		public virtual void Clear()
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			this._dic.Clear();
		}

		/// <summary>
		/// Determines if the table contains the specified key.
		/// </summary>
		/// <param name="keyName">The key name.</param>
		/// <returns></returns>
		public virtual bool Contains(string keyName)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.ContainsKey(keyName);
		}

		/// <summary>
		/// Determines if the table contains the specified key-value pair.
		/// </summary>
		/// <param name="pair">The key-value pair.</param>
		/// <returns></returns>
		public virtual bool Contains(KeyValuePair<string, string> pair)
		{
			foreach (var keyValuePair in this)
			{
				if (keyValuePair.Key == pair.Key && keyValuePair.Value == pair.Value)
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Determines if the table contains the specified key.
		/// </summary>
		/// <param name="keyName">The key name.</param>
		/// <returns></returns>
		public virtual bool ContainsKey(string keyName)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.ContainsKey(keyName);
		}

		/// <summary>
		/// Not implemented.
		/// </summary>
		/// <param name="keyValuePairs"></param>
		/// <param name="index"></param>
		public virtual void CopyTo(KeyValuePair<string, string>[] keyValuePairs, int index)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Determines whether the specified object is equal to the current object.
		/// </summary>
		/// <param name="obj">The object to compare with the current object.</param>
		/// <returns></returns>
		public override bool Equals(object obj)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			if (obj is KeyValuePairs)
			{
				return this._dic.Equals((obj as KeyValuePairs)._dic);
			}
			if (obj is Dictionary<string, string>)
			{
				return this._dic.Equals(obj as Dictionary<string, string>);
			}
			return false;
		}

		/// <summary>
		/// Returns an enumerator that iterates through the table.
		/// </summary>
		/// <returns></returns>
		IEnumerator<KeyValuePair<string, string>> IEnumerable<KeyValuePair<string, string>>.GetEnumerator()
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.GetEnumerator();
		}

		/// <summary>
		/// Returns an enumerator that iterates through the table.
		/// </summary>
		/// <returns></returns>
		IEnumerator IEnumerable.GetEnumerator()
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.GetEnumerator();
		}

		/// <summary>
		/// Returns the hash code of the instance.
		/// </summary>
		/// <returns></returns>
		public override int GetHashCode()
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.GetHashCode();
		}

		/// <summary>
		/// Removes the specified key.
		/// </summary>
		/// <param name="keyName">The key name to remove.</param>
		/// <returns></returns>
		public virtual bool Remove(string keyName)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.Remove(keyName);
		}

		/// <summary>
		/// Removes the specified key-value pair.
		/// </summary>
		/// <param name="keyValuePair">The key-value pair to remove.</param>
		/// <returns></returns>
		public virtual bool Remove(KeyValuePair<string, string> keyValuePair)
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return (this._dic.ContainsKey(keyValuePair.Key) && this._dic[keyValuePair.Key] == keyValuePair.Value) ? this.Remove(keyValuePair.Key) : false;
		}

		/// <summary>
		/// Gets a string that represents the current object.
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			// debug assertion
			Debug.Assert(this._dic != null);

			return this._dic.ToString();
		}

		/// <summary>
		/// Gets the value associated with the specified key.
		/// </summary>
		/// <param name="keyName">The key name.</param>
		/// <param name="value">The value.</param>
		/// <returns></returns>
		public bool TryGetValue(string keyName, out string value)
		{
			value = null;

			if (this.ContainsKey(keyName))
			{
				value = this[keyName];
				return true;
			}

			return false;
		}

		#endregion Method Overriding

		#region Operator Overriding

		/*
		public static bool operator ==(KeyValuePairs objA, object objB)
		{
			return objA.Equals(objB);
		}

		public static bool operator !=(KeyValuePairs objA, object objB)
		{
			return !objA.Equals(objB);
		}
		*/

		#endregion Operator Overriding
	}
}
