﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Common
{
	/// <summary>
	/// Provides a task scheduler that ensures a maximum concurrency level while running on top of the thread pool. 
	/// </summary>
	public class LimitedConcurrencyLevelTaskScheduler : TaskScheduler
	{
		// indicates whether the current thread is processing work items.
		[ThreadStatic]
		internal static bool _currentThreadIsProcessingItems;

		// the list of tasks to be executed  
		internal readonly LinkedList<Task> _tasks = new LinkedList<Task>(); // protected by lock(_tasks) 

		// the maximum concurrency level allowed by this scheduler.  
		internal readonly int _maxDegreeOfParallelism;

		// indicates whether the scheduler is currently processing work items.  
		internal int _delegatesQueuedOrRunning = 0;

		/// <summary>
		/// Initializes a new instance of the <see cref="LimitedConcurrencyLevelTaskScheduler"/> class.
		/// </summary>
		/// <param name="maxDegreeOfParallelism">The maximum degree of parallelism.</param>
		/// <exception cref="ArgumentOutOfRangeException">maxDegreeOfParallelism</exception>
		public LimitedConcurrencyLevelTaskScheduler(int maxDegreeOfParallelism)
		{
			if (maxDegreeOfParallelism < 1)
			{
				throw new ArgumentOutOfRangeException("maxDegreeOfParallelism");
			}
			this._maxDegreeOfParallelism = maxDegreeOfParallelism;
		}

		/// <summary>
		/// Queues a <see cref="T:System.Threading.Tasks.Task" /> to the scheduler.
		/// </summary>
		/// <param name="task">The <see cref="T:System.Threading.Tasks.Task" /> to be queued.</param>
		protected sealed override void QueueTask(Task task)
		{
			// add the task to the list of tasks to be processed. If there aren't enough delegates currently queued or running to process tasks, schedule another.  
			lock (this._tasks)
			{
				this._tasks.AddLast(task);
				if (_delegatesQueuedOrRunning < _maxDegreeOfParallelism)
				{
					++_delegatesQueuedOrRunning;
					NotifyThreadPoolOfPendingWork();
				}
			}
		}

		/// <summary>
		/// Informs the ThreadPool that there's work to be executed for this scheduler.  
		/// </summary>
		internal void NotifyThreadPoolOfPendingWork()
		{
			ThreadPool.UnsafeQueueUserWorkItem(_ =>
			{
				// note that the current thread is now processing work items. this is necessary to enable inlining of tasks into this thread.
				_currentThreadIsProcessingItems = true;
				try
				{
					// Process all available items in the queue. 
					while (true)
					{
						Task item;
						lock (this._tasks)
						{
							// when there are no more items to be processed, note that we're done processing, and get out. 
							if (this._tasks.Count == 0)
							{
								--this._delegatesQueuedOrRunning;
								break;
							}

							// get the next item from the queue
							item = this._tasks.First.Value;
							this._tasks.RemoveFirst();
						}

						// execute the task we pulled out of the queue 
						base.TryExecuteTask(item);
					}
				}
				// we're done processing items on the current thread 
				finally
				{
					_currentThreadIsProcessingItems = false;
				}
			}, null);
		}

		/// <summary>
		/// Determines whether the provided <see cref="T:System.Threading.Tasks.Task" /> can be executed synchronously in this call, and if it can, executes it.
		/// </summary>
		/// <param name="task">The <see cref="T:System.Threading.Tasks.Task" /> to be executed.</param>
		/// <param name="taskWasPreviouslyQueued">A Boolean denoting whether or not task has previously been queued. If this parameter is True, then the task may have been previously queued (scheduled); if False, then the task is known not to have been queued, and this call is being made in order to execute the task inline without queuing it.</param>
		/// <returns>
		/// A Boolean value indicating whether the task was executed inline.
		/// </returns>
		protected sealed override bool TryExecuteTaskInline(Task task, bool taskWasPreviouslyQueued)
		{
			// if this thread isn't already processing a task, we don't support inlining 
			if (!_currentThreadIsProcessingItems)
			{
				return false;
			}

			// if the task was previously queued, remove it from the queue
			return (!taskWasPreviouslyQueued || TryDequeue(task)) ? base.TryExecuteTask(task) : false;
		}

		/// <summary>
		/// Attempts to dequeue a <see cref="T:System.Threading.Tasks.Task" /> that was previously queued to this scheduler.
		/// </summary>
		/// <param name="task">The <see cref="T:System.Threading.Tasks.Task" /> to be dequeued.</param>
		/// <returns>
		/// A Boolean denoting whether the <paramref name="task" /> argument was successfully dequeued.
		/// </returns>
		protected sealed override bool TryDequeue(Task task)
		{
			lock (this._tasks)
			{
				return this._tasks.Remove(task);
			}
		}

		/// <summary>
		/// Indicates the maximum concurrency level this <see cref="T:System.Threading.Tasks.TaskScheduler" /> is able to support.
		/// </summary>
		public sealed override int MaximumConcurrencyLevel { get { return this._maxDegreeOfParallelism; } }

		/// <summary>
		/// For debugger support only, generates an enumerable of <see cref="T:System.Threading.Tasks.Task" /> instances currently queued to the scheduler waiting to be executed.
		/// </summary>
		/// <returns>
		/// An enumerable that allows a debugger to traverse the tasks currently queued to this scheduler.
		/// </returns>
		/// <exception cref="NotSupportedException"></exception>
		protected sealed override IEnumerable<Task> GetScheduledTasks()
		{
			bool lockTaken = false;
			try
			{
				Monitor.TryEnter(this._tasks, ref lockTaken);
				if (lockTaken)
				{
					return this._tasks;
				}
				else
				{
					throw new NotSupportedException();
				}
			}
			finally
			{
				if (lockTaken)
				{
					Monitor.Exit(this._tasks);
				}
			}
		}
	}
}
