﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media.Imaging;

namespace AirFusionCustomNotification
{
	/// <summary>
	/// Interaction logic for CustomNotificationView.xaml
	/// </summary>
	public partial class CustomNotificationView : Window, System.ComponentModel.INotifyPropertyChanged
	{
		#region ==================== Members ====================

		// _typeNameofCNViewWindow is used to recognize the custom notification from windows 
		private readonly string _typeNameofCNViewWindow = typeof(CustomNotificationView).Name;

		// _paddingRight and _paddingTop are composed the specific margin of each custom notification
		private double _paddingRight = 7;
		private double _paddingTop = 7;

		// _sumofDynamicButtons indicates the sum of the dynamically-created buttons
		private int _sumofDynamicButtons = 0;

		#endregion

		#region ==================== Constructors ====================

		public CustomNotificationView()
			: base()
		{
			InitializeComponent();
		}

		#endregion

		#region ==================== Events & Event Arguments ====================

		public class CNViewWindowActivatedEventArgs :EventArgs
		{
			public string NotificationKey { get; internal set; }
		}

		private delegate void CNViewWindowActivatedEventHandler(object sender, CNViewWindowActivatedEventArgs args);

		/// <summary>
		/// Triggered when the custom notification is being clicked
		/// </summary>
		private event CNViewWindowActivatedEventHandler CNViewWindowActivated;

		#endregion

		#region ==================== Event handlers ====================

		private void StoryboardLoadCompleted(object sender, System.EventArgs e)
		{
			if (!this.IsMouseOver && !this.lvQuickActionsList.IsMouseOver)
			{
				this.Close();
			}
		}

		private void StoryboardFadeCompleted(object sender, System.EventArgs e)
		{
			if (!this.IsMouseOver && !this.lvQuickActionsList.IsMouseOver)
			{
				this.Close();
			}
		}

		private void CNViewWindow_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
		{
			if (this.CNViewWindowActivated != null)
			{
				this.CNViewWindowActivated(this, new CNViewWindowActivatedEventArgs()
				{
					NotificationKey = this.NotificationKey,
				});
			}
		}

		private void imgClose_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
		{
			this.Close();
		}

		private void imgClose_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
		{
		}

		protected override void OnClosed(System.EventArgs e)
		{
			base.OnClosed(e);
		}

		#endregion

		#region ==================== Methods ====================

		/// <summary>
		/// Override the default Show() to creat the window with specific position and deal with the issue of keyboard focus
		/// </summary>
		public new void Show()
		{
			var lastFocusedElement = Keyboard.FocusedElement;

			base.Show();
			this.Owner = System.Windows.Application.Current.MainWindow;
			this.InitialPositionofCustomNotification();

			if (lastFocusedElement != null)
			{
				// Restore keyboard focus to the original element that had focus. That way if someone
				// was typing into a control we don't steal keyboard focus away from that control.
				lastFocusedElement.Focusable = true;
				Keyboard.Focus(lastFocusedElement);
			}
		}

		/// <summary>
		/// Override the default Close()
		/// </summary>
		public new void Close()
		{
			base.Close();
		}

		/// <summary>
		/// Show DynamicButtonField if the numebr of buttons is more than or equal to one
		/// Creat fit-sized buttons in DynamicButtonField according to the numebr of buttons if the numebr of buttons is less than three
		///	Creat a popup list to locate the rest of buttons and a button to open the popup list while the numebr of buttons is more than or equal to three
		/// </summary>
		/// <param name="icon">The icon of each button</param>
		/// <param name="name">The content of each button</param>
		/// <param name="action">The action for button click</param>
		public void RefreshDynamicButtonField(string name, Action action)
		{
			this._sumofDynamicButtons++;

			if (this._sumofDynamicButtons > 0)
			{
				if (this._sumofDynamicButtons <= 2)
				{
					ColumnDefinition _c = new ColumnDefinition();
					_c.Width = new GridLength(1, GridUnitType.Star);
					this.DynamicButtonField.ColumnDefinitions.Add(_c);

					if (this._sumofDynamicButtons == 1)
					{
						this.DynamicButtonField.Visibility = Visibility.Visible;
						this.button1 = this.SetDynamicButton(this.button1, icon, name, action);
						Grid.SetColumn(this.button1, 0);
					}
					else
					{
						this.button2.Visibility = Visibility.Visible;
						this.button2 = this.SetDynamicButton(this.button2, icon, name, action);
						Grid.SetColumn(this.button2, 1);
					}
				}
				else
				{
					Border _btn = this.SetDynamicButton(null, icon, name, action);
					this.lvQuickActionsList.Items.Add(_btn);

					if (this._sumofDynamicButtons == 3)
					{
						ColumnDefinition _c = new ColumnDefinition();
						_c.Width = new GridLength(45);
						this.DynamicButtonField.ColumnDefinitions.Add(_c);

						this.buttonPopUpButtonsList.Visibility = Visibility.Visible;
						this.buttonPopUpButtonsList.MouseLeftButtonUp += (s, e) =>
						{
							this.popUpButtonsList.IsOpen = true;
							foreach (Border border in this.lvQuickActionsList.Items)
							{
								border.Width = this.button2.ActualWidth;
								border.Height = this.button2.ActualHeight;
							}
						};
						Grid.SetColumn(this.buttonPopUpButtonsList, 2);
					}
				}
			}
		}

		private Border SetDynamicButton(Border targetButton, BitmapSource icon, string name, Action action)
		{
			Image _i = new Image()
			{
				Visibility = (icon == null) ? Visibility.Collapsed : Visibility.Visible,
				VerticalAlignment = System.Windows.VerticalAlignment.Center,
				HorizontalAlignment = System.Windows.HorizontalAlignment.Center,
				Source = icon,
				Margin = new Thickness(0,0,8,0),
			};

			TextBlock _t = new TextBlock()
			{
				VerticalAlignment = System.Windows.VerticalAlignment.Center,
				HorizontalAlignment = System.Windows.HorizontalAlignment.Center,
				Text = name,
			};

			StackPanel _s = new StackPanel()
			{
				Orientation = Orientation.Horizontal,
				Children = { _i, _t, },
			};

			if (targetButton == null) targetButton = new Border();

			targetButton.Child = _s;
			targetButton.Style = this.DynamicButtonField.FindResource("DynamicButtonStyle") as Style;

			targetButton.MouseLeftButtonUp += (s, e) =>
			{
				action.BeginInvoke(delegate(IAsyncResult result)
				{
					try
					{
						action.EndInvoke(result);
					}
					catch (Exception)
					{
					}
				}, null);
			};

			return targetButton;
		}

		private void InitialPositionofCustomNotification()
		{
			this.Topmost = true;

			double _left = SystemParameters.WorkArea.Right - (this.ActualWidth + this._paddingRight);
			double _top = SystemParameters.WorkArea.Top + this._paddingTop;

			foreach (Window _window in System.Windows.Application.Current.Windows)
			{
				string _typeNameofWindow = _window.GetType().Name;

				if (_typeNameofWindow.Equals(this._typeNameofCNViewWindow) && _window != this)
				{
					_window.Topmost = true;

					if (_left >= _window.Left)
					{
						_left = _window.Left;

						if (_top <= _window.Top)
						{
							_top = _window.Top + (this.ActualHeight + this._paddingTop);

							if ((_top + this.ActualHeight) > SystemParameters.WorkArea.Bottom)
							{
								_left -= (this.ActualWidth + this._paddingRight);
								_top = SystemParameters.WorkArea.Top + this._paddingTop;

								if (_left < SystemParameters.WorkArea.Left)
								{
									_left = SystemParameters.WorkArea.Right - (this.ActualWidth + this._paddingRight);
								}
							}
						}
					}
				}
			}

			this.Left = _left;
			this.Top = _top;
		}

		#endregion

		#region ==================== Data binding ====================

		// The window width of the custom notification
		private string _notificationWidth = string.Empty;
		public string NotificationWidth
		{
			get
			{
				return _notificationWidth;
			}

			set
			{
				_notificationWidth = value;
				this.OnPropertyChanged("NotificationWidth");
			}
		}

		// The window height of the custom notification
		private string _notificationHeight = string.Empty;
		public string NotificationHeight
		{
			get
			{
				return _notificationHeight;
			}

			set
			{
				_notificationHeight = value;
				this.OnPropertyChanged("NotificationHeight");
			}
		}

		// The icon of the custom notification
		private BitmapSource _notificationIcon = null;
		public BitmapSource NotificationIcon
		{
			get
			{
				return _notificationIcon;
			}

			set
			{
				_notificationIcon = value;
				this.OnPropertyChanged("NotificationIcon");
			}
		}

		// The title of the custom notification
		private string _notificationTitle = string.Empty;
		public string NotificationTitle
		{
			get
			{
				return _notificationTitle;
			}

			set
			{
				_notificationTitle = value;
				this.OnPropertyChanged("NotificationTitle");
			}
		}

		// The content of the custom notification
		private string _notificationContent = string.Empty;
		public string NotificationContent
		{
			get
			{
				return _notificationContent;
			}

			set
			{
				_notificationContent = value;
				this.OnPropertyChanged("NotificationContent");
			}
		}

		// The infomation including device and time about the custom notification
		private string _notificationInfomation = string.Empty;
		public string NotificationInfomation
		{
			get
			{
				return _notificationInfomation;
			}

			set
			{
				_notificationInfomation = value;
				this.OnPropertyChanged("NotificationInfomation");
			}
		}

		// The time to automatically close the custom notification
		private string _timeToCloseNotification = string.Empty;
		public string TimeToCloseNotification
		{
			get
			{
				return _timeToCloseNotification;
			}

			set
			{
				_timeToCloseNotification = value;
				this.OnPropertyChanged("TimeToCloseNotification");
			}
		}

		// The key to identify the custom notification
		private string _notificationKey = string.Empty;
		public string NotificationKey
		{
			get
			{
				return _notificationKey;
			}

			set
			{
				_notificationKey = value;
			}
		}

		public event System.ComponentModel.PropertyChangedEventHandler PropertyChanged;
		protected virtual void OnPropertyChanged(string targetName)
		{
			if (this.PropertyChanged != null)
			{
				this.PropertyChanged(this, new System.ComponentModel.PropertyChangedEventArgs(targetName));
			}
		}

		#endregion
	}
}
