﻿using AirFusion.Windows;
using System;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Forms;
using System.Windows.Input;
using System.Windows.Interop;
using System.Windows.Media;
using System.Windows.Media.Animation;

namespace AirFusionCustomToast
{
	public partial class GeneralToastPopUp : Window, System.ComponentModel.INotifyPropertyChanged
	{
		#region Members

		private bool _isRemovableFromPhone = false;

		// _paddingRight and _paddingTop are composed the specific margin of each custom notification

		private const double _paddingRight = 7;
		private const double _paddingTop = 7;

		private string title;
		private GeneralNotificationUIElement _currentNotification = null;

		public GeneralNotificationUIElement CurrentNotification
		{
			get
			{
				return _currentNotification;
			}

			set
			{
				_currentNotification = value;
				OnPropertyChanged("CurrentNotification");
			}
		}

		private Storyboard _storyboardLoad;
		private Storyboard _storyboardFade;

		#endregion Members

		#region Constructors

		private GeneralToastPopUp(string title)
			: base()
		{
			this.InitializeComponent();

			this.DataContext = this;
			this.tbTitle.Text = title;
			this.title = title;
			System.Windows.Application.Current.MainWindow.Closing += this.MainWindowClosing;

			this._storyboardLoad = this.FindResource("StoryboardLoad") as Storyboard;
			this._storyboardFade = this.FindResource("StoryboardFade") as Storyboard;
		}

		public GeneralToastPopUp(GeneralNotificationUIElement notificationItem)
			: this(notificationItem.Title)
		{
			this.CurrentNotification = notificationItem;
			this._isRemovableFromPhone = notificationItem.Removable;

			if (this.CurrentNotification.IsAllowMute == false)
				this.btnMute.Visibility = Visibility.Collapsed;

			//GenerateQuickActionButtons();
			GenerateDynamicButtonsField();
		}

		public GeneralToastPopUp(GeneralNotificationUIElement notificationItem, string filePath)
			: this(notificationItem.Title)
		{
			this.CurrentNotification = notificationItem;

			try
			{
				this.GenerateSnapshotDynamicButtonsField(filePath);
				this.btnMute.Visibility = Visibility.Collapsed;
			}
			catch (Exception ex)
			{
				string str = ex.ToString();
			}
		}

		private void GenerateSnapshotDynamicButtonsField(string filePath)
		{
			this.DynamicButtonField.Visibility = Visibility.Visible;
			this.btnOpenSnapShotFile.Visibility = Visibility.Visible;
			this.btnOpenSnapShotFolder.Visibility = Visibility.Visible;

			this.btnOpenSnapShotFile.Click += (s, e) =>
			{
				if (!String.IsNullOrEmpty(filePath) && System.IO.File.Exists(filePath))
				{
					System.Diagnostics.Process prc = new System.Diagnostics.Process();
					prc.StartInfo.FileName = filePath;
					prc.StartInfo.UseShellExecute = true;

					try
					{
						prc.Start();
					}
					catch
					{
					}
				}

				this.Close();
			};

			this.btnOpenSnapShotFolder.Click += (s, e) =>
			{
				if (!String.IsNullOrEmpty(filePath))
				{
					if (System.IO.Directory.Exists(System.IO.Path.GetDirectoryName(filePath)))
					{
						System.Diagnostics.Process prc = new System.Diagnostics.Process();
						prc.StartInfo.FileName = System.IO.Path.GetDirectoryName(filePath);
						prc.StartInfo.UseShellExecute = true;

						try
						{
							prc.Start();
						}
						catch
						{
						}
					}
				}

				this.Close();
			};

			this.DynamicButtonField.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Star) });
			Grid.SetColumn(this.btnOpenSnapShotFile, 0);
			Grid.SetColumn(this.btnOpenSnapShotFolder, 1);
		}

		#endregion Constructors

		#region Events Arguments & Events

		/// <summary>
		/// Represents the basic event arguments for the triggered event when user clicks the popup toast
		/// </summary>
		public class CustomToastEventRawArgs : EventArgs
		{
			/// <summary>
			/// Gets the notification id
			/// </summary>
			public string NotificationKey { get; internal set; }

			/// <summary>
			/// Gets the device id
			/// </summary>
			public Guid DeviceId { get; internal set; }
		}

		/// <summary>
		/// Represents the event arguments for the triggered event when user clicks the quick action button on the popup toast
		/// </summary>
		public class CustomToastQuickActionEventArgs : CustomToastEventRawArgs
		{
			/// <summary>
			/// Gets the id of the quick action
			/// </summary>
			public int QuickActionId { get; internal set; }

			/// <summary>
			/// Gets if the quick action has a popup content intent on phone
			/// </summary>
			public bool QuickActionActivityPopup { get; internal set; }
		}

		/// <summary>
		/// Represents the event arguments for the triggerred event when user clicks the local launch button on the popup toast
		/// </summary>
		public class CustomToastLocalLaunchEventArgs : CustomToastEventRawArgs
		{
			/// <summary>
			/// Gets the handler for the local launch button
			/// </summary>
			public AirFusion.Manager.Notification.LocalLaunch LocalLaunchHandler { get; internal set; }

			/// <summary>
			/// Gets the id for the local launch button
			/// </summary>
			public int LocalLaunchIndex { get; internal set; }
		}

		/// <summary>
		/// Represents the methods that handles the event when user clicks the toast
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args">CustomToastEventRawArgs</param>
		public delegate void AFCustomToastActivatedEventHandler(object sender, CustomToastEventRawArgs args);

		/// <summary>
		/// Occurs when user clicks the toast
		/// </summary>
		public event AFCustomToastActivatedEventHandler CustomToastActivated;

		/// <summary>
		/// Represents the methods that handles the event when user clicks the launch button of the toast
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args">CustomToastEventRawArgs</param>
		public delegate void AFCustomToastLaunchedEventHandler(object sender, CustomToastEventRawArgs args);

		/// <summary>
		/// Occurs when user clicks the lauch button of the toast
		/// </summary>
		public event AFCustomToastLaunchedEventHandler CustomToastLaunched;

		/// <summary>
		/// Represents the methods that handles the event when user clicks the dismiss button of the toast
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args">CustomToastEventRawArgs</param>
		public delegate void AFCustomToastDismissedEventHandler(object sender, CustomToastEventRawArgs args);

		/// <summary>
		/// Represents the methods that handles the event when the toast closed
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args">CustomToastEventRawArgs</param>
		public delegate void AFCustomToastWindowClosedEventHandler(object sender, CustomToastEventRawArgs args);

		/// <summary>
		/// Occurs when user clicks the dismiss button of the toast
		/// </summary>
		public event AFCustomToastDismissedEventHandler CustomToastDismissed;

		/// <summary>
		/// Occurs when the toast closed
		/// </summary>
		public event AFCustomToastDismissedEventHandler CustomToastWindowClosed;

		/// <summary>
		/// Represents the methods that handles the event when user clicks the ignore button of the toast
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args">CustomToastEventRawArgs</param>
		public delegate void AFCustomToastIgnoreEventHandler(object sender, CustomToastEventRawArgs args);

		/// <summary>
		/// Occurs when user clicks the ignore button of the toast
		/// </summary>
		public event AFCustomToastIgnoreEventHandler CustomToastIgnore;

		/// <summary>
		/// Represents the methods that handles the event when user clicks the quick action buttons of the toast
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args">CustomToastQuickActionEventArgs</param>
		public delegate void AFCustomToastQuickActionActivatedEventHandler(object sender, CustomToastQuickActionEventArgs args);

		/// <summary>
		/// Occurs when user clicks the quick action buttons of the toast
		/// </summary>
		public event AFCustomToastQuickActionActivatedEventHandler CustomToastQuickActionActivated;

		/// <summary>
		/// Represents the methods that handles the event when user clicks the local launch button of the toast
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="args">CustomToastLocalLaunchEventArgs</param>
		public delegate void AFCustomToastLocalLaunchEventHandler(object sender, CustomToastLocalLaunchEventArgs args);

		/// <summary>
		/// Occurs when user clicks the local launch button of the toast
		/// </summary>
		public event AFCustomToastLocalLaunchEventHandler CustomToastLocalLaunch;

		#endregion Events Arguments & Events

		#region Public Methods

		/// <summary>
		/// Opens a window and returns without waiting for the newly opened window to close.
		/// </summary>
		public new void Show()
		{
			var focusedElement = Keyboard.FocusedElement;

			base.Show();

			try
			{
				this.Owner = System.Windows.Application.Current.MainWindow;
			}
			catch (Exception ex)
			{ }

			try
			{
				if (!GeneralNotificationHelper.ScreenDetectionn.isForegroundWindowFullScreen())
				{
					this.Topmost = true;
					this.Topmost = false;
				}
			}
			catch (Exception ex)
			{ }

			if (focusedElement != null)
			{
				// Restore keyboard focus to the original element that had focus. That way if someone
				// was typing into a control we don't steal keyboard focus away from that control.
				focusedElement.Focusable = true;
				Keyboard.Focus(focusedElement);
			}
		}

		#endregion Public Methods

		#region Private Methods

		//private void GenerateQuickActionButtons()
		//{
		//	//Grid
		//	if (CurrentNotification.QAItemList != null && CurrentNotification.QAItemList.Count > 0)
		//	{
		//		grdNotificationQAButtons.Visibility = System.Windows.Visibility.Visible;

		//		for (int i = 0; i < CurrentNotification.QAItemList.Count; ++i)
		//		{
		//			//Grid
		//			grdNotificationQAButtons.ColumnDefinitions.Add(
		//				new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Star) });

		//			//TextBlock
		//			TextBlock txtBlock = new TextBlock()
		//			{
		//				VerticalAlignment = System.Windows.VerticalAlignment.Center,
		//				HorizontalAlignment = System.Windows.HorizontalAlignment.Center,
		//				Margin = new Thickness(5, 0, 0, 0),
		//				Text = CurrentNotification.QAItemList[i].QATitle,
		//				Foreground = new SolidColorBrush(Colors.Green),
		//				TextAlignment = TextAlignment.Center,
		//			};

		//			//Image
		//			System.Windows.Controls.Image img = new System.Windows.Controls.Image();
		//			if (!string.IsNullOrEmpty(CurrentNotification.QAItemList[i].QAIconPath) &&
		//				System.IO.File.Exists(CurrentNotification.QAItemList[i].QAIconPath))
		//			{
		//				Bitmap bitmap = new Bitmap(System.Drawing.Image.FromFile(CurrentNotification.QAItemList[i].QAIconPath));
		//				System.Windows.Media.Imaging.BitmapSource bitmapSource = System.Windows.Interop.Imaging.CreateBitmapSourceFromHBitmap(
		//						bitmap.GetHbitmap(),
		//						IntPtr.Zero,
		//						System.Windows.Int32Rect.Empty,
		//						System.Windows.Media.Imaging.BitmapSizeOptions.FromWidthAndHeight(
		//							bitmap.Width,
		//							bitmap.Height));
		//				bitmapSource.Freeze();
		//				img.Source = bitmapSource;
		//			}
		//			else
		//			{
		//				img.Source = null;
		//			}

		//			//StackPanel
		//			StackPanel skPanel = new StackPanel()
		//			{
		//				Orientation = System.Windows.Controls.Orientation.Horizontal,
		//				Children =
		//				{
		//					img,
		//					txtBlock
		//				}
		//			};

		//			//Button
		//			System.Windows.Controls.Button customBtn = new System.Windows.Controls.Button
		//			{
		//				Background = new SolidColorBrush(System.Windows.Media.Color.FromRgb(0xe4, 0xe4, 0xe4)),
		//				BorderBrush = new SolidColorBrush(System.Windows.Media.Color.FromRgb(0xD4, 0xD4, 0xD4)),
		//				BorderThickness = new Thickness(1, 1, 1, 1),
		//				Content = skPanel,
		//				Tag = new CustomToastQuickActionEventArgs()
		//				{
		//					NotificationKey = CurrentNotification.NotificationKey,
		//					QuickActionId = CurrentNotification.QAItemList[i].QAIndex.Value,
		//					QuickActionActivityPopup = CurrentNotification.QAItemList[i].QAActivityPopup,
		//					DeviceId = CurrentNotification.DeviceInstanceID
		//				}
		//			};
		//			customBtn.Click += new RoutedEventHandler((sender, args) =>
		//			{
		//				if (this.CustomToastQuickActionActivated != null)
		//				{
		//					object customInfo = (sender as System.Windows.Controls.Button).Tag;
		//					if (customInfo != null)
		//					{
		//						CustomToastQuickActionEventArgs eventArg = customInfo as CustomToastQuickActionEventArgs;
		//						this.CustomToastQuickActionActivated(this, eventArg);
		//					}
		//				}
		//			});

		//			Grid.SetColumn(customBtn, i);
		//			grdNotificationQAButtons.Children.Add(customBtn);
		//		}
		//	}
		//}

		private void GenerateDynamicButtonsField()
		{
			this.DynamicButtonField.Visibility = Visibility.Collapsed;

			int sumofDynamicButtons = 0;
			this.DynamicButtonField.Visibility = Visibility.Visible;

			switch (CurrentNotification.Type)
			{
				case AirFusion.Manager.Notification.NotificationType.Basic:
					{
						if (CurrentNotification.QAItemList.Count == 0)
							this.DynamicButtonField.Visibility = Visibility.Collapsed;

						break;
					}
				case AirFusion.Manager.Notification.NotificationType.BasicWithLaunch:
					{
						if (CurrentNotification.QAItemList.Count == 0)
							this.DynamicButtonField.Visibility = Visibility.Collapsed;

						break;
					}
				case AirFusion.Manager.Notification.NotificationType.FileShare:
					{
						this.btnOpenFile.Visibility = Visibility.Visible;
						this.setPositionofButton(this.btnOpenFile, sumofDynamicButtons, false, false);
						sumofDynamicButtons++;
						this.btnOpenFolder.Visibility = Visibility.Visible;
						this.setPositionofButton(this.btnOpenFolder, sumofDynamicButtons, false, false);
						sumofDynamicButtons++;

						break;
					}
				case AirFusion.Manager.Notification.NotificationType.FileShareWithLaunch:
					{
						this.btnOpenFile.Visibility = Visibility.Visible;
						this.setPositionofButton(this.btnOpenFile, sumofDynamicButtons, false, false);
						sumofDynamicButtons++;
						this.btnOpenFolder.Visibility = Visibility.Visible;
						this.setPositionofButton(this.btnOpenFolder, sumofDynamicButtons, false, false);
						sumofDynamicButtons++;

						break;
					}
				case AirFusion.Manager.Notification.NotificationType.UrlShare:
					{
						this.btnLocalLaunch.Visibility = Visibility.Visible;
						this.setPositionofButton(this.btnLocalLaunch, sumofDynamicButtons, false, false);
						sumofDynamicButtons++;

						break;
					}
				case AirFusion.Manager.Notification.NotificationType.UrlShareWithLaunch:
					{
						this.btnLocalLaunch.Visibility = Visibility.Visible;
						this.setPositionofButton(this.btnLocalLaunch, sumofDynamicButtons, false, false);
						sumofDynamicButtons++;

						break;
					}
				default:
					break;
			}

			if (this.CurrentNotification.LocalLaunchHandler != null && CurrentNotification.Type != AirFusion.Manager.Notification.NotificationType.FileShare
				&& CurrentNotification.Type != AirFusion.Manager.Notification.NotificationType.UrlShare
				&& CurrentNotification.Type != AirFusion.Manager.Notification.NotificationType.FileShareWithLaunch
				&& CurrentNotification.Type != AirFusion.Manager.Notification.NotificationType.UrlShareWithLaunch)
			{
				System.Windows.Controls.Button _button = new System.Windows.Controls.Button();
				_button.Click += btnLocalLaunch_Click;
				_button.Content = btnLocalLaunch.Content;
				_button.Style = this.DynamicButtonField.FindResource("DynamicButtonStyle") as Style;

				this.setPositionofButton(_button, sumofDynamicButtons, true, false);
				sumofDynamicButtons++;
			}

			if (CurrentNotification.QAItemList != null && CurrentNotification.QAItemList.Count > 0)
			{
				for (int i = 0; i < CurrentNotification.QAItemList.Count; ++i)
				{
					////TextBlock
					//TextBlock txtBlock = new TextBlock()
					//{
					//	//VerticalAlignment = System.Windows.VerticalAlignment.Center,
					//	//HorizontalAlignment = System.Windows.HorizontalAlignment.Center,
					//	//Margin = new Thickness(5, 0, 0, 0),
					//	Text = CurrentNotification.QAItemList[i].QATitle,
					//	//Foreground = new SolidColorBrush(Colors.Green),
					//	//TextAlignment = TextAlignment.Center,
					//};

					//Image
					System.Windows.Controls.Image img = new System.Windows.Controls.Image();
					if (!string.IsNullOrEmpty(CurrentNotification.QAItemList[i].QAIconPath) &&
						System.IO.File.Exists(CurrentNotification.QAItemList[i].QAIconPath))
					{
						Bitmap bitmap = new Bitmap(System.Drawing.Image.FromFile(CurrentNotification.QAItemList[i].QAIconPath));
						System.Windows.Media.Imaging.BitmapSource bitmapSource = System.Windows.Interop.Imaging.CreateBitmapSourceFromHBitmap(
							bitmap.GetHbitmap(),
							IntPtr.Zero,
							System.Windows.Int32Rect.Empty,
							System.Windows.Media.Imaging.BitmapSizeOptions.FromWidthAndHeight(
							bitmap.Width,
							bitmap.Height));
						bitmapSource.Freeze();
						img.Source = bitmapSource;
					}
					else
					{
						img.Source = null;
					}

					////StackPanel
					//StackPanel skPanel = new StackPanel()
					//{
					//	Orientation = System.Windows.Controls.Orientation.Horizontal,
					//	VerticalAlignment = System.Windows.VerticalAlignment.Center,
					//	HorizontalAlignment = System.Windows.HorizontalAlignment.Center,
					//	Children =
					//	{
					//		img,
					//		txtBlock
					//	}
					//};

					//Button
					System.Windows.Controls.Button _button = new System.Windows.Controls.Button()
					{
						Content = CurrentNotification.QAItemList[i].QATitle,
						Style = this.DynamicButtonField.FindResource("DynamicButtonStyle") as Style,
						Tag = new CustomToastQuickActionEventArgs()
						{
							NotificationKey = CurrentNotification.NotificationKey,
							QuickActionId = CurrentNotification.QAItemList[i].QAIndex.Value,
							QuickActionActivityPopup = CurrentNotification.QAItemList[i].QAActivityPopup,
							DeviceId = CurrentNotification.DeviceInstanceID
						},
						Visibility = Visibility.Visible,
					};

					if (CurrentNotification.isMediaStyle && img != null)
					{
						_button.Style = this.DynamicButtonField.FindResource("DynamicButtonMediaStyle") as Style;
						_button.Content = img;
					}

					_button.Click += new RoutedEventHandler((sender, args) =>
					{
						if (this.CustomToastQuickActionActivated != null)
						{
							object customInfo = (sender as System.Windows.Controls.Button).Tag;
							if (customInfo != null)
							{
								CustomToastQuickActionEventArgs eventArg = customInfo as CustomToastQuickActionEventArgs;
								this.CustomToastQuickActionActivated(this, eventArg);
							}
						}

						this.Close();
					});

					this.setPositionofButton(_button, sumofDynamicButtons, true, this.CurrentNotification.isMediaStyle);
					sumofDynamicButtons++;
				}
			}
		}

		private void setPositionofButton(System.Windows.Controls.Button button, int sumofDynamicButtons, bool isDynamicallyCreated, bool isMediaStyle)
		{
			if (sumofDynamicButtons == 0)
			{
				if (isDynamicallyCreated)
					this.DynamicButtonField.Children.Add(button);
				Grid.SetColumn(button, sumofDynamicButtons);

				button.VerticalAlignment = System.Windows.VerticalAlignment.Stretch;
				button.HorizontalAlignment = System.Windows.HorizontalAlignment.Stretch;
				button.VerticalContentAlignment = System.Windows.VerticalAlignment.Center;
				button.HorizontalContentAlignment = System.Windows.HorizontalAlignment.Center;
			}
			else if (sumofDynamicButtons == 1)
			{
				this.DynamicButtonField.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Star) });
				if (isDynamicallyCreated)
					this.DynamicButtonField.Children.Add(button);
				Grid.SetColumn(button, sumofDynamicButtons);

				button.VerticalAlignment = System.Windows.VerticalAlignment.Stretch;
				button.HorizontalAlignment = System.Windows.HorizontalAlignment.Stretch;
				button.VerticalContentAlignment = System.Windows.VerticalAlignment.Center;
				button.HorizontalContentAlignment = System.Windows.HorizontalAlignment.Center;
			}
			else
			{
				if (isMediaStyle)
				{
					if (sumofDynamicButtons == 2)
					{
						this.DynamicButtonField.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Star) });
						if (isDynamicallyCreated)
							this.DynamicButtonField.Children.Add(button);
						Grid.SetColumn(button, sumofDynamicButtons);

						button.VerticalAlignment = System.Windows.VerticalAlignment.Stretch;
						button.HorizontalAlignment = System.Windows.HorizontalAlignment.Stretch;
						button.VerticalContentAlignment = System.Windows.VerticalAlignment.Center;
						button.HorizontalContentAlignment = System.Windows.HorizontalAlignment.Center;
					}
					else
					{
						if (sumofDynamicButtons == 3)
						{
							this.DynamicButtonField.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(45) });

							this.buttonPopUpButtonsList.Visibility = Visibility.Visible;
							this.buttonPopUpButtonsList.Click += (s, e) =>
							{
								this.popUpButtonsList.IsOpen = !this.popUpButtonsList.IsOpen;
							};
							Grid.SetColumn(this.buttonPopUpButtonsList, sumofDynamicButtons);
						}

						button.Height = 40;
						this.stkpQuickActionsList.Children.Add(button);

						button.VerticalAlignment = System.Windows.VerticalAlignment.Stretch;
						button.HorizontalAlignment = System.Windows.HorizontalAlignment.Stretch;
						button.VerticalContentAlignment = System.Windows.VerticalAlignment.Center;
						button.HorizontalContentAlignment = System.Windows.HorizontalAlignment.Center;
					}
				}
				else
				{
					if (sumofDynamicButtons == 2)
					{
						this.DynamicButtonField.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(45) });

						this.buttonPopUpButtonsList.Visibility = Visibility.Visible;
						this.buttonPopUpButtonsList.Click += (s, e) =>
						{
							this.popUpButtonsList.IsOpen = !this.popUpButtonsList.IsOpen;
						};
						Grid.SetColumn(this.buttonPopUpButtonsList, sumofDynamicButtons);
					}

					button.Height = 40;
					this.stkpQuickActionsList.Children.Add(button);

					button.VerticalAlignment = System.Windows.VerticalAlignment.Stretch;
					button.HorizontalAlignment = System.Windows.HorizontalAlignment.Stretch;
					button.VerticalContentAlignment = System.Windows.VerticalAlignment.Center;
					button.HorizontalContentAlignment = System.Windows.HorizontalAlignment.Center;
				}
			}
		}

		/// <summary>
		/// Adjusts the windows.
		/// </summary>
		private void AdjustWindows()
		{
			double left = SystemParameters.WorkArea.Right - (this.ActualWidth + _paddingRight);
			double top = SystemParameters.WorkArea.Top + _paddingTop;

			foreach (Window window in System.Windows.Application.Current.Windows)
			{
				if (window == null || !window.GetType().Equals(typeof(GeneralToastPopUp)) || window == this) continue;

				if (window.Title.Equals(this.CurrentNotification.NotificationKey))
				{
					this.Left = window.Left;
					this.Top = window.Top;

					this.Dispatcher.BeginInvoke(System.Windows.Threading.DispatcherPriority.Background, (Action)delegate()
					{
						window.Close();
					});

					this._storyboardFade.Begin();

					return;
				}
				else if ((left + this.ActualWidth) > window.Left)
				{
					left = window.Left;

					if (top < (window.Top + window.ActualHeight))
					{
						top = window.Top + (window.ActualHeight + _paddingTop);

						if ((top + this.ActualHeight) > SystemParameters.WorkArea.Bottom)
						{
							left -= (this.ActualWidth + _paddingRight);
							top = SystemParameters.WorkArea.Top + _paddingTop;

							if (left < SystemParameters.WorkArea.Left)
							{
								left = SystemParameters.WorkArea.Right - (this.ActualWidth + _paddingRight);
							}
						}
					}
				}
			}

			this._storyboardLoad.Begin();

			this.Left = left;
			this.Top = top;

			return;
		}

		public event System.ComponentModel.PropertyChangedEventHandler PropertyChanged;

		protected virtual void OnPropertyChanged(string propertyName)
		{
			if (this.PropertyChanged != null)
			{
				this.PropertyChanged(this, new System.ComponentModel.PropertyChangedEventArgs(propertyName));
			}
		}

		private string AdjustDateTime(int now)
		{
			if (now < 10)
			{
				return ("0" + now.ToString());
			}
			else
			{
				return now.ToString();
			}
		}

		#endregion Private Methods

		#region UI controls event

		//private void btnLaunch_Click(object sender, RoutedEventArgs e)
		//{
		//    if (this.CustomToastLaunched != null)
		//    {
		//        this.CustomToastLaunched(this, new CustomToastEventRawArgs()
		//        {
		//            NotificationKey = CurrentNotification.NotificationKey,
		//            DeviceId = CurrentNotification.DeviceInstanceID,
		//        });
		//    }

		//    this.Close();
		//}

		private void btnClose_Click(object sender, RoutedEventArgs e)
		{
			//if (this._isRemovableFromPhone)
			//{
			//	if (this.CustomToastDismissed != null)
			//	{
			//		this.CustomToastDismissed(this, new CustomToastEventRawArgs()
			//		{
			//			NotificationKey = CurrentNotification.NotificationKey,
			//			DeviceId = CurrentNotification.DeviceInstanceID
			//		});
			//	}
			//}

			this.Close();
		}

		private void GeneralNotificationWindow_Closed(object sender, EventArgs e)
		{
			if (this.CustomToastWindowClosed != null)
			{
				this.CustomToastWindowClosed(this, new CustomToastEventRawArgs()
				{
					NotificationKey = CurrentNotification.NotificationKey,
					DeviceId = CurrentNotification.DeviceInstanceID
				});
			}

			//this.CurrentNotification.BitmapIcon = null;
			//UpdateLayout();
			//GC.Collect();
			//GC.WaitForPendingFinalizers();
			//GC.Collect();
		}

		private void btnMute_Click(object sender, RoutedEventArgs e)
		{
			this.MouseEnter -= this.GeneralNotificationWindow_MouseEnter;
			this.MouseLeave -= this.GeneralNotificationWindow_MouseLeave;
			this.stkpQuickActionsList.MouseEnter -= this.stkpQuickActionsList_MouseEnter;
			this.stkpQuickActionsList.MouseLeave -= this.stkpQuickActionsList_MouseLeave;

			this._storyboardLoad.Remove();
			this._storyboardFade.Remove();

			string GuestAccountContent = string.Format(System.Windows.Application.Current.TryFindResource("txtNotificationMuteDetail") as String, this.CurrentNotification.AppName);
			string ButtonContent = (string)System.Windows.Application.Current.TryFindResource("txtOK");
			string ButtonContent2 = (string)System.Windows.Application.Current.TryFindResource("txtCancel");
			CustomMessageBox guestMessageBox = new CustomMessageBox(new AirFusion.Windows.CustomMessageBox.CustomMessageBoxStructInfo(
															GuestAccountContent, CustomMessageBox.CustomImageIcon.Warring, ButtonContent, ButtonContent2));

			guestMessageBox.btn_CustomMessagebtn1.Click += this.approveMute;
			guestMessageBox.btn_CustomMessagebtn2.Click += this.rejectMute;
			guestMessageBox.btn_Close.Click += this.rejectMute;

			guestMessageBox.ShowDialog();
		}

		private void rejectMute(object sender, RoutedEventArgs e)
		{
			this.MouseEnter += this.GeneralNotificationWindow_MouseEnter;
			this.MouseLeave += this.GeneralNotificationWindow_MouseLeave;
			this.stkpQuickActionsList.MouseEnter += this.stkpQuickActionsList_MouseEnter;
			this.stkpQuickActionsList.MouseLeave += this.stkpQuickActionsList_MouseLeave;

			this._storyboardFade.Begin();
		}

		private void approveMute(object sender, RoutedEventArgs e)
		{
			if (this.CustomToastIgnore != null)
			{
				this.CustomToastIgnore(this, new CustomToastEventRawArgs()
				{
					NotificationKey = CurrentNotification.NotificationKey,
					DeviceId = CurrentNotification.DeviceInstanceID
				});
			}

			this.Close();
		}

		private void btnLocalLaunch_Click(object sender, RoutedEventArgs e)
		{
			if (this.CustomToastLocalLaunch != null)
			{
				this.CustomToastLocalLaunch(this, new CustomToastLocalLaunchEventArgs()
				{
					NotificationKey = CurrentNotification.NotificationKey,
					LocalLaunchHandler = CurrentNotification.LocalLaunchHandler,
					DeviceId = CurrentNotification.DeviceInstanceID,
					LocalLaunchIndex = 0, //no define here
				});
			}

			this.Close();
		}

		private void btnOpenFolder_Click(object sender, RoutedEventArgs e)
		{
			if (this.CustomToastLocalLaunch != null)
			{
				this.CustomToastLocalLaunch(this, new CustomToastLocalLaunchEventArgs()
				{
					NotificationKey = CurrentNotification.NotificationKey,
					LocalLaunchHandler = CurrentNotification.LocalLaunchHandler,
					DeviceId = CurrentNotification.DeviceInstanceID,
					LocalLaunchIndex = 0, //0 -> represents to open folder
				});
			}

			this.Close();
		}

		private void btnOpenFile_Click(object sender, RoutedEventArgs e)
		{
			if (this.CustomToastLocalLaunch != null)
			{
				this.CustomToastLocalLaunch(this, new CustomToastLocalLaunchEventArgs()
				{
					NotificationKey = CurrentNotification.NotificationKey,
					LocalLaunchHandler = CurrentNotification.LocalLaunchHandler,
					DeviceId = CurrentNotification.DeviceInstanceID,
					LocalLaunchIndex = 1, //1 -> represents to open file
				});
			}

			this.Close();
		}

		private void NotificaitonWindow_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
		{
			if (this.CurrentNotification.Type == AirFusion.Manager.Notification.NotificationType.BasicWithLaunch ||
				this.CurrentNotification.Type == AirFusion.Manager.Notification.NotificationType.UrlShareWithLaunch ||
				this.CurrentNotification.Type == AirFusion.Manager.Notification.NotificationType.FileShareWithLaunch)
			{
				if (this.CustomToastActivated != null)
				{
					this.CustomToastActivated(this, new CustomToastEventRawArgs()
					{
						NotificationKey = CurrentNotification.NotificationKey,
						DeviceId = CurrentNotification.DeviceInstanceID
					});
				}

				if (this.CustomToastLaunched != null)
				{
					this.CustomToastLaunched(this, new CustomToastEventRawArgs()
					{
						NotificationKey = CurrentNotification.NotificationKey,
						DeviceId = CurrentNotification.DeviceInstanceID,
					});
				}

				this.Close();
			}
		}

		/// <summary>
		/// Doubles the animation completed.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
		private void DoubleAnimationCompleted(object sender, EventArgs e)
		{
			if (!this.IsMouseOver && !this.stkpQuickActionsList.IsMouseOver)
			{
				this.Close();

				// Time-out close won't remove the notification on phone
				//
				//if (this.CustomToastDismissed != null)
				//{
				//	this.CustomToastDismissed(this, new CustomToastEventRawArgs() { NotificationKey = CurrentNotification.NotificationKey });
				//}
			}
		}

		/// <summary>
		/// Mains the window closing.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.ComponentModel.CancelEventArgs"/> instance containing the event data.</param>
		private void MainWindowClosing(object sender, System.ComponentModel.CancelEventArgs e)
		{
			foreach (Window window in System.Windows.Application.Current.Windows)
			{
				if (window.GetType().Equals(typeof(GeneralToastPopUp)))
				{
					window.Close();
				}
			}
		}

		private void DynamicButtonField_IsVisibleChanged(object sender, DependencyPropertyChangedEventArgs e)
		{
			if (this.DynamicButtonField.Visibility == Visibility.Visible)
			{
				this.NotificationBackground.Style = this.FindResource("rectActionBackground") as Style;
			}
		}

		private void GeneralNotificationWindow_Loaded(object sender, RoutedEventArgs e)
		{
			WindowInteropHelper wndHelper = new WindowInteropHelper(this);

			int exStyle = (int)GetWindowLong(wndHelper.Handle, (int)GetWindowLongFields.GWL_EXSTYLE);

			exStyle |= (int)ExtendedWindowStyles.WS_EX_TOOLWINDOW;
			SetWindowLong(wndHelper.Handle, (int)GetWindowLongFields.GWL_EXSTYLE, (IntPtr)exStyle);

			this.AdjustWindows();
		}

		private void GeneralNotificationWindow_MouseEnter(object sender, System.Windows.Input.MouseEventArgs e)
		{
			this._storyboardLoad.Remove();
			this._storyboardFade.Remove();
		}

		private void GeneralNotificationWindow_MouseLeave(object sender, System.Windows.Input.MouseEventArgs e)
		{
			this._storyboardFade.Begin();

			if (!this.stkpQuickActionsList.IsMouseOver)
				this.popUpButtonsList.IsOpen = false;
		}

		private void stkpQuickActionsList_MouseEnter(object sender, System.Windows.Input.MouseEventArgs e)
		{
			this._storyboardLoad.Remove();
			this._storyboardFade.Remove();
		}

		private void stkpQuickActionsList_MouseLeave(object sender, System.Windows.Input.MouseEventArgs e)
		{
			this._storyboardFade.Begin();
			this.popUpButtonsList.IsOpen = false;
		}

		protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
		{
			this.CurrentNotification.BitmapIcon = null;
			this.imageLeft.UpdateLayout();

			base.OnClosing(e);
		}

		#endregion UI controls event

		#region Window styles

		[Flags]
		public enum ExtendedWindowStyles
		{
			// ...
			WS_EX_TOOLWINDOW = 0x00000080,

			// ...
		}

		public enum GetWindowLongFields
		{
			// ...
			GWL_EXSTYLE = (-20),

			// ...
		}

		[DllImport("user32.dll")]
		public static extern IntPtr GetWindowLong(IntPtr hWnd, int nIndex);

		public static IntPtr SetWindowLong(IntPtr hWnd, int nIndex, IntPtr dwNewLong)
		{
			int error = 0;
			IntPtr result = IntPtr.Zero;
			// Win32 SetWindowLong doesn't clear error on success
			SetLastError(0);

			if (IntPtr.Size == 4)
			{
				// use SetWindowLong
				Int32 tempResult = IntSetWindowLong(hWnd, nIndex, IntPtrToInt32(dwNewLong));
				error = Marshal.GetLastWin32Error();
				result = new IntPtr(tempResult);
			}
			else
			{
				// use SetWindowLongPtr
				result = IntSetWindowLongPtr(hWnd, nIndex, dwNewLong);
				error = Marshal.GetLastWin32Error();
			}

			if ((result == IntPtr.Zero) && (error != 0))
			{
				throw new System.ComponentModel.Win32Exception(error);
			}

			return result;
		}

		[DllImport("user32.dll", EntryPoint = "SetWindowLongPtr", SetLastError = true)]
		private static extern IntPtr IntSetWindowLongPtr(IntPtr hWnd, int nIndex, IntPtr dwNewLong);

		[DllImport("user32.dll", EntryPoint = "SetWindowLong", SetLastError = true)]
		private static extern Int32 IntSetWindowLong(IntPtr hWnd, int nIndex, Int32 dwNewLong);

		private static int IntPtrToInt32(IntPtr intPtr)
		{
			return unchecked((int)intPtr.ToInt64());
		}

		[DllImport("kernel32.dll", EntryPoint = "SetLastError")]
		public static extern void SetLastError(int dwErrorCode);

		#endregion Window styles
	}
}