﻿using AirFusion.Common;
using System;
using System.Runtime.InteropServices;
using System.Runtime.Serialization;

namespace AirFusion.Common.ProximityDetection.SoundTrigger
{
	internal class AFSoundException : Exception, ISerializable
	{
		public AFSoundException()
			: base("Unknown SoundError") { }

		public AFSoundException(string message)
			: base(message) { }

		public AFSoundException(string message, Exception inner)
			: base(message, inner) { }

		protected AFSoundException(SerializationInfo info, StreamingContext context)
			: base(info, context) { }
	}

	internal static class AFSoundTrigger
	{
		#region Imported C++ function

		/// <summary>
		/// To start listen sound message, import from C++ library
		/// </summary>
		[DllImport("SoundTriggerforDesktop.dll", EntryPoint = "StartListen")]
		private static extern bool StartSoundListen();

		/// <summary>
		/// To stop listen sound message, import from C++ library
		/// </summary>
		[DllImport("SoundTriggerforDesktop.dll", EntryPoint = "StopListen")]
		private static extern bool StopSoundListen();

		/// <summary>
		/// To get the last error code, import from C++ library
		/// </summary>
		[DllImport("SoundTriggerforDesktop.dll", EntryPoint = "GetSTLLastError")]
		private static extern int GetSoundLastError();

		[DllImport("SoundTriggerforDesktop.dll", EntryPoint = "UnInit")]
		private static extern void UnregisterSound();

		/// <summary>
		/// To get the last error code, import from C++ library
		/// </summary>
		/// <param name="nType">Define the action, to listen or to send sound message</param>
		/// <param name="appKey">A key which should match between sound transmitter and sound receiver</param>
		/// <param name="callback">A callback to receive sound status and message</param>
		[DllImport("SoundTriggerforDesktop.dll", EntryPoint = "Init", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Auto)]
		private static extern bool RegisterSound(
			int nType,
			[MarshalAs(UnmanagedType.LPStr)] string appKey,
			[MarshalAs(UnmanagedType.FunctionPtr)] AFSoundCallback callback);

		[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
		private delegate void AFSoundCallback(int message, int param1, int param2, IntPtr data, int dataSize);

		#endregion Imported C++ function

		#region Sound flag enumeration

		internal enum AFSoundAction : int
		{
			// action
			SOUND_INIT_PLAY = 0x01,

			SOUND_INIT_LISTEN = 0x02,
		}

		internal enum AFSoundStatus : int
		{
			// Init status
			SOUND_INIT_MSG_READY_TO_PLAY = 0x00000001,

			SOUND_INIT_MSG_READY_TO_LISTEN = 0x00000002,

			// Send status
			SOUND_SEND_MSG_PREPARE = 0x00000011,

			SOUND_SEND_MSG_PLAYING = 0x00000012,
			SOUND_SEND_MSG_PLAYED = 0x00000013,
			SOUND_SEND_MSG_PLAY_FINISH = 0x00000014,
			SOUND_SEND_MSG_VOLUME_TOO_LOW = 0x00000015,	// Volume is too low

			// Listening status
			SOUND_LISTEN_MSG_LISTENING = 0x00000020,

			SOUND_LISTEN_MSG_RECEIVING_DATA = 0x00000021,
			SOUND_LISTEN_MSG_RECEIVED_DATA = 0x00000022,

			// Device Property Change status
			SOUND_DEVICE_PROPERTIES_CHANGE = 0x00000040,
		}

		internal enum AFSoundError : int
		{
			// Error
			SOUND_ERR_NO_INITIALIZED = 0x00000001,	// no initialized

			SOUND_ERR_INIT_AGAIN = 0x00000002,	// init again
			SOUND_ERR_INITIAL = 0x00000004,	// no initialized
			SOUND_ERR_PLAY_NOT_READY = 0x00000008,	// Play is not ready
			SOUND_ERR_PLAY_STATUS_NOT_PLAYING = 0x00000010,	// Play status is not Playing
			SOUND_ERR_PLAY_DEVICE_NOT_FOUND = 0x00000020,	// Play device is not found
			SOUND_ERR_LISTEN_DEVICE_NOT_FOUND = 0x00000040,	// Listen device is not found
			SOUND_ERR_RECORDER_SAMPLE_RATE = 0x00000080,	// Recorder sample rate is not 48kHz
			SOUND_ERR_LISTEN_NOT_READY = 0x00000100,	// Listen is not ready
			SOUND_ERR_LISTEN_NOT_START = 0x00000200,	// Listen is not start
			SOUND_ERR_LISTEN_IS_START = 0x00000400,	// Listen is start
			SOUND_ERR_BAD_DATA_TYPE = 0x00000800,	// Data type is out of range (0 ~7)
			SOUND_ERR_BAD_DATA = 0x00001000,	// Data is invalid
			SOUND_ERR_BAD_REPEAT_TIMES = 0x00002000,	// Repeat times is out of range (>=0)Play device is not found
		}

		#endregion Sound flag enumeration

		#region Delegates

		/// <summary>
		/// Represents the methods that will handle the event of incoming sound message.
		/// </summary>
		/// <param name="soundMessage">The received sound message, might be null or empty.</param>
		public delegate void AFSoundReceiveEventHandler(byte[] soundMessage);

		#endregion Delegates

		#region Events

		/// <summary>
		/// Occurs when a sound message is received.
		/// </summary>
		public static event AFSoundReceiveEventHandler SoundReceived;

		private static AFSoundCallback _callback = SoundCallbackFunc;

		#endregion Events

		#region Methods

		/// <summary>
		/// Registers sound callback function and start listen.
		/// </summary>
		/// <exception cref="AFSoundException"></exception>
		public static bool RegisterAndStartListen(string appKey)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("appKey: " + appKey);

				bool bResult = RegisterSound((int)AFSoundAction.SOUND_INIT_LISTEN, appKey, _callback);

				if (!bResult)
				{
					int error = GetSoundLastError();
					SoundErrorExceptionHandler(error);
				}

				return bResult;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stops listening and unregister sound callback.
		/// </summary>
		public static void UnregisterAndStopListen()
		{
			try
			{
				StaticLogger.TrackEnter();

				bool success = StopSoundListen();
				if (!success)
				{
					int error = GetSoundLastError();
					SoundErrorExceptionHandler(error);
				}
				UnregisterSound();
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a sound-related message is received.
		/// </summary>
		/// <param name="message">sound message type.</param>
		/// <param name="param1">TBD</param>
		/// <param name="param2">TBD</param>
		/// <param name="data">The received sound message.</param>
		/// <param name="dataSize">The length of received sound message.</param>
		private static void SoundCallbackFunc(int message, int param1, int param2, IntPtr data, int dataSize)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log paramters
				StaticLogger.LogDebug("message: " + message);
				StaticLogger.LogDebug("param1: " + param1);
				StaticLogger.LogDebug("param2: " + param2);
				StaticLogger.LogDebug("data: " + data);
				StaticLogger.LogDebug("dataSize: " + dataSize);

				// get status
				AFSoundStatus status = (AFSoundStatus)message;
				StaticLogger.LogInfo("status: " + status);

				switch (status)
				{
					case AFSoundStatus.SOUND_LISTEN_MSG_RECEIVED_DATA:
						{
							if (data != IntPtr.Zero && dataSize != 0)
							{
								byte[] byteArray = new byte[dataSize];
								Marshal.Copy(data, byteArray, 0, dataSize);

								StaticLogger.LogInfo("data: " + byteArray.ToHexString());

								if (SoundReceived != null)
								{
									SoundReceived(byteArray);
								}
							}
						}
						break;

					case AFSoundStatus.SOUND_INIT_MSG_READY_TO_LISTEN:
						{
							bool success = StartSoundListen();
							if (!success)
							{
								int error = GetSoundLastError();
								SoundErrorExceptionHandler(error);
							}
						}
						break;

					case AFSoundStatus.SOUND_DEVICE_PROPERTIES_CHANGE:
						break;
				}
			}
			catch (Exception ex)
			{
				StaticLogger.LogWarn("Sound exception, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Throws sound exception message.
		/// </summary>
		/// <param name="errorCode">Last error code retrived from c++ dll</param>
		private static void SoundErrorExceptionHandler(int errorCode)
		{
			AFSoundError soundError = (AFSoundError)errorCode;
			switch (soundError)
			{
				case AFSoundError.SOUND_ERR_NO_INITIALIZED:
					throw new AFSoundException("Device is not initialized");
				case AFSoundError.SOUND_ERR_INIT_AGAIN:
					throw new AFSoundException("Device already be initialized");
				case AFSoundError.SOUND_ERR_INITIAL:
					throw new AFSoundException("Initialization fail");
				case AFSoundError.SOUND_ERR_PLAY_NOT_READY:
					throw new AFSoundException("Play is not ready");
				case AFSoundError.SOUND_ERR_PLAY_STATUS_NOT_PLAYING:
					throw new AFSoundException("Play status is not Playing");
				case AFSoundError.SOUND_ERR_PLAY_DEVICE_NOT_FOUND:
					throw new AFSoundException("Could not find the device to play");
				case AFSoundError.SOUND_ERR_LISTEN_DEVICE_NOT_FOUND:
					throw new AFSoundException("Could not find the device to listen");
				case AFSoundError.SOUND_ERR_RECORDER_SAMPLE_RATE:
					throw new AFSoundException("Recorder sample rate is incorrect");
				case AFSoundError.SOUND_ERR_LISTEN_NOT_READY:
					throw new AFSoundException("Listen device is not ready to listen");
				case AFSoundError.SOUND_ERR_LISTEN_NOT_START:
					throw new AFSoundException("Listen device is not start");
				case AFSoundError.SOUND_ERR_LISTEN_IS_START:
					throw new AFSoundException("Listen device already started");
				case AFSoundError.SOUND_ERR_BAD_DATA_TYPE:
					throw new AFSoundException("Data type is out of range (0 ~7)");
				case AFSoundError.SOUND_ERR_BAD_DATA:
					throw new AFSoundException("Invalid data");
				case AFSoundError.SOUND_ERR_BAD_REPEAT_TIMES:
					throw new AFSoundException("Repeat times is out of range. Should be > 0");
			}
		}

		#endregion Methods
	}
}