﻿using AirFusion.Common;
using AirFusion.Dandelion;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Net.NetworkInformation;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class UserDeviceManager
	{
		/// <summary>
		/// Provides the functionality of Dandelion.
		/// </summary>
		internal sealed partial class DandelionAdapter
		{
			#region Properties

			/// <summary>
			/// Gets a value indicating whether the advertisement is processing.
			/// </summary>
			public bool IsAdvertising { get; private set; }

			/// <summary>
			/// Gets a value indicating whether the <see cref="DandelionAdapter"/> instance is initialized.
			/// </summary>
			public bool IsInitialized { get; private set; }

			#endregion Properties

			#region Public Methods

			/// <summary>
			/// Initializes a new instance of the <see cref="DandelionAdapter"/> class.
			/// </summary>
			/// <param name="host">The host.</param>
			/// <param name="port">The port.</param>
			/// <param name="isSecure">Indicates whether use HTTPS.</param>
			public DandelionAdapter(string host, ushort? port = null, bool isSecure = true)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("host: " + host);
					StaticLogger.LogDebug("port: " + (port.HasValue ? port.Value.ToString() : "N/A"));
					StaticLogger.LogDebug("isSecure: " + isSecure);

					this._aodController = new AODController(host, port, isSecure);
					this.IsAdvertising = false;
					this.IsInitialized = false;
					this._localInstanceId = null;
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Initializes a new instance of the <see cref="DandelionAdapter"/> class.
			/// </summary>
			public DandelionAdapter()
			{
				this._aodController = new AODController();
				this.IsAdvertising = false;
				this.IsInitialized = false;
				this._localInstanceId = null;
			}

			/// <summary>
			/// Initializes this instance.
			/// </summary>
			/// <param name="instanceId">The instance identifier.</param>
			/// <exception cref="System.ArgumentException" />
			/// <exception cref="System.Exception" />
			public void Initialize(Guid instanceId)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("instanceId: " + instanceId.GetStringFormat());

					// check parameter
					if (instanceId.Equals(Guid.Empty))
					{
						throw new ArgumentException("The 'instanceId' should not be empty.");
					}

					// initialize the AODController
					try
					{
						StaticLogger.LogTrace("Call AODController.Initialize()...");
						Debug.Assert(this._aodController != null);
						this._aodController.Initialize(instanceId.GetStringFormat());
						StaticLogger.LogTrace("AODController.Initialize() is completed without exception.");
						this._localInstanceId = instanceId;
						this.IsInitialized = true;
					}
					catch (Exception ex)
					{
						throw new Exception("Failed initializing AODController.", ex);
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Sets the profile.
			/// </summary>
			/// <param name="profile">The profile.</param>
			/// <exception cref="System.ArgumentNullException" />
			/// <exception cref="System.Exception" />
			/// <exception cref="System.InvalidOperationException" />
			public void SetProfile(InstanceProfile profile)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameter
					StaticLogger.LogDebug("profile: " + (profile != null ? "\r\n" + profile.ToString() : "N/A"));

					// check parameter
					if (profile == null)
					{
						throw new ArgumentNullException("profile");
					}

					// check prerequisite
					if (!this.IsInitialized)
					{
						throw new InvalidOperationException("This instance is not initialized yet.");
					}

					// get profile string
					string localProfile = null;
					try
					{
						localProfile = profile.ToBase64String();
						StaticLogger.LogInfo("Encoded profile string: " + localProfile);
					}
					catch (Exception ex)
					{
						throw new Exception("Failed getting profile string.", ex);
					}

					// set profile via AODController
					try
					{
						StaticLogger.LogTrace("Call AODController.SetDeviceProfile()...");
						Debug.Assert(this._aodController != null);
						this._aodController.SetDeviceProfile(localProfile);
						StaticLogger.LogTrace("AODController.SetDeviceProfile() is completed without exception.");
					}
					catch (Exception ex)
					{
						throw new Exception("Failed setting profile.", ex);
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Sends pairing invitation to the specified device asynchronously.
			/// </summary>
			/// <param name="targetInstanceId">The instance ID of target device.</param>
			/// <returns>True if the operation is done successfully; false otherwise.</returns>
			/// <exception cref="System.ArgumentException" />
			/// <exception cref="System.Exception" />
			/// <exception cref="System.InvalidOperationException" />
			public async Task SendPairingInvitationAsync(Guid targetInstanceId)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameter
					StaticLogger.LogDebug("targetInstanceId: " + targetInstanceId.GetStringFormat());

					// check parameter
					if (targetInstanceId.Equals(Guid.Empty))
					{
						throw new ArgumentException("The 'targetInstanceId' should not be empty.");
					}

					// check prerequisite
					if (!this.IsInitialized)
					{
						throw new InvalidOperationException("This instance is not initialized yet.");
					}

					try
					{
						AODCommon.PairingInvitationTaskResult result;
						try
						{
							Debug.Assert(this._aodController != null);
							StaticLogger.LogTrace("Call AODController.SendPairingInvitationAsync()...");
							result = await this._aodController.SendPairingInvitationAsync(targetInstanceId.GetStringFormat()).ConfigureAwait(continueOnCapturedContext: false);
							StaticLogger.LogTrace("AODController.SendPairingInvitationAsync() is completed without exception.");
							StaticLogger.LogInfo("result.IsSuccess: " + result.IsSuccess);
						}
						catch (Exception ex)
						{
							throw new Exception("AODController.SendPairingInvitationAsync() failed.", ex);
						}

						if (result.Exception != null)
						{
							throw result.Exception;
						}

						if (result.IsSuccess == false)
						{
							throw new Exception("isSuccess=false");
						}
					}
					catch (Exception ex)
					{
						throw new Exception("Failed sending pairing invitation through Dandelion.", ex);
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Gets the profile of the specified device asynchronously.
			/// </summary>
			/// <param name="instanceId">The instance ID.</param>
			/// <returns>The profile of the specified instance.</returns>
			/// <exception cref="System.ArgumentNullException" />
			/// <exception cref="System.InvalidOperationException" />
			/// <exception cref="System.Exception" />
			/// <exception cref="System.FormatException" />
			public async Task<InstanceProfile> GetProfileAsync(Guid instanceId)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("instanceId: " + instanceId.GetStringFormat());

					// check prerequisite
					if (!this.IsInitialized)
					{
						throw new InvalidOperationException("This instance is not initialized yet.");
					}

					string profile = null;
					try
					{
						Debug.Assert(this._aodController != null);
						StaticLogger.LogTrace("Call AODController.GetDeviceProfileAsync()...");
						var result = await this._aodController.GetDeviceProfileAsync(instanceId.GetStringFormat()).ConfigureAwait(continueOnCapturedContext: false);
						StaticLogger.LogTrace("AODController.GetDeviceProfileAsync() is completed without exception.");
						StaticLogger.LogInfo("result.IsSuccess: " + result.IsSuccess);

						if (result.Exception != null)
						{
							throw result.Exception;
						}
						if (result.IsSuccess == false)
						{
							throw new Exception("AODController.GetDeviceProfileAsync() indicates the operation is returned unsuccessfully.");
						}

						profile = result.Profile;
						StaticLogger.LogInfo("Response profile: " + profile);
					}
					catch (Exception ex)
					{
						throw new Exception("AODController.GetDeviceProfileAsync() failed.", ex);
					}

					if (string.IsNullOrWhiteSpace(profile))
					{
						throw new Exception("The profile of device '" + instanceId.GetStringFormat() + "' is empty.");
					}

					// return the result
					try
					{
						return InstanceProfile.FromBase64String(profile);
					}
					catch (Exception ex)
					{
						throw new FormatException("Failed parsing the profile string '" + profile + "'.", ex);
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Gets the endpoint of the specified device asynchronously.
			/// </summary>
			/// <param name="instanceId">The instance ID.</param>
			/// <returns></returns>
			/// <exception cref="System.ArgumentNullException" />
			/// <exception cref="System.InvalidOperationException" />
			/// <exception cref="System.Exception" />
			/// <exception cref="System.FormatException" />
			public async Task<IPEndPoint[]> GetEndpointAsync(Guid instanceId)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("instanceId: " + instanceId.GetStringFormat());

					// check prerequisite
					if (!this.IsInitialized)
					{
						throw new InvalidOperationException("This instance is not initialized yet.");
					}

					InstanceProfile profile = null;
					try
					{
						profile = await this.GetProfileAsync(instanceId).ConfigureAwait(continueOnCapturedContext: false);
						if (profile == null)
						{
							throw new Exception("The profile is null.");
						}
					}
					catch (Exception ex)
					{
						throw new Exception("Failed getting profile from Dandelion.", ex);
					}

					if (profile.Network == null || profile.Network.Length == 0 || profile.WebServicePort.HasValue == false || profile.WebServicePort.Value == 0)
					{
						throw new Exception("The profile doesn't contain network information or port number.");
					}

					if (profile.Network == null || profile.Network.Count() == 0 || profile.WebServicePort.Value == 0)
					{
						throw new Exception("The address or port number is invalid from the profile.");
					}

					return profile.Network.Select(item => new IPEndPoint(item.Address, (int)profile.WebServicePort.Value)).ToArray();
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Begins advertisement.
			/// </summary>
			/// <exception cref="System.InvalidOperationException" />
			public void BeginAdvertise()
			{
				try
				{
					StaticLogger.TrackEnter();

					// check prerequisite
					if (!this.IsInitialized)
					{
						throw new InvalidOperationException("This instance is not initialized yet.");
					}
					if (this.IsAdvertising)
					{
						throw new InvalidOperationException("It's already being advertising.");
					}

					this.IsAdvertising = true;
					this.Advertise();
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Ends the advertisement.
			/// </summary>
			/// <exception cref="System.InvalidOperationException" />
			public void EndAdvertise()
			{
				try
				{
					StaticLogger.TrackEnter();

					// check prerequisite
					if (!this.IsInitialized)
					{
						throw new InvalidOperationException("This instance is not initialized yet.");
					}

					this.IsAdvertising = false;
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Gets the nearby profile list.
			/// </summary>
			/// <returns></returns>
			/// <exception cref="System.InvalidOperationException" />
			/// <exception cref="System.Exception" />
			public IEnumerable<KeyValuePair<Guid, InstanceProfile>> GetNearbyDevices()
			{
				try
				{
					StaticLogger.TrackEnter();

					// check prerequisite
					if (!this.IsInitialized)
					{
						throw new InvalidOperationException("This instance is not initialized yet.");
					}

					// initialize
					AutoResetEvent eventHandle = new AutoResetEvent(false);
					bool completed = false;
					Exception exception = null;
					ConcurrentBag<KeyValuePair<Guid, InstanceProfile>> profileBag = new ConcurrentBag<KeyValuePair<Guid, InstanceProfile>>();
					Stopwatch watch = new Stopwatch();
					watch.Start();

					// start pairing without getting the list
					try
					{
						Debug.Assert(this._aodController != null);
						StaticLogger.LogTrace("Call AODController.StartPairing()...");
						this._aodController.StartPairing(PairingHoldTime, (s, e) =>
						{
							// first: Get device ID list
							// second ~ nth: Get instance ID & profile

							try
							{
								if (!e.IsSuccess)
								{
									if (e.Exception != null)
									{
										exception = e.Exception;
									}
									else
									{
										exception = new Exception("AODController.StartPairing() failed.");
									}
								}
								else if (!string.IsNullOrEmpty(e.InstanceID) && !string.IsNullOrEmpty(e.Profile))
								{
									StaticLogger.LogInfo("Get a nearby device:\r\ninstance ID: " + e.InstanceID + "\r\nprofile: " + e.Profile);
									try
									{
										Guid instanceID = Guid.Parse(e.InstanceID);
										InstanceProfile profile = InstanceProfile.FromBase64String(e.Profile);
										profileBag.Add(new KeyValuePair<Guid, InstanceProfile>(instanceID, profile));
									}
									catch (Exception ex)
									{
										StaticLogger.LogWarn("Invalid instance ID or profile string, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
									}
								}

								completed = e.IsCompleted;

								if (e.IsCompleted)
								{
									eventHandle.Set(); // set the event handle when the IsCompleted is true
								}
							}
							catch (Exception ex)
							{
								completed = true;
								exception = ex;
							}
						});
						StaticLogger.LogTrace("AODController.StartPairing() is returned without exception.");
					}
					catch (Exception ex)
					{
						throw new Exception("AODController.StartPairing() failed.", ex);
					}

					while (!completed)
					{
						eventHandle.WaitOne((PairingHoldTime + 10) * 1000); // 10 is an arbitrary number

						KeyValuePair<Guid, InstanceProfile> result;
						if (profileBag.TryTake(out result))
						{
							yield return result;
						}
					}

					watch.Stop();

					StaticLogger.LogInfo("AODController.StartPairing() is completed in " + watch.ElapsedMilliseconds + "ms.");

					if (exception != null)
					{
						throw exception;
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Gets the nearby profile list.
			/// </summary>
			/// <param name="callback">The callback function.</param>
			/// <exception cref="System.ArgumentNullException" />
			/// <exception cref="System.InvalidOperationException" />
			/// <exception cref="System.Exception" />
			public void BeginGetNearbyDevices(NearbyScanCallback callback)
			{
				try
				{
					StaticLogger.TrackEnter();

					// check parameters
					if (callback == null)
					{
						throw new ArgumentNullException("callback");
					}

					// check prerequisite
					if (!this.IsInitialized)
					{
						throw new InvalidOperationException("This instance is not initialized yet.");
					}

					// start pairing without getting the list
					try
					{
						Debug.Assert(this._aodController != null);
						StaticLogger.LogTrace("Call AODController.StartPairing()...");
						this._aodController.StartPairing(PairingHoldTime, (s, e) =>
						{
							// first: Get device ID list
							// second ~ nth: Get instance ID & profile

							try
							{
								if (!e.IsSuccess)
								{
									callback(new NearbyScanCallbackArg(e.Exception ?? new Exception("AODController.StartPairing() failed."), e.IsCompleted));
								}
								else if (!string.IsNullOrEmpty(e.InstanceID) && !string.IsNullOrEmpty(e.Profile))
								{
									StaticLogger.LogInfo("Get a nearby device:\r\ninstance ID: " + e.InstanceID + "\r\nprofile: " + e.Profile);
									try
									{
										Guid instanceID = Guid.Parse(e.InstanceID);
										InstanceProfile profile = InstanceProfile.FromBase64String(e.Profile);
										callback(new NearbyScanCallbackArg(instanceID, profile, e.IsCompleted));
									}
									catch (Exception ex)
									{
										StaticLogger.LogWarn("Invalid instance ID or profile string, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
									}
								}
							}
							catch (Exception ex)
							{
								StaticLogger.LogError("Unhandled exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
							}
						});
						StaticLogger.LogTrace("AODController.StartPairing() is returned without exception.");
					}
					catch (Exception ex)
					{
						throw new Exception("AODController.StartPairing() failed.", ex);
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			#endregion Public Methods
		}
	}
}