﻿using AirFusion.Common;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class UserDeviceManager
	{
		internal sealed partial class MulticastDnsAdapter
		{
			/// <summary>
			/// Called when RegisterService() is callback.
			/// </summary>
			/// <param name="args">The <see cref="DnsSdHandler.RegistrarCallbackArgs"/> instance containing the callback arguments.</param>
			private void OnRegisterCallback(DnsSdHandler.RegistrarCallbackArgs args)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("args: " + (args != null ? args.ToString() : "N/A"));

					// check parameter
					if (args == null)
					{
						StaticLogger.LogError("args=NULL");
						return;
					}

					// debug assertion
					Debug.Assert(args.Handler != null);

					if (args.IsFaulted)
					{
						// clean the instance name
						this.RegisteredInstance = null;

						StaticLogger.LogError("Failed registering the instance '" + args.InstanceName + "." + args.Handler.ServiceType + "." + args.Handler.Domain + ".'.");
					}
					else
					{
						// update the instance name in case there's any collision during the registration
						if (this.RegisteredInstance != null)
						{
							this.RegisteredInstance.InstanceName = args.InstanceName;
						}

						StaticLogger.LogInfo("This device has registered the instance name '" + args.InstanceName + "'.");
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// The callback of StartBrowse().
			/// </summary>
			/// <param name="args">The <see cref="DnsSdHandler.BrowserCallbackArgs"/> instance containing the callback arguments.</param>
			private void OnBrowserCallback(DnsSdHandler.BrowserCallbackArgs args)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("args: " + (args != null ? args.ToString() : "N/A"));

					// check parameter
					if (args == null)
					{
						StaticLogger.LogError("args=NULL");
						return;
					}

					// debug assertion
					Debug.Assert(args.Handler != null);

					if (args.IsFaulted)
					{
						StaticLogger.LogError("Failed browsing '" + args.Handler.ServiceType + "'.");
					}
					else if (args.EventType.HasValue)
					{
						// check parameter
						if (string.IsNullOrEmpty(args.InstanceName))
						{
							StaticLogger.LogWarn("instance name is empty or null, skip this.");
							return;
						}

						// skip if this is not fired by our observer
						if (this._observationProcessorId.HasValue && args.ProcessorID != this._observationProcessorId.Value)
						{
							StaticLogger.LogInfo("Skip since this event is not fired by our observer.");
							return;
						}

						// skip local device
						if (this.RegisteredInstance != null && string.Compare(args.InstanceName, this.RegisteredInstance.InstanceName) == 0)
						{
							StaticLogger.LogInfo("Skip local instance.");
							return;
						}

						switch (args.EventType.Value)
						{
							case DnsSdHandler.BrowserCallbackArgs.BrowserEventType.Found:
								this.OnServiceFound(args);
								break;

							case DnsSdHandler.BrowserCallbackArgs.BrowserEventType.Lost:
								this.OnServiceLost(args);
								break;
						}
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogError(CommonUtils.GetExceptionMessages(ex));
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Called when a service instance is found.
			/// </summary>
			/// <param name="args">The <see cref="DnsSdHandler.BrowserCallbackArgs"/> instance containing the callback arguments.</param>
			private void OnServiceFound(DnsSdHandler.BrowserCallbackArgs args)
			{
				try
				{
					StaticLogger.TrackEnter();

					// debug assertion
					Debug.Assert(args != null);
					Debug.Assert(args.Handler != null);
					Debug.Assert(args.InstanceName != null);
					Debug.Assert(args.InterfaceIndex.HasValue);

					// create the instance
					var instance = new MDnsInstance()
					{
						Handler = args.Handler,
						InstanceName = args.InstanceName
					};
					instance.AddAddress(args.InterfaceIndex.Value, null);

					// add or update the instance
					try
					{
						Debug.Assert(this.InstanceStore != null);
						instance = this.InstanceStore.AddOrUpdate(
							args.InstanceName,
							instance,
							delegate(string instanceName, MDnsInstance existingInstance)
							{
								if (existingInstance != null)
								{
									existingInstance.Handler = args.Handler;
									existingInstance.InstanceName = args.InstanceName;
									existingInstance.AddAddress(args.InterfaceIndex.Value, null);

									return existingInstance;
								}

								return instance;
							}
						);
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed adding or updating the instance, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						return;
					}

					// debug assertion
					Debug.Assert(instance != null);

					// start resolver
					lock (CommonUtils.GetLocker(new Tuple<object, object, string>(this, instance, LNResolver)))
					{
						try
						{
							if (!instance.IsResolving)
							{
								instance.StartResolve((resolvedInstanceRecord) =>
								{
									// fire ServiceFound event
									if (this.ServiceFound != null)
									{
										StaticLogger.LogTrace("Fire ServiceFound event...");
										CommonUtils.InvokeEvent<ServiceFoundEventArgs>(new EventHandler<ServiceFoundEventArgs>(this.ServiceFound), this, new ServiceFoundEventArgs(resolvedInstanceRecord));
									}

									// stop resolver
									lock (CommonUtils.GetLocker(new Tuple<object, object, string>(this, instance, LNResolver)))
									{
										try
										{
											if (instance.IsResolving)
											{
												instance.StopResolve();
											}
										}
										catch (Exception ex)
										{
											StaticLogger.LogError("Failed to stop resolving the instance, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
											return;
										}
									}

									// start query address
									lock (CommonUtils.GetLocker(new Tuple<object, object, string>(this, instance, LNAddrQuerier)))
									{
										try
										{
											if (!instance.IsAddrQuerying)
											{
												instance.StartQueryAddress((answeredInstanceRecord, ifIndex, address, isRemoved) =>
												{
													// fire AddressUpdated event
													if (this.AddressUpdated != null)
													{
														StaticLogger.LogTrace("Fire AddressUpdated event...");
														CommonUtils.InvokeEvent<AddressUpdatedEventArgs>(new EventHandler<AddressUpdatedEventArgs>(this.AddressUpdated), this, new AddressUpdatedEventArgs(answeredInstanceRecord, ifIndex, address, isRemoved));
													}

													/*
													// release waiting threads that waiting for new instance with the desired instance ID
													if (!isRemoved && resolvedInstanceRecord.InstanceID.HasValue)
													{
														try
														{
															HashSet<AutoResetEvent> eventHandleBags;
															if (this._reconfirmedInstanceIDs.TryGetValue(resolvedInstanceRecord.InstanceID.Value, out eventHandleBags))
															{
																foreach (var eventHandle in eventHandleBags.ToArray())
																{
																	eventHandle.Set();
																}
															}

															this._reconfirmedInstanceIDs.TryRemove(resolvedInstanceRecord.InstanceID.Value, out eventHandleBags);
														}
														catch (Exception ex)
														{
															StaticLogger.LogWarn("Unhandled exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
														}
													}
													*/
												});
											}
										}
										catch (Exception ex)
										{
											StaticLogger.LogError("Failed starting to query the instance, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
											return;
										}
									}
								});
							}
						}
						catch (Exception ex)
						{
							StaticLogger.LogError("Failed starting to resolve the instance, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
							return;
						}
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogError(CommonUtils.GetExceptionMessages(ex));
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Called when a service instance is lost.
			/// </summary>
			/// <param name="args">The <see cref="DnsSdHandler.BrowserCallbackArgs"/> instance containing the callback arguments.</param>
			private void OnServiceLost(DnsSdHandler.BrowserCallbackArgs args)
			{
				try
				{
					StaticLogger.TrackEnter();

					// debug assertion
					Debug.Assert(args != null);
					Debug.Assert(args.Handler != null);
					Debug.Assert(args.InstanceName != null);
					Debug.Assert(args.InterfaceIndex.HasValue);

					// get the instance from the store
					Debug.Assert(this.InstanceStore != null);
					MDnsInstance instance = null;
					if (this.InstanceStore.TryGetValue(args.InstanceName, out instance) && instance != null)
					{
						// remove the address from the address pool
						var removedAddress = instance.RemoveAddress(args.InterfaceIndex.Value);

						// fire AddressUpdate event
						if (this.AddressUpdated != null)
						{
							StaticLogger.LogTrace("Fire AddressUpdated event...");
							CommonUtils.InvokeEvent<AddressUpdatedEventArgs>(new EventHandler<AddressUpdatedEventArgs>(this.AddressUpdated), this, new AddressUpdatedEventArgs(instance, args.InterfaceIndex.Value, null, true));
						}

						// remove from the store if it's not present in any interface
						if (instance.InterfaceIndexes.Count == 0)
						{
							// stop resolving
							lock (CommonUtils.GetLocker(new Tuple<object, object, string>(this, instance, LNResolver)))
							{
								try
								{
									if (instance.IsResolving)
									{
										instance.StopResolve();
									}
								}
								catch (Exception ex)
								{
									StaticLogger.LogError("Failed stopping resolving the instance, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
								}
							}

							// stop address query
							lock (CommonUtils.GetLocker(new Tuple<object, object, string>(this, instance, LNAddrQuerier)))
							{
								try
								{
									if (instance.IsAddrQuerying)
									{
										instance.StopQueryAddress();
									}
								}
								catch (Exception ex)
								{
									StaticLogger.LogError("Failed stopping querying address of the instance, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
								}
							}

							this.InstanceStore.TryRemove(args.InstanceName, out instance);

							// fire ServiceLost event
							if (this.ServiceLost != null)
							{
								StaticLogger.LogTrace("Fire ServiceLost event...");
								CommonUtils.InvokeEvent<ServiceLostEventArgs>(new EventHandler<ServiceLostEventArgs>(this.ServiceLost), this, new ServiceLostEventArgs(instance));
							}
						}
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogError(CommonUtils.GetExceptionMessages(ex));
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			public AutoResetEvent HashSet { get; set; }
		}
	}
}
