﻿using AirFusion.Common;
using Bonjour;
using System;
using System.Diagnostics;
using System.Linq;
using System.Threading;

namespace AirFusion.Manager.DeviceManager
{
	public partial class DnsSdHandler
	{
		/// <summary>
		/// Invokes the action asynchronously if it's not on UI thread or synchronousely if it's on UI thread.
		/// </summary>
		/// <param name="action">The action.</param>
		/// <param name="state">The state.</param>
		private void RunOnUIThread(Action<object> action, object state)
		{
			Debug.Assert(this._synchronizationContext != null);
			if (this._synchronizationContext == SynchronizationContext.Current)
			{
				action.Invoke(state);
			}
			else
			{
				this._synchronizationContext.Post(new SendOrPostCallback(action), state);
			}
		}

		/// <summary>
		/// Stops all processors.
		/// </summary>
		private void StopAllProcessors()
		{
			try
			{
				StaticLogger.TrackEnter();

				// Bonjour uses STA COM threading model, below code segment must be invoked on synchronization context,
				// otherwise there could be unexpected behavior
				this.RunOnUIThread(delegate(object state)
				{
					Debug.Assert(state is DnsSdHandler);
					DnsSdHandler handler = state as DnsSdHandler;

					StaticLogger.LogTrace("Total processors: " + this._processorDic.Count());

					var browserCount = handler._processorDic.Where(entry => entry.Value is BrowserData).Count();
					var resolverCount = handler._processorDic.Where(entry => entry.Value is ResolverData).Count();
					var recordQuerierCount = handler._processorDic.Where(entry => entry.Value is RecordQuerierData).Count();
					var addrQuerierCount = handler._processorDic.Where(entry => entry.Value is AddrQuerierData).Count();

					StaticLogger.LogInfo("# of browsers: " + browserCount);
					StaticLogger.LogInfo("# of resolvers: " + resolverCount);
					StaticLogger.LogInfo("# of record queriers: " + recordQuerierCount);
					StaticLogger.LogInfo("# of address queriers: " + addrQuerierCount);

					foreach (var entry in handler._processorDic)
					{
						try
						{
							if (entry.Value != null)
							{
								if (entry.Value.Processor != null)
								{
									entry.Value.Processor.Stop();
									StaticLogger.LogTrace("DNSSDService.Stop() (Processor) finished.");
								}
								if (entry.Value.Service != null)
								{
									entry.Value.Service.Stop();
									StaticLogger.LogTrace("DNSSDService.Stop() (Service) finished.");
								}
							}
						}
						catch (Exception ex)
						{
							StaticLogger.LogWarn("DNSSDService.Stop() failed of processor '" + entry.Key + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
						}
					}

					foreach (var key in handler._processorDic.Keys.ToArray())
					{
						handler.RemoveProcessorID(key);
					}
				}, this);
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Tries to find the associated processor ID with the <see cref="DNSSDService"/> object among processors.
		/// </summary>
		/// <param name="service">The <see cref="DNSSDService"/> object.</param>
		/// <param name="processorID">The assoicated processor ID.</param>
		/// <param name="data">The processor data</param>
		/// <returns>True if a processor is found with the specified DNSSDService object.</returns>
		private bool TryFindProcessor(DNSSDService service, out uint processorID, out ProcessorData data)
		{
			// assign a default output value
			processorID = uint.MaxValue;
			data = null;

			Debug.Assert(this._processorDic != null);
			lock (this._processorDic)
			{
				foreach (var entry in this._processorDic)
				{
					if (entry.Value is ProcessorData)
					{
						ProcessorData value = entry.Value as ProcessorData;
						if (value.Processor.Equals(service) || value.Service.Equals(service))
						{
							processorID = entry.Key;
							data = value;
							return true;
						}
					}
				}
			}

			return false;
		}

		/// <summary>
		/// Gets an available processor ID.
		/// </summary>
		/// <param name="processorID">Output the available processor ID.</param>
		/// <param name="processorData">The processor data.</param>
		/// <returns>True if there's an available processor ID.</returns>
		private bool GetAnAvailableProcessorID(out uint processorID, ProcessorData processorData)
		{
			// assign default output value
			processorID = uint.MaxValue;

			Debug.Assert(this._processorDic != null);
			lock (this._processorDic)
			{
				for (uint i = 0; i <= uint.MaxValue; ++i)
				{
					ushort newProcessorID = (ushort)(this._processorID + i);
					if (this._processorDic.TryAdd(newProcessorID, processorData) == false)
					{
						continue;
					}
					else
					{
						processorID = newProcessorID;
						processorData.ProcessorID = processorID;
						processorData.State = ProcessorState.ScheduleToRun;
						this._processorID = (ushort)(processorID + 1);
						return true;
					}
				}
			}

			return false;
		}

		/// <summary>
		/// Removes the specified processor ID.
		/// </summary>
		/// <param name="processorID">The processor ID.</param>
		private void RemoveProcessorID(uint processorID)
		{
			Debug.Assert(this._processorDic != null);
			ProcessorData data;
			this._processorDic.TryRemove(processorID, out data);
		}

		/// <summary>
		/// Determines if the service type is desired for this instance.
		/// </summary>
		/// <param name="serviceType">The service type. ex: _acer-extend._tcp or _vnc._sub._acer-extend._tcp</param>
		/// <param name="serviceName">The service name part of the service type. ex: _acer-extend</param>
		/// <param name="subType">The subtype part of the service type. ex: _vnc</param>
		/// <param name="connProtocol">The connection protocol part of the service type. ex: _tcp</param>
		/// <returns>Returns true if the service type is desired, false otherwise.</returns>
		private bool IsServiceTypeDesired(string serviceType, out string serviceName, out string subType, out ConnectionProtocol connProtocol)
		{
			// initialize output
			{
				serviceName = null;
				subType = null;
				connProtocol = ConnectionProtocol.Unknown;
			}

			if (string.IsNullOrWhiteSpace(serviceType))
			{
				StaticLogger.LogWarn("'serviceType' is unspecified.");
				return false;
			}

			string[] serviceTypeLabels = serviceType.TrimEnd('.').Split('.');
			if (serviceTypeLabels.Length < 2 || serviceTypeLabels.Length > 4)
			{
				StaticLogger.LogWarn("Bad format of service type: " + serviceType);
				return false;
			}

			serviceName = serviceTypeLabels[serviceTypeLabels.Length - 2];
			if (string.Compare(serviceName, this.ServiceName) != 0)
			{
				StaticLogger.LogWarn("Undesired service name: " + serviceName);
				return false;
			}

			if (serviceTypeLabels.Length == 4)
			{
				subType = serviceTypeLabels[0];
			}

			Enum connProcolValue = serviceTypeLabels[serviceTypeLabels.Length - 1].GetEnumValue(typeof(ConnectionProtocol));
			connProtocol = connProcolValue != null ? (ConnectionProtocol)connProcolValue : ConnectionProtocol.Unknown;
			if (connProtocol == ConnectionProtocol.Unknown || connProtocol != this.ConnProtocol)
			{
				StaticLogger.LogWarn("Undesired connection protocol: " + serviceTypeLabels[serviceTypeLabels.Length - 1]);
				return false;
			}

			return true;
		}

		/// <summary>
		/// Determines if the service type is desired for this instance.
		/// </summary>
		/// <param name="serviceType">The service type. ex: _acer-extend._tcp or _vnc._sub._acer-extend._tcp</param>
		/// <returns>Returns true if the service type is desired, false otherwise.</returns>
		private bool IsServiceTypeDesired(string serviceType)
		{
			string serviceName, subType;
			ConnectionProtocol connProtocol;
			return this.IsServiceTypeDesired(serviceType, out serviceName, out subType, out connProtocol);
		}

		/// <summary>
		/// Determines if the full name is desired for this instance.
		/// </summary>
		/// <param name="fullName">The full name without subtype. ex: Seoker._acer-extend._tcp.local.</param>
		/// <param name="instanceName">The instance name part of the full name. ex: Seoker</param>
		/// <param name="serviceName">The service name part of the full name. ex: _acer-extend</param>
		/// <param name="connProtocol">The connection protocol part of the service type. ex: _tcp</param>
		/// <param name="domain">The domain part of the full name. ex: local</param>
		/// <returns>Returns true if the full name is desired, false otherwise.</returns>
		private bool IsFullNameDesired(string fullName, out string instanceName, out string serviceName, out ConnectionProtocol connProtocol, out string domain)
		{
			// initialize output
			{
				instanceName = null;
				serviceName = null;
				connProtocol = ConnectionProtocol.Unknown;
				domain = null;
			}

			if (string.IsNullOrWhiteSpace(fullName))
			{
				StaticLogger.LogWarn("'fullName' is unspecified.");
				return false;
			}

			string[] labels = fullName.TrimEnd('.').Split('.');
			if (labels.Length < 4)
			{
				StaticLogger.LogWarn("Bad format of fullname: " + fullName);
				return false;
			}

			// get instance name
			instanceName = string.Join(".", labels.Take(labels.Length - 3));
			if (string.IsNullOrEmpty(instanceName))
			{
				StaticLogger.LogWarn("Instance name is unspecified.");
				return false;
			}

			// unescape
			instanceName = UnescapeInstanceName(instanceName);

			// get service name
			serviceName = labels[labels.Length - 3];
			if (string.Compare(serviceName, this.ServiceName) != 0)
			{
				StaticLogger.LogWarn("Undesired service name: " + serviceName);
				return false;
			}

			// get protocol
			Enum connProcolValue = labels[labels.Length - 2].GetEnumValue(typeof(ConnectionProtocol));
			connProtocol = connProcolValue != null ? (ConnectionProtocol)connProcolValue : ConnectionProtocol.Unknown;
			if (connProtocol == ConnectionProtocol.Unknown || connProtocol != this.ConnProtocol)
			{
				StaticLogger.LogWarn("Undesired connection protocol: " + labels[2]);
				return false;
			}

			// get domain
			domain = labels[labels.Length - 1];
			if (string.Compare(domain, this.Domain) != 0)
			{
				StaticLogger.LogWarn("Undesired domain: " + domain);
				return false;
			}

			return true;
		}

		/// <summary>
		/// Determines if the full name is desired for this instance.
		/// </summary>
		/// <param name="fullName">The full name without subtype. ex: Seoker._acer-extend._tcp.local.</param>
		/// <returns>Returns true if the full name is desired, false otherwise.</returns>
		private bool IsFullNameDesired(string fullName)
		{
			string instanceName, serviceName, domain;
			ConnectionProtocol connProtocol;
			return this.IsFullNameDesired(fullName, out instanceName, out serviceName, out connProtocol, out domain);
		}

		/// <summary>
		/// Escapes the instance name.
		/// </summary>
		/// <param name="instanceName">The instance name to escape.</param>
		/// <returns>The escaped instance name.</returns>
		private static string EscapeInstanceName(string instanceName)
		{
			try
			{
				// . => \.
				// \ => \\
				instanceName = System.Text.RegularExpressions.Regex.Replace(instanceName, @"([.\\])", "\\$1", System.Text.RegularExpressions.RegexOptions.Compiled);

				// unprintable characters (includes spaces) => \001 ~ \032
				instanceName = System.Text.RegularExpressions.Regex.Replace(instanceName, @"(?<value>[\x01-\x20])", m =>
				{
					try
					{
						return "\\0" + ((int)m.Groups["value"].Value[0]).ToString();
					}
					catch
					{
						return "\\" + m.Groups["value"].Value;
					}
				}, System.Text.RegularExpressions.RegexOptions.Compiled);
			}
			catch { }

			return instanceName;
		}

		/// <summary>
		/// Unescapes the instance name.
		/// </summary>
		/// <param name="instanceName">The instance name to unescape.</param>
		/// <returns>The unescaped instance name.</returns>
		private static string UnescapeInstanceName(string instanceName)
		{
			try
			{
				// \. => .
				// \\ => \
				instanceName = System.Text.RegularExpressions.Regex.Replace(instanceName, @"\\([\.\\])", "$1", System.Text.RegularExpressions.RegexOptions.Compiled);

				// \001 ~ \032 => unprintable characters (includes spaces)
				instanceName = System.Text.RegularExpressions.Regex.Replace(instanceName, @"\\(?<value>[01][0-9][0-9]|2[0-4][0-9]|25[0-5])", m =>
				{
					try
					{
						return ((char)int.Parse(m.Groups["value"].Value, System.Globalization.NumberStyles.Integer)).ToString();
					}
					catch
					{
						return "\\" + m.Groups["value"].Value;
					}
				}, System.Text.RegularExpressions.RegexOptions.Compiled);
			}
			catch { }

			return instanceName;
		}
	}
}