﻿using AirFusion.Common;
using Bonjour;
using System.Net;
using System.Net.Sockets;
using System.Text;

namespace AirFusion.Manager.DeviceManager
{
	public partial class DnsSdHandler
	{
		#region Enumerations

		/// <summary>
		/// Specifies the query method.
		/// </summary>
		public enum AddrQueryMethod
		{
			/// <summary>
			/// Signifies the default method.
			/// </summary>
			Default,

			/// <summary>
			/// Signifies the query should be performed exclusively via multicast DNS.
			/// </summary>
			ForceMulticast,

			/// <summary>
			/// Signifies the query is created as a long-lived unicast query.
			/// </summary>
			LongLivedQuery
		}

		#endregion
		#region Classes

		/// <summary>
		/// Represents the callback arguments of StartQueryAddress().
		/// </summary>
		public class AddressQuerierCallbackArgs : BaseCallbackArgs
		{
			/// <summary>
			/// Gets the hostname.
			/// </summary>
			public string Hostname { get; internal set; }

			/// <summary>
			/// Gets the the index of the interface that the address is found.
			/// </summary>
			public uint? InterfaceIndex { get; internal set; }

			/// <summary>
			/// Gets the address.
			/// </summary>
			public IPAddress Address { get; internal set; }

			/// <summary>
			/// Gets the query method.
			/// </summary>
			public AddrQueryMethod? QueryMethod { get; internal set; }

			/// <summary>
			/// Gets the indication that indicates if there's more data coming.
			/// </summary>
			public bool? MoreComing { get; internal set; }

			/// <summary>
			///  Gets the indication that indicates if the address is removed.
			/// </summary>
			public bool? IsRemoved { get; internal set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="AddressQuerierCallbackArgs"/> class.
			/// </summary>
			internal AddressQuerierCallbackArgs()
				: base()
			{
				this.Hostname = null;
				this.InterfaceIndex = null;
				this.Address = null;
				this.QueryMethod = null;
				this.MoreComing = null;
				this.IsRemoved = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + " (Base):\r\n" + base.ToString());
				builder.AppendLine(this.GetType().GetFullName() + ".Hostname: " + this.Hostname);
				builder.AppendLine(this.GetType().GetFullName() + ".InterfaceIndex: " + (this.InterfaceIndex.HasValue ? this.InterfaceIndex.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Address: " + (this.Address != null ? this.Address.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".QueryMethod: " + (this.QueryMethod.HasValue ? this.QueryMethod.Value.GetStringValue() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".MoreComing: " + (this.MoreComing.HasValue ? this.MoreComing.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".IsRemoved: " + (this.IsRemoved.HasValue ? this.IsRemoved.Value.ToString() : "N/A"));

				return builder.ToString().TrimEnd();
			}
		}

		#endregion
	}

	#region Extensions

	/// <summary>
	/// Extensions.
	/// </summary>
	internal static partial class AFExtensions
	{
		internal static Bonjour.DNSSDAddressFamily Transform(this AddressFamily addressFamily)
		{
			if (addressFamily == AddressFamily.InterNetwork)
			{
				return Bonjour.DNSSDAddressFamily.kDNSSDAddressFamily_IPv4;
			}
			else if (addressFamily == AddressFamily.InterNetworkV6)
			{
				return Bonjour.DNSSDAddressFamily.kDNSSDAddressFamily_IPv6;
			}
			else
			{
				return 0;
			}
		}

		internal static AddressFamily Transform(this DNSSDAddressFamily addressFamily)
		{
			if (addressFamily == DNSSDAddressFamily.kDNSSDAddressFamily_IPv4)
			{
				return AddressFamily.InterNetwork;
			}
			else if (addressFamily == DNSSDAddressFamily.kDNSSDAddressFamily_IPv6)
			{
				return AddressFamily.InterNetworkV6;
			}
			else
			{
				return AddressFamily.Unknown;
			}
		}
	}

	#endregion
}
