﻿using AirFusion.Common;
using System;
using System.Net;
using System.Text;

namespace AirFusion.Manager.DeviceManager
{
	public partial class DnsSdHandler
	{
		#region Enumerations

		/// <summary>
		/// Represents the resource record type.
		/// </summary>
		public enum DnsSdRRType
		{
			/// <summary>
			/// Unknown.
			/// </summary>
			Unknown = 0,

			/// <summary>
			/// A.
			/// </summary>
			[ParameterValue("A")]
			TypeA = 1,

			/// <summary>
			/// PTR.
			/// </summary>
			[ParameterValue("PTR")]
			TypePTR = 12,

			/// <summary>
			/// TXT.
			/// </summary>
			[ParameterValue("TXT")]
			TypeTXT = 16,

			/// <summary>
			/// AAAA.
			/// </summary>
			[ParameterValue("AAAA")]
			TypeAAAA = 28,

			/// <summary>
			/// SRV.
			/// </summary>
			[ParameterValue("SRV")]
			TypeSRV = 33,

			/// <summary>
			/// ANY.
			/// </summary>
			[ParameterValue("ANY")]
			TypeANY = 255
		}

		/// <summary>
		/// Represents the query method.
		/// </summary>
		public enum QueryRecordMethod
		{
			/// <summary>
			/// Default.
			/// </summary>
			Default,

			/// <summary>
			/// Signifies the query should be performed exclusively via multicast DNS.
			/// </summary>
			ForceMulticast,

			/// <summary>
			/// Signifies the query is created as a long-lived unicast query.
			/// </summary>
			LongLivedQuery
		}

		#endregion
		#region Resource Record Data

		/// <summary>
		/// Represents the interface of resource record data.
		/// </summary>
		public interface IResourceRecordData
		{
			/// <summary>
			/// Gets the raw data.
			/// </summary>
			/// <returns></returns>
			byte[] GetRawData();

			/// <summary>
			/// Sets the raw data.
			/// </summary>
			/// <param name="rawData"></param>
			void SetRawData(byte[] rawData);

			/// <summary>
			/// Gets the data which is being processed by the specified type.
			/// </summary>
			/// <param name="type">The type of the data.</param>
			/// <returns></returns>
			object GetData(out Type type);
		}

		// TODO:
		// implementations of IResourceRecordData: AAAA, SRV, TXT, PTR...

		/// <summary>
		/// Provides the 'A' type resource record.
		/// </summary>
		public class ResourceRecordDataA : IResourceRecordData
		{
			internal byte[] _rawData;
			internal IPAddress _data;

			/// <summary>
			/// Initializes the instance.
			/// </summary>
			/// <param name="rawData">The raw data of RR data.</param>
			public ResourceRecordDataA(byte[] rawData)
			{
				this._data = null;
				this.SetRawData(rawData);
			}

			/// <summary>
			/// Gets the raw data.
			/// </summary>
			/// <returns></returns>
			public byte[] GetRawData()
			{
				return this._rawData;
			}

			/// <summary>
			/// Sets the raw data.
			/// </summary>
			/// <param name="rawData"></param>
			public void SetRawData(byte[] rawData)
			{
				this._rawData = rawData;
			}

			/// <summary>
			/// Gets the data which is being processed by the specified type.
			/// </summary>
			/// <param name="type">The type of the data.</param>
			/// <returns></returns>
			public object GetData(out Type type)
			{
				type = typeof(IPAddress);

				if (this._data == null)
				{
					uint bits = BitConverter.ToUInt32(this._rawData, 0);
					this._data = new System.Net.IPAddress(bits);
				}

				return this._data;
			}

			/// <summary>
			/// Gets the string that represents the object.
			/// </summary>
			/// <returns></returns>
			public override string ToString()
			{
				Type type;
				object data = this.GetData(out type);
				return data != null ? data.ToString() : base.ToString();
			}
		}

		#endregion
		#region Classes

		/// <summary>
		/// Represents the callback argument of StartQueryRecord().
		/// </summary>
		public class RecordQuerierCallbackArgs : BaseCallbackArgs
		{
			/// <summary>
			/// Gets the instance name.
			/// </summary>
			public string InstanceName { get; internal set; }

			/// <summary>
			/// Gets the index of the interface that the record is answered.
			/// </summary>
			public uint? InterfaceIndex { get; internal set; }

			/// <summary>
			/// Gets the type of resource record.
			/// </summary>
			public DnsSdRRType? RRType { get; internal set; }

			/// <summary>
			/// Gets the data of resource record.
			/// </summary>
			public IResourceRecordData RRData { get; internal set; }

			/// <summary>
			/// Gets the query method.
			/// </summary>
			public QueryRecordMethod? QueryMethod { get; internal set; }

			/// <summary>
			/// Gets the indication that indicates if there's more data coming.
			/// </summary>
			public bool? MoreComing { get; internal set; }

			/// <summary>
			/// Gets the indication that indicates if the record is removed.
			/// </summary>
			public bool? IsRemoved { get; internal set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="RecordQuerierCallbackArgs"/> class.
			/// </summary>
			internal RecordQuerierCallbackArgs()
				: base()
			{
				this.InstanceName = null;
				this.InterfaceIndex = null;
				this.RRType = null;
				this.RRData = null;
				this.QueryMethod = null;
				this.MoreComing = null;
				this.IsRemoved = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + " (Base):\r\n" + base.ToString());
				builder.AppendLine(this.GetType().GetFullName() + ".InstanceName: " + this.InstanceName);
				builder.AppendLine(this.GetType().GetFullName() + ".InterfaceIndex: " + (this.InterfaceIndex.HasValue ? this.InterfaceIndex.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".RRType: " + (this.RRType.HasValue ? this.RRType.Value.GetStringValue() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".RRData: " + (this.RRData != null ? "\r\n" + this.RRData.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".QueryMethod: " + (this.QueryMethod.HasValue ? this.QueryMethod.Value.GetStringValue() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".MoreComing: " + (this.MoreComing.HasValue ? this.MoreComing.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".IsRemoved: " + (this.IsRemoved.HasValue ? this.IsRemoved.Value.ToString() : "N/A"));

				return builder.ToString().TrimEnd();
			}
		}

		#endregion
	}

	#region Extensions

	/// <summary>
	/// Extensions.
	/// </summary>
	internal static partial class AFExtensions
	{
		internal static Bonjour.DNSSDRRType Transform(this DnsSdHandler.DnsSdRRType rrType)
		{
			int value = (int)rrType;
			if (Enum.IsDefined(typeof(Bonjour.DNSSDRRType), value))
				return (Bonjour.DNSSDRRType)Enum.ToObject(typeof(Bonjour.DNSSDRRType), value);

			return (Bonjour.DNSSDRRType)0;
		}

		internal static DnsSdHandler.DnsSdRRType Transform(this Bonjour.DNSSDRRType rrType)
		{
			int value = (int)rrType;
			if (Enum.IsDefined(typeof(DnsSdHandler.DnsSdRRType), value))
				return (DnsSdHandler.DnsSdRRType)Enum.ToObject(typeof(DnsSdHandler.DnsSdRRType), value);

			return DnsSdHandler.DnsSdRRType.Unknown;
		}
	}

	#endregion
}
