﻿using AirFusion.Common;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class UserDeviceManager
	{
		internal sealed partial class MulticastDnsAdapter
		{
			/// <summary>
			/// Represents an instance of mDNS service of AirFusion.
			/// </summary>
			public sealed partial class MDnsInstance
			{
				#region Definitions

				// TXT record
				private const string TRKeyTxtRecordVersion = "TXTVER";

				private const string TRKeyDeviceName = "DNAME";
				private const string TRKeyUserName = "UNAME";
				private const string TRKeyInstanceID = "IID";
				private const string TRKeyOS = "OS";
				private const string TRValueTxtRecordVersion = "1";

				// others
				private const uint ResolverTimeout = 60000; // in milliseconds

				private const uint ReconfirmTimeout = 5000; // in milliseconds

				#endregion Definitions

				#region Delegates

				/// <summary>
				/// Represents the methods that will handle the callback of StartResolve().
				/// </summary>
				/// <param name="instance">The instance.</param>
				public delegate void ResolvedCallback(MDnsInstance instance);

				/// <summary>
				/// Represents the methods that will handle the callback of StartQueryAddress().
				/// </summary>
				/// <param name="instance">The instance.</param>
				/// <param name="interfaceIndex">The index of interface.</param>
				/// <param name="address">The address.</param>
				/// <param name="isRemoved">A value indicating whether the address is removed.</param>
				public delegate void AddressAnsweredCallback(MDnsInstance instance, uint interfaceIndex, IPAddress address, bool isRemoved);

				#endregion Delegates

				#region Properties

				/// <summary>
				/// Gets the version of TXT record.
				/// </summary>
				public string Version { get; internal set; }

				/// <summary>
				/// Gets the instance ID.
				/// </summary>
				public Guid? InstanceID { get; internal set; }

				/// <summary>
				/// Gets the device name.
				/// </summary>
				public string DeviceName { get; internal set; }

				/// <summary>
				/// Gets the user name.
				/// </summary>
				public string UserName { get; internal set; }

				/// <summary>
				/// Gets the type of operating system of the device.
				/// </summary>
				public AFUserDevice.OSTypeValue? OS { get; internal set; }

				/// <summary>
				/// Gets the addresses of the instance.
				/// </summary>
				public IDictionary<uint, IPAddress> Addresses
				{
					get
					{
						Debug.Assert(this._addresses != null);
						return this._addresses;
					}
				}

				/// <summary>
				/// Gets an address of the instance
				/// </summary>
				/// <remarks>If there's no address found yet, the value will be null.</remarks>
				public IPAddress Address
				{
					get
					{
						Debug.Assert(this.Addresses != null);
						return this.Addresses.Count > 0 ? this.Addresses.First().Value : null;
					}
				}

				/// <summary>
				/// Gets the port number of the instance.
				/// </summary>
				public ushort? Port { get; internal set; }

				/// <summary>
				/// Gets the hostname.
				/// </summary>
				/// <remarks>Leave empty for default value.</remarks>
				public string Hostname { get; internal set; }

				/// <summary>
				/// Gets the instance name.
				/// </summary>
				public string InstanceName { get; internal set; }

				/// <summary>
				/// Gets the collection of interface index.
				/// </summary>
				public ICollection<uint> InterfaceIndexes
				{
					get
					{
						return this._addresses != null ? this._addresses.Keys : null;
					}
				}

				/// <summary>
				/// Gets the TXT record.
				/// </summary>
				public DnsSdHandler.TxtRecord TxtRecord
				{
					get
					{
						DnsSdHandler.TxtRecord txtRecord = new DnsSdHandler.TxtRecord();

						if (!string.IsNullOrEmpty(this.Version))
						{
							txtRecord.SetValue(TRKeyTxtRecordVersion, this.Version);
						}
						if (this.InstanceID.HasValue)
						{
							txtRecord.SetValue(TRKeyInstanceID, this.InstanceID.Value.GetStringFormat());
						}
						if (!string.IsNullOrEmpty(this.DeviceName))
						{
							txtRecord.SetValue(TRKeyDeviceName, this.DeviceName);
						}
						if (!string.IsNullOrEmpty(this.UserName))
						{
							txtRecord.SetValue(TRKeyUserName, this.UserName);
						}
						if (this.OS.HasValue)
						{
							txtRecord.SetValue(TRKeyOS, this.OS.GetStringValue());
						}

						return txtRecord;
					}
					set
					{
						if (value != null)
						{
							if (value.ContainsKey(TRKeyTxtRecordVersion))
							{
								this.Version = value.GetValueForKey(TRKeyTxtRecordVersion);
							}
							if (value.ContainsKey(TRKeyInstanceID))
							{
								try
								{
									this.InstanceID = Guid.Parse(value.GetValueForKey(TRKeyInstanceID));
								}
								catch
								{
									this.InstanceID = null;
								}
							}
							if (value.ContainsKey(TRKeyDeviceName))
							{
								this.DeviceName = value.GetValueForKey(TRKeyDeviceName);
							}
							if (value.ContainsKey(TRKeyUserName))
							{
								this.UserName = value.GetValueForKey(TRKeyUserName);
							}
							if (value.ContainsKey(TRKeyOS))
							{
								try
								{
									this.OS = (AFUserDevice.OSTypeValue)value.GetValueForKey(TRKeyOS).GetEnumValue(typeof(AFUserDevice.OSTypeValue));
								}
								catch
								{
									this.OS = null;
								}
							}
						}
					}
				}

				/// <summary>
				/// Gets a value indicating whether there's a resolver running for this instance.
				/// </summary>
				public bool IsResolving
				{
					get
					{
						return this._resolverProcessorId.HasValue;
					}
				}

				/// <summary>
				/// Gets a value indicating whether there's an address querier for this instance.
				/// </summary>
				public bool IsAddrQuerying
				{
					get
					{
						return this._addrQuerierProcessorId.HasValue;
					}
				}

				/// <summary>
				/// Gets or sets the bound handler.
				/// </summary>
				internal DnsSdHandler Handler { get; set; }

				#endregion Properties

				#region Private/Protected/Internal Variables

				private uint? _resolverProcessorId;									// the processor ID of the resolver
				private uint? _addrQuerierProcessorId;								// the processor ID of the address querier
				private System.Timers.Timer _resolverTimer;							// the timer of resolver (the resolver can make network burden so that we'll stop it  with a timeout)
				private ConcurrentDictionary<uint, IPAddress> _addresses;			// the address pool
				private ResolvedCallback _resolvedCallback;							// the callback of resolver
				private AddressAnsweredCallback _addressAnsweredCallback;			// the callback of address answered

				#endregion Private/Protected/Internal Variables

				#region Methods

				/// <summary>
				/// Initializes a new instance of the <see cref="MDnsInstance"/> class.
				/// </summary>
				internal MDnsInstance()
				{
					this.Version = TRValueTxtRecordVersion;
					this.InstanceID = null;
					this.DeviceName = null;
					this.UserName = null;
					this.OS = null;
					this.Port = null;
					this.Hostname = null;
					this.InstanceName = null;
					this.Handler = null;
					this._resolverProcessorId = null;
					this._addrQuerierProcessorId = null;
					this._resolverTimer = null;
					this._addresses = new ConcurrentDictionary<uint, IPAddress>();
					this._resolvedCallback = null;
					this._addressAnsweredCallback = null;
				}

				/// <summary>
				/// Adds an address into the address pool.
				/// </summary>
				/// <param name="ifIndex">If index.</param>
				/// <param name="address">The address.</param>
				internal void AddAddress(uint ifIndex, IPAddress address)
				{
					this._addresses.AddOrUpdate(ifIndex, address, delegate(uint key, IPAddress value)
					{
						return address;
					});
				}

				/// <summary>
				/// Removes the address from the address pool.
				/// </summary>
				/// <param name="ifIndex">If index.</param>
				internal IPAddress RemoveAddress(uint ifIndex)
				{
					IPAddress address;
					this._addresses.TryRemove(ifIndex, out address);
					return address;
				}

				/// <summary>
				/// Starts to resolve the instance.
				/// </summary>
				/// <exception cref="System.ArgumentException" />
				/// <exception cref="System.ArgumentNullException" />
				/// <exception cref="System.InvalidOperationException" />
				/// <remarks>If callback is not specified, the caller won't be acknowledged as it is resolved.</remarks>
				public void StartResolve(ResolvedCallback callback = null)
				{
					try
					{
						StaticLogger.TrackEnter();

						// check prerequisite
						if (this.Handler == null)
						{
							throw new InvalidOperationException("There's no bundle DNS-SD handler for this instance.");
						}
						if (string.IsNullOrEmpty(this.InstanceName))
						{
							throw new InvalidOperationException("There's no instance name, cannot resolve this instance.");
						}

						lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNResolver)))
						{
							if (this.IsResolving)
							{
								throw new InvalidOperationException("There's already a resolver for this instance.");
							}

							this._resolvedCallback = callback;

							this._resolverProcessorId = this.Handler.StartResolveService(this.InstanceName, 0, true,
								(args) =>
								{
									this.OnServiceResolved(args, this._resolvedCallback);
								});

							// start timeout timer
							this._resolverTimer = new System.Timers.Timer(ResolverTimeout);
							this._resolverTimer.AutoReset = false;
							this._resolverTimer.Elapsed += (s, e) =>
							{
								this._resolverTimer.Dispose();
								this._resolverTimer = null;

								try
								{
									if (this.IsResolving)
									{
										StaticLogger.LogInfo("Resolver timeout, stop the resolver of '" + this.InstanceName + "'.");
										this.StopResolve();
									}
								}
								catch (Exception ex)
								{
									StaticLogger.LogError("Failed stopping resolver, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
								}
							};
							this._resolverTimer.Start();
						}
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}

				/// <summary>
				/// Stops resolving the instance.
				/// </summary>
				/// <exception cref="System.Exception" />
				/// <exception cref="System.InvalidOperationException" />
				public void StopResolve()
				{
					try
					{
						StaticLogger.TrackEnter();

						// check prerequisite
						if (this.Handler == null)
						{
							throw new InvalidOperationException("There's no bundle DNS-SD handler for this instance.");
						}

						lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNResolver)))
						{
							if (!this.IsResolving)
							{
								throw new InvalidOperationException("There's no resolver for this instance.");
							}

							this._resolvedCallback = null;

							try
							{
								if (this.Handler.IsResolving(this._resolverProcessorId.Value))
								{
									this.Handler.StopProcessor(this._resolverProcessorId.Value);
								}
							}
							catch (Exception ex)
							{
								throw new Exception("Failed stopping the resolver of '" + this.InstanceName + "'.", ex);
							}

							this._resolverProcessorId = null;

							// stop timer if such timer exist
							if (this._resolverTimer != null)
							{
								this._resolverTimer.Dispose();
								this._resolverTimer = null;
							}
						}
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}

				/// <summary>
				/// Starts to query the address.
				/// </summary>
				/// <exception cref="System.ArgumentException" />
				/// <exception cref="System.Exception" />
				/// <exception cref="System.InvalidOperationException" />
				/// <exception cref="System.NotSupportedException" />
				/// <remarks>If callback is not specified, the caller won't be acknowledged as it is resolved.</remarks>
				public void StartQueryAddress(AddressAnsweredCallback callback = null)
				{
					try
					{
						StaticLogger.TrackEnter();

						// check prerequisite
						if (this.Handler == null)
						{
							throw new InvalidOperationException("There's no bundle DNS-SD handler for this instance.");
						}
						if (string.IsNullOrEmpty(this.Hostname))
						{
							throw new InvalidOperationException("There's no hostname, cannot query the address of this instance.");
						}

						lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNAddrQuerier)))
						{
							if (this.IsAddrQuerying)
							{
								throw new InvalidOperationException("There's already an address querier for this instance.");
							}

							this._addressAnsweredCallback = callback;

							this._addrQuerierProcessorId = this.Handler.StartQueryAddress(this.Hostname, System.Net.Sockets.AddressFamily.InterNetwork, DnsSdHandler.AddrQueryMethod.LongLivedQuery, 0,
								(args) =>
								{
									this.OnAddressQuerierCallback(args, callback);
								});
						}
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}

				/// <summary>
				/// Stops querying the address.
				/// </summary>
				/// <exception cref="System.Exception" />
				/// <exception cref="System.InvalidOperationException" />
				public void StopQueryAddress()
				{
					try
					{
						StaticLogger.TrackEnter();

						// check prerequisite
						if (this.Handler == null)
						{
							throw new InvalidOperationException("There's no bundle DNS-SD handler for this instance.");
						}

						lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNAddrQuerier)))
						{
							if (!this.IsAddrQuerying)
							{
								throw new InvalidOperationException("There's no address querier for this instance.");
							}

							this._addressAnsweredCallback = null;

							try
							{
								if (this.Handler.IsAddressQuerying(this._addrQuerierProcessorId.Value))
								{
									this.Handler.StopProcessor(this._addrQuerierProcessorId.Value);
								}
							}
							catch (Exception ex)
							{
								throw new Exception("Failed stopping the address querier of '" + this.InstanceName + "'.", ex);
							}

							this._addrQuerierProcessorId = null;
						}
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}

				/// <summary>
				/// Reconfirms the endpoint (IP address &amp; port) asynchronously.
				/// </summary>
				/// <returns>
				/// The reconfirmed IP address, null if there's no IP address being reconfirmed.
				/// </returns>
				/// <exception cref="System.InvalidOperationException" />
				/// <exception cref="System.ArgumentException" />
				/// <remarks>
				/// 1. This function cannot solve AcerWireless cache issue, only solve Bonjour cache issue.
				/// 2. Since it's async call, only first endpoint will be returned.
				/// </remarks>
				public async Task<IPEndPoint[]> ReconfirmEndpointAsync()
				{
					try
					{
						StaticLogger.TrackEnter();

						// check prerequisite
						if (this.Handler == null)
						{
							throw new InvalidOperationException("There's no bundle DNS-SD handler for this instance.");
						}
						if (string.IsNullOrEmpty(this.InstanceName))
						{
							throw new InvalidOperationException("There's no instance name, cannot reconfirm the address of this instance.");
						}

						var tcs = new TaskCompletionSource<IPEndPoint[]>();
						lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNAddrQuerier)))
						{
							// start resolver first since the hostname might change but instance name doesn't change
							// (not use StartResolve function to avoid thread-safe issue, just create a new resolver/address querier)
							var resolverProcessorId = this.Handler.StartResolveService(this.InstanceName, 0, true,
								(resolverEventArgs) =>
								{
									if (resolverEventArgs.IsFaulted)
									{
										StaticLogger.LogError("Failed resolving '" + this.InstanceName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(resolverEventArgs.Exception));
										tcs.TrySetException(resolverEventArgs.Exception);
										return;
									}

									if (resolverEventArgs.MoreComing.HasValue && resolverEventArgs.MoreComing.Value)
									{
										StaticLogger.LogWarn("There's more data coming, skipped this record.");
										return;
									}

									// debug assertion
									Debug.Assert(this.Handler != null);
									Debug.Assert(resolverEventArgs.TxtRecord != null);
									Debug.Assert(resolverEventArgs.Hostname != null);
									Debug.Assert(resolverEventArgs.Port.HasValue);

									// stop original long-lived querier
									var restartAddressQuerier = false;
									var callback = this._addressAnsweredCallback; // keep the callback since we'll have to call another StartQueryAddress() later
									if (this.Hostname.Equals(resolverEventArgs.Hostname, StringComparison.InvariantCultureIgnoreCase) == false)
									{
										try
										{
											this.StopQueryAddress();
											restartAddressQuerier = true;
										}
										catch (Exception ex)
										{
											StaticLogger.LogWarn("Failed to stop querying address of the instance '" + this.InstanceName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
										}
									}

									// update values
									this.TxtRecord = resolverEventArgs.TxtRecord;
									this.Hostname = resolverEventArgs.Hostname;
									this.Port = resolverEventArgs.Port;

									// stop the resolver
									try
									{
										this.Handler.StopProcessor(resolverEventArgs.ProcessorID);
									}
									catch (Exception ex)
									{
										StaticLogger.LogWarn("Failed to stop the resolver, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
									}

									// start another long-lived address querier: this function should be called after the hostname is updated
									if (restartAddressQuerier)
									{
										try
										{
											this.StartQueryAddress(callback);
										}
										catch (Exception ex)
										{
											StaticLogger.LogWarn("Failed to start querying address of the instance '" + this.InstanceName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
										}
									}

									var addressList = new List<IPAddress>();

									// start query the address
									var querierProcessorId = this.Handler.StartQueryAddress(this.Hostname, System.Net.Sockets.AddressFamily.InterNetwork, DnsSdHandler.AddrQueryMethod.Default, 0,
										(querierCallbackArgs) =>
										{
											if (querierCallbackArgs.IsFaulted)
											{
												StaticLogger.LogError("Failed querying the address of '" + this.Hostname + "', exception:\r\n" + CommonUtils.GetExceptionMessages(querierCallbackArgs.Exception));
												return;
											}

											if (querierCallbackArgs.MoreComing.HasValue && querierCallbackArgs.MoreComing.Value && querierCallbackArgs.IsRemoved.HasValue && querierCallbackArgs.IsRemoved.Value)
											{
												StaticLogger.LogWarn("There's more data coming and the it's marked as removed, skipped this record.");
												return;
											}

											// add/update the address of the instance
											Debug.Assert(querierCallbackArgs.InterfaceIndex.HasValue);
											this.AddAddress(querierCallbackArgs.InterfaceIndex.Value, querierCallbackArgs.Address);

											// output
											if (querierCallbackArgs.Address != null)
											{
												addressList.Add(querierCallbackArgs.Address);
											}

											// if there's no more data and there's at least one address, set result of the task
											if ((querierCallbackArgs.MoreComing.HasValue || !querierCallbackArgs.MoreComing.Value) && addressList.Count() > 0 && !tcs.Task.IsCompleted)
											{
												if (this.Port.HasValue && this.Port.Value > 0)
												{
													tcs.TrySetResult(addressList.Select(address => new IPEndPoint(address, this.Port.Value)).ToArray());
												}
												else
												{
													tcs.TrySetException(new Exception("Failed getting the port number."));
												}
											}
										});

									// start timeout timer
									var querierTimer = new System.Timers.Timer(ReconfirmTimeout);
									querierTimer.AutoReset = false;
									querierTimer.Elapsed += (s, e) =>
									{
										lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNAddrQuerier)))
										{
											if (this.Handler.IsAddressQuerying(querierProcessorId))
											{
												this.Handler.StopProcessor(querierProcessorId);
											}

											if (!tcs.Task.IsCompleted)
											{
												tcs.TrySetException(new TimeoutException("The operation has timed out with " + ReconfirmTimeout + "ms."));
											}
										}
										querierTimer.Dispose();
										querierTimer = null;
									};
									querierTimer.Start();
								});

							// start timeout timer
							var resolverTimer = new System.Timers.Timer(ReconfirmTimeout);
							resolverTimer.AutoReset = false;
							resolverTimer.Elapsed += (s, e) =>
							{
								lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNAddrQuerier)))
								{
									if (this.Handler.IsResolving(resolverProcessorId))
									{
										this.Handler.StopProcessor(resolverProcessorId);
									}

									if (!tcs.Task.IsCompleted)
									{
										tcs.TrySetException(new TimeoutException("The operation has timed out with " + ReconfirmTimeout + "ms."));
									}
								}
								resolverTimer.Dispose();
								resolverTimer = null;
							};
							resolverTimer.Start();
						}

						return await tcs.Task;

						/*
						return await Task<IPEndPoint>.Run(delegate()
						{
							// stop resolver if any
							try
							{
								if (this.IsResolving)
								{
									this.StopResolve();
								}
							}
							catch (Exception ex)
							{
								StaticLogger.LogWarn("Failed to stop resolving the instance '" + this.InstanceName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
							}

							var eventHandle = new AutoResetEvent(false); // create an event handle
							IPAddress address = null;
							int port = 0;

							// start resolver first since the hostname might change but instance name doesn't change
							// (not use StartResolve function to avoid thread-safe issue, just create a new resolver/address querier)
							var resolverProcessorId = this.Handler.StartResolveService(this.InstanceName, 0, true,
								(resolverEventArgs) =>
								{
									if (resolverEventArgs.IsFaulted)
									{
										StaticLogger.LogError("Failed resolving '" + this.InstanceName + "', exception:\r\n" + CommonUtils.GetExceptionMessages(resolverEventArgs.Exception));
										return;
									}

									if (resolverEventArgs.MoreComing.HasValue && resolverEventArgs.MoreComing.Value)
									{
										StaticLogger.LogWarn("There's more data coming, skipped this record.");
										return;
									}

									// debug assertion
									Debug.Assert(this.Handler != null);
									Debug.Assert(resolverEventArgs.TxtRecord != null);
									Debug.Assert(resolverEventArgs.Hostname != null);
									Debug.Assert(resolverEventArgs.Port.HasValue);

									// update values
									this.TxtRecord = resolverEventArgs.TxtRecord;
									this.Hostname = resolverEventArgs.Hostname;
									this.Port = resolverEventArgs.Port;

									// output
									port = resolverEventArgs.Port.Value;

									// stop the resolver
									try
									{
										this.Handler.StopProcessor(resolverEventArgs.ProcessorID);
									}
									catch (Exception ex)
									{
										StaticLogger.LogWarn("Failed to stop the resolver, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
									}

									// TODO:
									// If the hostname is changed, need to restart another long-lived address query

									// start query the address
									var querierProcessorId = this.Handler.StartQueryAddress(this.Hostname, System.Net.Sockets.AddressFamily.InterNetwork, DnsSdHandler.AddrQueryMethod.Default, 0,
										(querierCallbackArgs) =>
										{
											if (querierCallbackArgs.IsFaulted)
											{
												StaticLogger.LogError("Failed querying the address of '" + this.Hostname + "', exception:\r\n" + CommonUtils.GetExceptionMessages(querierCallbackArgs.Exception));
												return;
											}

											if (querierCallbackArgs.MoreComing.HasValue && querierCallbackArgs.MoreComing.Value)
											{
												StaticLogger.LogWarn("There's more data coming, skipped this record.");
												return;
											}

											// add/update the address of the instance
											Debug.Assert(querierCallbackArgs.InterfaceIndex.HasValue);
											this.AddAddress(querierCallbackArgs.InterfaceIndex.Value, querierCallbackArgs.Address);

											// output
											address = querierCallbackArgs.Address;

											try
											{
												if (eventHandle != null)
												{
													eventHandle.Set();
												}
											}
											catch { }
										});
								});

							try
							{
								if (eventHandle != null)
								{
									if (eventHandle.WaitOne(timeout) == false)
									{
										throw new TimeoutException("Failed getting mDNS record within " + timeout + "ms.");
									}
								}
							}
							catch { }

							if (address == null || address.Equals(IPAddress.None) || port == 0)
							{
								throw new Exception("The address or port number is invalid from mDNS.");
							}

							return new IPEndPoint(address, port);
						});
						*/
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}

				/// <summary>
				/// Gets a value indicating whether the instance is available to connect, i.e. the IP address &amp; port number are valid.
				/// </summary>
				/// <returns></returns>
				/// <remarks>The port number of Android device might be 0 due to an issue, in that case the return value will be false.</remarks>
				public bool IsConnectable()
				{
					return this.Address != null && !this.Address.Equals(IPAddress.None) && this.Port.HasValue && this.Port.Value != 0;
				}

				/// <summary>
				/// Returns a <see cref="System.String" /> that represents this instance.
				/// </summary>
				/// <returns>
				/// A <see cref="System.String" /> that represents this instance.
				/// </returns>
				public override string ToString()
				{
					StringBuilder builder = new StringBuilder();
					builder.AppendLine(this.GetType().GetFullName() + ".Version: " + (this.Version ?? "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".InstanceID: " + (this.InstanceID.HasValue ? this.InstanceID.Value.GetStringFormat() : "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".DeviceName: " + (this.DeviceName ?? "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".UserName: " + (this.UserName ?? "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".OS: " + (this.OS.HasValue ? this.OS.Value.ToString() : "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".Addresses: " + (this.Addresses != null && this.Addresses.Count > 0 ? string.Join(", ", this.Addresses.Values) : "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".Port: " + (this.Port.HasValue ? this.Port.Value.ToString() : "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".Hostname: " + (this.Hostname ?? "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".InstanceName: " + (this.InstanceName ?? "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".InterfaceIndex: " + (this.InterfaceIndexes != null && this.InterfaceIndexes.Count > 0 ? string.Join(", ", this.InterfaceIndexes) : "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".TxtRecord: " + (this.TxtRecord != null ? "\r\n" + this.TxtRecord.ToString() : "N/A"));
					builder.AppendLine(this.GetType().GetFullName() + ".IsResolving: " + this.IsResolving);
					builder.AppendLine(this.GetType().GetFullName() + ".IsAddrQuerying: " + this.IsAddrQuerying);

					return builder.ToString().TrimEnd();
				}

				#endregion Methods

				#region Private/Protected/Internal Methods

				/// <summary>
				/// The callback of StartResolveService().
				/// </summary>
				/// <param name="args">The <see cref="DnsSdHandler.ResolverCallbackArgs"/> instance containing the callback arguments.</param>
				/// <param name="callback">The callback function.</param>
				private void OnServiceResolved(DnsSdHandler.ResolverCallbackArgs args, ResolvedCallback callback)
				{
					try
					{
						StaticLogger.TrackEnter();

						// log parameters
						StaticLogger.LogDebug("args: " + (args != null ? args.ToString() : "N/A"));

						// check parameter
						if (args == null)
						{
							StaticLogger.LogError("args=NULL");
							return;
						}

						// debug assertion
						Debug.Assert(args.Handler != null);

						if (args.IsFaulted)
						{
							StaticLogger.LogError("Failed resolving '" + args.InstanceName + "'.");
						}
						else
						{
							// debug assertion
							Debug.Assert(args.InstanceName != null);
							Debug.Assert(args.InterfaceIndex.HasValue);
							Debug.Assert(args.Port.HasValue);
							Debug.Assert(args.Hostname != null);
							Debug.Assert(args.TxtRecord != null);
							Debug.Assert(args.MoreComing.HasValue);

							// skip the record if there's more data coming
							if (args.MoreComing.Value)
							{
								StaticLogger.LogInfo("Skip this record since there's more data coming.");
								return;
							}

							// update values
							this.Port = args.Port.Value;
							this.Hostname = args.Hostname;
							this.TxtRecord = args.TxtRecord;
							(this.Addresses as ConcurrentDictionary<uint, IPAddress>).AddOrUpdate(args.InterfaceIndex.Value, IPAddress.None,
								(ifIndex, address) =>
								{
									return null;
								});
						}

						if (callback != null)
						{
							try
							{
								callback(this);
							}
							catch (Exception ex)
							{
								StaticLogger.LogError("Unhandled exception of callback function, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
							}
						}
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}

				/// <summary>
				/// The callback of StartQueryAddress().
				/// </summary>
				/// <param name="args">The <see cref="DnsSdHandler.AddressQuerierCallbackArgs"/> instance containing the callback arguments.</param>
				/// <param name="callback">The callback function.</param>
				private void OnAddressQuerierCallback(DnsSdHandler.AddressQuerierCallbackArgs args, AddressAnsweredCallback callback)
				{
					try
					{
						StaticLogger.TrackEnter();

						// log parameters
						StaticLogger.LogDebug("args: " + (args != null ? args.ToString() : "N/A"));

						// check parameter
						if (args == null)
						{
							StaticLogger.LogWarn("args=NULL");
							return;
						}

						// debug assertion
						Debug.Assert(args.Handler != null);

						if (args.IsFaulted)
						{
							StaticLogger.LogError("Failed querying '" + args.Hostname + "'.");
						}
						else
						{
							// debug assertion
							Debug.Assert(args.Hostname != null);
							Debug.Assert(args.QueryMethod.HasValue);
							Debug.Assert(args.InterfaceIndex.HasValue);
							Debug.Assert(args.MoreComing.HasValue);
							Debug.Assert(args.IsRemoved.HasValue);

							// add or remove the address
							if (args.IsRemoved.HasValue && args.InterfaceIndex.HasValue && args.IsRemoved.Value)
							{
								this.Addresses.Remove(args.InterfaceIndex.Value);
							}
							if (args.IsRemoved.HasValue && args.InterfaceIndex.HasValue && !args.IsRemoved.Value && args.Address != null && !args.Address.Equals(IPAddress.None))
							{
								this.Addresses[args.InterfaceIndex.Value] = args.Address;
							}

							// fire event if it's added event or MoreComing flag is unspecified
							if ((!args.IsRemoved.Value || !args.MoreComing.Value))
							{
								if (callback != null)
								{
									try
									{
										callback(this, args.InterfaceIndex.Value, args.Address, args.IsRemoved.Value);
									}
									catch (Exception ex)
									{
										StaticLogger.LogError("Unhandled exception of callback function, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
									}
								}
							}
						}
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}

				/// <summary>
				/// Handles the record querier callback of A record.
				/// </summary>
				/// <param name="args">The <see cref="DnsSdHandler.RecordQuerierCallbackArgs"/> instance containing the callback arguments.</param>
				/// <param name="address">The result IP address.</param>
				/// <param name="moreCallback">A value indicating if there's more callback.</param>
				/// <returns>Indicating whether the callback is valid.</returns>
				private bool HandleRecordAQuerierCallback(DnsSdHandler.RecordQuerierCallbackArgs args, out IPAddress address, out bool moreCallback)
				{
					try
					{
						StaticLogger.TrackEnter();

						// assign default output values
						address = null;
						moreCallback = false;

						// log parameters
						StaticLogger.LogDebug("args: " + (args != null ? args.ToString() : "N/A"));

						// check parameter
						if (args == null)
						{
							StaticLogger.LogWarn("args=NULL");
							return false;
						}

						// debug assertion
						Debug.Assert(args.Handler != null);

						if (args.IsFaulted)
						{
							StaticLogger.LogError("Failed reconfirming the address of '" + args.InstanceName + "'.");
							return false;
						}
						else
						{
							// debug assertion
							Debug.Assert(args.InstanceName != null);
							Debug.Assert(args.QueryMethod.HasValue);
							Debug.Assert(args.InterfaceIndex.HasValue);
							Debug.Assert(args.MoreComing.HasValue);
							Debug.Assert(args.IsRemoved.HasValue);

							// add or remove the address
							if (args.IsRemoved.HasValue && args.InterfaceIndex.HasValue && args.IsRemoved.Value)
							{
								this.Addresses.Remove(args.InterfaceIndex.Value);
							}
							if (args.IsRemoved.HasValue && args.InterfaceIndex.HasValue && !args.IsRemoved.Value && args.RRData != null && args.RRData is DnsSdHandler.ResourceRecordDataA)
							{
								DnsSdHandler.ResourceRecordDataA recordA = args.RRData as DnsSdHandler.ResourceRecordDataA;
								address = this.Addresses[args.InterfaceIndex.Value] = recordA._data;
							}

							moreCallback = args.MoreComing.Value;

							return true;
						}
					}
					finally
					{
						StaticLogger.TrackLeave();
					}
				}

				#endregion Private/Protected/Internal Methods
			}
		}
	}

	#region Extensions

	internal static partial class AFExtensions
	{
		/// <summary>
		/// Gets the <see cref="AFUserDevice"/> form for the specified mDNS instance.
		/// </summary>
		/// <param name="instance">The instance.</param>
		/// <returns></returns>
		internal static AFUserDevice ToDevice(this UserDeviceManager.MulticastDnsAdapter.MDnsInstance instance)
		{
			return new AFUserDevice()
			{
				InstanceID = instance.InstanceID.HasValue ? instance.InstanceID.Value : Guid.Empty,
				DeviceName = instance.DeviceName,
				UserName = instance.UserName,
				OSType = instance.OS.HasValue ? instance.OS.Value : AFUserDevice.OSTypeValue.Unknown,
				Address = instance.Addresses != null && instance.Addresses.Count > 0 ? instance.Addresses.First().Value : null,
				WebServicePort = instance.Port.HasValue ? instance.Port.Value : (ushort)0
			};
		}

		/// <summary>
		/// Gets the mDNS instance form for the specified device.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <returns></returns>
		internal static UserDeviceManager.MulticastDnsAdapter.MDnsInstance ToInstance(this AFUserDevice device)
		{
			return new UserDeviceManager.MulticastDnsAdapter.MDnsInstance()
			{
				InstanceName = device.InstanceID.GetStringFormat(),
				InstanceID = device.InstanceID,
				DeviceName = device.DeviceName,
				UserName = device.UserName,
				OS = device.OSType,
				Port = device.WebServicePort
			};
		}

		/// <summary>
		/// Fills from the specified <see cref="UserDeviceManager.MulticastDnsAdapter.MDnsInstance"/> object if such values are specified.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <param name="instance">The instance.</param>
		internal static void FillFromIfSpecified(this AFUserDevice device, UserDeviceManager.MulticastDnsAdapter.MDnsInstance instance)
		{
			// debug assertion
			Debug.Assert(instance != null);

			if (instance.InstanceID.HasValue && instance.InstanceID.Value.Equals(Guid.Empty) == false)
			{
				instance.InstanceID = instance.InstanceID.Value;
			}

			if (instance.DeviceName != null)
			{
				device.DeviceName = instance.DeviceName;
			}

			if (instance.UserName != null)
			{
				device.UserName = instance.UserName;
			}

			if (instance.OS.HasValue)
			{
				device.OSType = instance.OS.Value;
			}

			// TODO: need to support multiple addresses
			if (instance.InterfaceIndexes.Count > 0)
			{
				device.Address = instance.Addresses[instance.InterfaceIndexes.First()];
			}

			if (instance.Port.HasValue)
			{
				device.WebServicePort = instance.Port.Value;
			}
		}
	}

	#endregion Extensions
}
