﻿using AirFusion.Common;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class UserDeviceManager
	{
		/// <summary>
		/// Provides the functionality of mDNS.
		/// </summary>
		internal sealed partial class MulticastDnsAdapter
		{
			#region Properties

			/// <summary>
			/// Gets or sets the service name. e.g. _air-fusion
			/// </summary>
			public string ServiceName { get; private set; }

			/// <summary>
			/// Gets or sets the name of registered instance.
			/// </summary>
			public MDnsInstance RegisteredInstance { get; private set; }

			/// <summary>
			/// Gets the instance store.
			/// </summary>
			private ConcurrentDictionary<string, MDnsInstance> InstanceStore { get; set; }

			/// <summary>
			/// Gets a value that indicates whether the observer is live.
			/// </summary>
			public bool IsObserving
			{
				get
				{
					return this._handler != null && this._observationProcessorId.HasValue && this._handler.IsBrowsing(this._observationProcessorId.Value);
				}
			}

			#endregion Properties

			#region Events

			/// <summary>
			/// Occurs when a service instance is found.
			/// </summary>
			public event ServiceFoundEventHandler ServiceFound;

			/// <summary>
			/// Occurs when a service instance is lost.
			/// </summary>
			public event ServiceLostEventHandler ServiceLost;

			/// <summary>
			/// Occurs when the address of a service instance is updated.
			/// </summary>
			public event AddressUpdatedEventHandler AddressUpdated;

			#endregion Events

			#region Public Methods

			/// <summary>
			/// Initializes a new instance of the <see cref="MulticastDnsAdapter"/> class.
			/// </summary>
			/// <param name="serviceName">The service name. e.g. _air-fusion</param>
			/// <exception cref="System.ArgumentException" />
			/// <exception cref="System.Exception" />
			/// <exception cref="System.NotSupportedException" />
			/// <remarks>This instance must be initialized by the main thread.</remarks>
			public MulticastDnsAdapter(string serviceName)
			{
				if (RegisteredServiceNames != null && RegisteredServiceNames.Contains(serviceName))
				{
					throw new ArgumentException("There's already an existing instance with the specified service name '" + serviceName + "'.");
				}

				this._handler = new DnsSdHandler(serviceName, DnsSdHandler.ConnectionProtocol.Tcp, null); // this instance must be initialized by the main thread.

				this.ServiceName = serviceName;
				this.RegisteredInstance = null;
				this.InstanceStore = new ConcurrentDictionary<string, MDnsInstance>();
				this._observationProcessorId = null;
				this._registrarProcessorId = null;
				//this._reconfirmedInstanceIDs = new ConcurrentDictionary<Guid, HashSet<AutoResetEvent>>();

				RegisteredServiceNames.Add(serviceName);
			}

			/// <summary>
			/// Initializes this instance.
			/// </summary>
			/// <exception cref="System.ArgumentException" />
			/// <exception cref="System.Exception" />
			/// <exception cref="System.NotSupportedException" />
			public void Initialize()
			{
				try
				{
					StaticLogger.TrackEnter();

					this._handler.Initialize();
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Registers the service with the instance name.
			/// </summary>
			/// <param name="instanceRecord">The record of the instance.</param>
			/// <exception cref="System.ArgumentNullException" />
			/// <exception cref="System.Exception" />
			/// <exception cref="System.InvalidOperationException" />
			public void RegisterService(MDnsInstance instanceRecord)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("instance: " + (instanceRecord != null ? "\r\n" + instanceRecord.ToString() : "N/A"));

					// check parameter
					if (instanceRecord == null)
					{
						throw new ArgumentNullException("device");
					}
					if (!instanceRecord.Port.HasValue || instanceRecord.Port.Value == 0)
					{
						throw new ArgumentException("The specified port number is invalid.");
					}
					if (instanceRecord.Handler != null && instanceRecord.Handler != this._handler)
					{
						throw new InvalidOperationException("The specified instance has already being bound with another handler.");
					}

					// check prerequisite
					if (this._handler == null)
					{
						throw new InvalidOperationException("The instance is not initialized yet.");
					}

					// register the instance
					lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNRegister)))
					{
						if (this.RegisteredInstance != null)
						{
							throw new InvalidOperationException("There's already a registered instance.");
						}

						try
						{
							Debug.Assert(this._handler != null);
							instanceRecord.Handler = this._handler;
							this._registrarProcessorId = this._handler.RegisterService(instanceRecord.InstanceName, null, instanceRecord.Port.Value, instanceRecord.TxtRecord, 0, this.OnRegisterCallback);
							this.RegisteredInstance = instanceRecord;
						}
						catch (Exception ex)
						{
							throw new Exception("Failed registering mDNS service.", ex);
						}
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Unregisters the service.
			/// </summary>
			public void UnregisterService()
			{
				try
				{
					StaticLogger.TrackEnter();

					// check prerequisite
					if (this._handler == null)
					{
						throw new InvalidOperationException("The instance is not initialized yet.");
					}

					// unregister the instance
					lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNRegister)))
					{
						if (this.RegisteredInstance == null)
						{
							throw new InvalidOperationException("There's no registered instance.");
						}

						try
						{
							Debug.Assert(this._handler != null);
							if (this._registrarProcessorId.HasValue)
							{
								this._handler.StopProcessor(this._registrarProcessorId.Value);
							}
							this.RegisteredInstance = null;
						}
						catch (Exception ex)
						{
							throw new Exception("Failed unregistering mDNS service.", ex);
						}
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Starts the observation.
			/// </summary>
			/// <exception cref="System.ArgumentNullException" />
			/// <exception cref="System.Exception" />
			/// <exception cref="System.InvalidOperationException" />
			public void StartObservation()
			{
				try
				{
					StaticLogger.TrackEnter();

					lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNObserver)))
					{
						if (this.IsObserving)
						{
							throw new InvalidOperationException("There's already an observer for this instance.");
						}

						Debug.Assert(this._handler != null);
						this._observationProcessorId = this._handler.StartBrowseService(0, this.OnBrowserCallback);
						StaticLogger.LogInfo("Observation processor ID: " + this._observationProcessorId);
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Stops the observation.
			/// </summary>
			/// <exception cref="System.ArgumentException" />
			/// <exception cref="System.InvalidOperationException" />
			public void StopObservation()
			{
				try
				{
					StaticLogger.TrackEnter();

					lock (CommonUtils.GetLocker(new Tuple<object, string>(this, LNObserver)))
					{
						if (!this.IsObserving)
						{
							throw new InvalidOperationException("There's no observer for this instance.");
						}

						Debug.Assert(this._handler != null);
						Debug.Assert(this._observationProcessorId.HasValue);
						this._handler.StopProcessor(this._observationProcessorId.Value);
						this._observationProcessorId = null;
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Starts an immediate browser to trigger the multicast query.
			/// </summary>
			/// <param name="duration">The duration, in milliseconds.</param>
			/// <exception cref="System.ArgumentException" />
			/// <exception cref="System.Exception" />
			/// <exception cref="System.InvalidOperationException" />
			public void ImmediateBrowse(double duration = 1000)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("duration: " + duration);

					// check parameters
					if (duration <= 0)
					{
						throw new ArgumentException("The duration must larger than zero.");
					}

					try
					{
						Debug.Assert(this._handler != null);
						uint processorId = this._handler.StartBrowseService(0, null);
						System.Timers.Timer timer = new System.Timers.Timer(duration);
						timer.AutoReset = false;
						timer.Elapsed += (s, e) =>
						{
							try
							{
								this._handler.StopProcessor(processorId);
							}
							catch (Exception ex)
							{
								StaticLogger.LogWarn("Failed stopping the browser, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
							}
						};
						timer.Start();
					}
					catch (Exception ex)
					{
						throw new Exception("Failed starting the browser.", ex);
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Gets the instance record of the specified instance ID.
			/// </summary>
			/// <param name="instanceID">The instance ID.</param>
			/// <returns></returns>
			public MulticastDnsAdapter.MDnsInstance GetInstance(Guid instanceID)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("instanceID: " + instanceID.GetStringFormat());

					try
					{
						Debug.Assert(this.InstanceStore != null);
						lock (CommonUtils.GetLocker(this.InstanceStore))
						{
							return this.InstanceStore.Where(entry => entry.Value.InstanceID.HasValue && entry.Value.InstanceID.Value.Equals(instanceID)).FirstOrDefault().Value;
						}
					}
					catch
					{
						return null;
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Gets the address of the specified instance ID.
			/// </summary>
			/// <param name="instanceID">The instance ID.</param>
			/// <returns>The address dictionary of the specified instance ID, null if such instance is not present.</returns>
			public IDictionary<uint, IPAddress> GetAddresses(Guid instanceID)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("instanceID: " + instanceID.GetStringFormat());

					try
					{
						Debug.Assert(this.InstanceStore != null);
						lock (CommonUtils.GetLocker(this.InstanceStore))
						{
							return this.InstanceStore.Where(entry => entry.Value.InstanceID.HasValue && entry.Value.InstanceID.Value.Equals(instanceID)).Select(entry => entry.Value.Addresses).FirstOrDefault();
						}
					}
					catch
					{
						return null;
					}
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Reconfirms the endpoint (IP address &amp; port) asynchronously.
			/// </summary>
			/// <param name="instanceID">The instance ID.</param>
			/// <returns></returns>
			/// <remarks>
			/// 1. Currently, this function cannot solve AcerWireless cache issue, only solve Bonjour cache issue.
			/// 2. Since it's async call, only first endpoint will be returned.
			/// </remarks>
			public async Task<IPEndPoint[]> ReconfirmEndpointAsync(Guid instanceID)
			{
				try
				{
					StaticLogger.TrackEnter();

					// log parameters
					StaticLogger.LogDebug("instanceID: " + instanceID.GetStringFormat());

					var instance = this.GetInstance(instanceID);
					if (instance == null)
					{
						throw new Exception("The instance of instance ID '" + instanceID + "' doesn't exist, cannot reconfirm the address.");
					}

					/*
					// Two scenarios:
					// 1. In generic AP, the IP address of Android might changed without changing the instance name,
					//    we'll resolve the instance name again and get the updated IP address.
					//    => Bonjour cache issue
					// 2. In AcerWireless, there's a mDNS controller on the infra (wireless controller), the records would be cached by its own.
					//    In this case, we cannot guarantee to get the updated one, need try to start another immediate browser to check if there's another instance with the instance ID created.
					//    Also, we need to reconfirm the SRV & A record by DNSSDService.QueryRecord()
					//    => Cisco + Bonjour cahce issue

					var taskList = new List<Task<IPAddress>>();

					// solution 1. just to resolve the instance again
					taskList.Add(instance.ReconfirmAddressAsync(timeout));

					// solution 2. start another browser, check if another instance created with the instance ID
					// TODO:
					// Rewrite below, use TaskCompletionSource to convert APM model to TAP model
					taskList.Add(Task<IPAddress>.Factory.StartNew(() =>
					{
						IPAddress address = null;

						// get the bag
						var eventHandleBags = this._reconfirmedInstanceIDs.AddOrUpdate(instanceID, new HashSet<AutoResetEvent>(),
							(id, eventHandleBag) =>
							{
								return eventHandleBag;
							});

						// create a new event handle and insert into the bag
						var eventHandle = new AutoResetEvent(false);
						eventHandleBags.Add(eventHandle);

						// immediate browse
						try
						{
							this.ImmediateBrowse(timeout);
						}
						catch (Exception ex)
						{
							StaticLogger.LogWarn("Failed triggering an immediate browse, exception:\r\n" + AFCommonUtils.GetExceptionMessages(ex));
						}

						try
						{
							if (eventHandle != null)
							{
								eventHandle.WaitOne(timeout);
								if (eventHandleBags != null)
								{
									eventHandleBags.Remove(eventHandle);
								}
								if (eventHandleBags.Count == 0)
								{
									this._reconfirmedInstanceIDs.TryRemove(instanceID, out eventHandleBags);
								}
							}
						}
						catch (Exception ex)
						{
							StaticLogger.LogWarn("Unhandled exception:\r\n" + AFCommonUtils.GetExceptionMessages(ex));
						}

						return address;
					}));

					if (taskList != null && taskList.Count > 0)
					{
						int completed = 0;
						while (completed < taskList.Count)
						{
							Task<IPAddress> task = await Task.WhenAny(taskList);
							completed++;
							if (!task.IsFaulted && task.Result != null)
							{
								StaticLogger.LogInfo("By reconfirmation through mDNS, the IP address of device '" + instanceID + "' is " + task.Result);
								return task.Result;
							}
						}
					}

					return null;
					*/

					// currently, just check the instance record with original instance name
					return await instance.ReconfirmEndpointAsync().ConfigureAwait(continueOnCapturedContext: false);
				}
				finally
				{
					StaticLogger.TrackLeave();
				}
			}

			/// <summary>
			/// Gets the collection of live mDNS instances.
			/// </summary>
			/// <returns></returns>
			public ICollection<MDnsInstance> GetInstances()
			{
				Debug.Assert(this.InstanceStore != null);
				lock (CommonUtils.GetLocker(this.InstanceStore))
				{
					return this.InstanceStore.Values;
				}
			}

			/// <summary>
			/// Gets the collection of instance IDs from live mDNS instances.
			/// </summary>
			/// <returns></returns>
			public ICollection<Guid> GetInstanceIDs()
			{
				Debug.Assert(this.InstanceStore != null);
				lock (CommonUtils.GetLocker(this.InstanceStore))
				{
					return this.InstanceStore.Values.Where(instance => instance.InstanceID.HasValue).Select(instance => instance.InstanceID.Value).ToArray();
				}
			}

			/// <summary>
			/// Determines if the an instance exists with the specified instance ID.
			/// </summary>
			/// <param name="instanceID">The instance ID.</param>
			/// <returns></returns>
			public bool ExistInstance(Guid instanceID)
			{
				Debug.Assert(this.InstanceStore != null);
				lock (CommonUtils.GetLocker(this.InstanceStore))
				{
					return this.InstanceStore.Where(instance => instance.Value.InstanceID.HasValue && instance.Value.InstanceID.Value.Equals(instanceID)).Count() > 0;
				}
			}

			#endregion Public Methods
		}

		#region Disposable

		internal partial class MulticastDnsAdapter : IDisposable
		{
			internal bool _disposed = false;

			internal void Dispose(bool disposing)
			{
				if (!this._disposed)
				{
					// If disposing equals true, dispose all managed and unmanaged resources.
					if (disposing)
					{
						try
						{
							if (this._handler != null)
							{
								this._handler.Dispose();
							}
						}
						catch { }
					}

					// cleanup unmanaged resources
					//

					this._disposed = true;
				}
			}

			/// <summary>
			/// Disposes the object.
			/// </summary>
			public void Dispose()
			{
				this.Dispose(true);

				// This object will be cleaned up by the Dispose method.
				// Therefore, you should call GC.SupressFinalize to take this object off the finalization queue
				// and prevent finalization code for this object from executing a second time.
				GC.SuppressFinalize(this);
			}
		}

		#endregion Disposable
	}
}
