﻿using AirFusion.Common;
using AirFusion.WebService.Adapters;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;

namespace AirFusion.Manager.DeviceManager
{
	/// <summary>
	/// Represents the web service wake up approach.
	/// </summary>
	internal sealed class WebServiceWakeUp : IWakeUp
	{
		#region Private/Protected/Internal

		private Dictionary<IPEndPoint, Task> _wakeUpTask;

		#endregion Private/Protected/Internal

		#region Public

		public WebServiceWakeUp()
		{
			this._wakeUpTask = new Dictionary<IPEndPoint, Task>();
		}

		#endregion Public

		#region Override IWakeUp

		/// <summary>
		/// Wakes up the specified device asynchronously.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <returns></returns>
		/// <exception cref="System.NotSupportedException" />
		public async Task WakeUpAsync(AFUserDevice device)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check prerequisite
				if (device == null)
				{
					throw new ArgumentNullException("device");
				}
				if (device.Owner == null)
				{
					throw new InvalidOperationException("Please set owner first.");
				}
				if (device.InstanceID.CompareTo(Guid.Empty) == 0)
				{
					throw new ArgumentException("The instance ID of the device is unknown.");
				}
				if (device.Address == null || device.Address.Equals(IPAddress.None))
				{
					StaticLogger.LogError("The address is not specified, might due to the device is not found yet.");
					throw new InvalidOperationException("The address is not specified, might due to the device is not found yet.");
				}

				var address = device.Address;
				var port = device.WebServicePort;
				var endPoint = new IPEndPoint(address, port);

				// check if there's any previous connecting task, if yes, wait for its result
				if (this._wakeUpTask.ContainsKey(endPoint))
				{
					var task = this._wakeUpTask[endPoint];
					if (task != null && !task.IsCompleted)
					{
						try
						{
							StaticLogger.LogTrace("There's a previous connecting task running, wait for the result...");
							await task;
							StaticLogger.LogInfo("The device has been waken up successfully by previous task.");
							return;
						}
						catch (WebServiceVersionNotMatchedException ex)
						{
							StaticLogger.LogError("WakeUpAsynce Version'" + this + "': " + ex.Message);
							throw new WebServiceVersionNotMatchedException(new HttpResponseMessage());
						}
						catch (Exception ex)
						{
							throw new Exception("Failed waking up the '" + device + "' by the previous task.", ex);
						}
					}
				}

				StaticLogger.LogTrace("Try to wake up the device by sending wakeup HTTP request to '" + address + ":" + port + "'...");

				try
				{
					Debug.Assert(device.Owner.ConnectionHelperService != null);
					var task = device.Owner.ConnectionHelperService.WakeUpAsync(address, port);
					this._wakeUpTask[endPoint] = task;
					var response = await task.ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("ConnectionHelperService.WakeUpAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("ConnectionHelperService.WakeUpAsync() is done successfully.");
					}

					StaticLogger.LogInfo("The device has been waken up by web service successfully.");
				}
				catch (WebServiceVersionNotMatchedException ex)
				{
					StaticLogger.LogError("WakeUpAsynce Version'" + this + "': " + ex.Message);
					throw ex;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("ConnectionHelperService.WakeUpAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					throw new Exception("Failed waking up the device '" + device + "' by web service through '" + address + ":" + port + "'.", ex);
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Override IWakeUp
	}
}