﻿using AirFusion.Common;
using AirFusion.Common.RegistryHolders;
using AirFusion.WebService.Adapters;
using System;
using System.Diagnostics;
using System.Text;

namespace AirFusion.Manager.DeviceManager
{
	public sealed partial class AFUserDevice
	{
		/// <summary>
		/// Represents the display information of the device.
		/// </summary>
		public sealed partial class DisplayInfo
		{
			/// <summary>
			/// Gets the rotation of the display.
			/// </summary>
			[RegistryStoreTypeAttribute(StoreType.Value)]
			public DisplayRotation? Rotation { get; internal set; }

			/// <summary>
			/// Gets the width of the display.
			/// </summary>
			[RegistryStoreTypeAttribute(StoreType.Value)]
			public int? Width { get; internal set; }

			/// <summary>
			/// Gets the height of the display.
			/// </summary>
			[RegistryStoreTypeAttribute(StoreType.Value)]
			public int? Height { get; internal set; }

			/// <summary>
			/// Gets a value that indicates if the navigation bar is displayed virtually.
			/// </summary>
			[RegistryStoreTypeAttribute(StoreType.Value)]
			public bool? Navigation { get; internal set; }

			/// <summary>
			/// Gets a value that indicates if the screen is ON or not.
			/// </summary>
			[RegistryStoreTypeAttribute(StoreType.Value)]
			public bool? ScreenOn { get; internal set; }

			/// <summary>
			/// Gets a value that indicates if the AutoRotation is ON or not.
			/// </summary>
			[RegistryStoreTypeAttribute(StoreType.Value)]
			public bool? AutoRotation { get; internal set; }

			#region Private/Protected/Internal

			/// <summary>
			/// Initializes a new instance of the <see cref="DisplayInfo"/> class.
			/// </summary>
			internal DisplayInfo()
			{
				this.Rotation = null;
				this.Width = null;
				this.Height = null;
				this.Navigation = null;
				this.ScreenOn = null;
				this.AutoRotation = null;
			}

			/// <summary>
			/// Fills from the specified <see cref="DisplayInfo"/> object if such values are specified.
			/// </summary>
			/// <param name="source">The source object.</param>
			internal void FillFromIfSpecified(DisplayInfo source)
			{
				Debug.Assert(source != null);

				if (source.Rotation.HasValue)
				{
					this.Rotation = source.Rotation.Value;
				}
				if (source.Width.HasValue)
				{
					this.Width = source.Width;
				}
				if (source.Height.HasValue)
				{
					this.Height = source.Height;
				}
				if (source.Navigation.HasValue)
				{
					this.Navigation = source.Navigation;
				}
				if (source.ScreenOn.HasValue)
				{
					this.ScreenOn = source.ScreenOn;
				}
				if (source.AutoRotation.HasValue)
				{
					this.AutoRotation = source.AutoRotation;
				}
			}

			/// <summary>
			/// Fills from the specified <see cref="DeviceInformation.DisplayInfo"/> object if such values are specified.
			/// </summary>
			/// <param name="source">The source object.</param>
			internal void FillFromIfSpecified(DeviceInformation.DisplayInfo source)
			{
				Debug.Assert(source != null);

				if (source.Rotation.HasValue)
				{
					this.Rotation = source.Rotation.Value.Transform();
				}
				if (source.Width.HasValue)
				{
					this.Width = source.Width;
				}
				if (source.Height.HasValue)
				{
					this.Height = source.Height;
				}
				if (source.Navigation.HasValue)
				{
					this.Navigation = source.Navigation;
				}
				if (source.ScreenOn.HasValue)
				{
					this.ScreenOn = source.ScreenOn;
				}
				if (source.AutoRotation.HasValue)
				{
					this.AutoRotation = source.AutoRotation;
				}
			}

			/// <summary>
			/// Fills from the specified <see cref="DisplayConfigurationUpdateAdapter.DisplayConfiguration"/> object if such values are specified.
			/// </summary>
			/// <param name="source">The source object.</param>
			internal void FillFromIfSpecified(DisplayConfigurationUpdateAdapter.DisplayConfiguration source)
			{
				Debug.Assert(source != null);

				if (source.Rotation.HasValue)
				{
					this.Rotation = source.Rotation.Value.Transform();
				}
				if (source.Width.HasValue)
				{
					this.Width = source.Width;
				}
				if (source.Height.HasValue)
				{
					this.Height = source.Height;
				}
				if (source.Navigation.HasValue)
				{
					this.Navigation = source.Navigation;
				}
				if (source.ScreenOn.HasValue)
				{
					this.ScreenOn = source.ScreenOn;
				}
				if (source.AutoRotation.HasValue)
				{
					this.AutoRotation = source.AutoRotation;
				}
			}

			/// <summary>
			/// Fills the values into the specified <see cref="DeviceInformation.DisplayInfo"/> object.
			/// </summary>
			/// <param name="target">The target object.</param>
			internal void FillTo(DeviceInformation.DisplayInfo target)
			{
				Debug.Assert(target != null);

				if (this.Rotation.HasValue)
				{
					target.Rotation = this.Rotation.Value.Transform();
				}
				else
				{
					target.Rotation = null;
				}
				target.Width = this.Width;
				target.Height = this.Height;
				target.Navigation = this.Navigation;
				target.ScreenOn = this.ScreenOn;
				target.AutoRotation = this.AutoRotation;
			}

			/// <summary>
			/// Fills the values into the specified <see cref="DisplayConfigurationUpdateAdapter.DisplayConfiguration"/>  object.
			/// </summary>
			/// <param name="target">The target object.</param>
			internal void FillTo(DisplayConfigurationUpdateAdapter.DisplayConfiguration target)
			{
				Debug.Assert(target != null);

				if (this.Rotation.HasValue)
				{
					target.Rotation = this.Rotation.Value.Transform();
				}
				else
				{
					target.Rotation = null;
				}

				target.Width = this.Width;
				target.Height = this.Height;
				target.Navigation = this.Navigation;
				target.ScreenOn = this.ScreenOn;
				target.AutoRotation = this.AutoRotation;
			}

			#endregion Private/Protected/Internal

			#region Override

			/// <summary>
			/// Determines whether the specified object is equal to the current object.
			/// </summary>
			/// <param name="obj">The object to compare with the current object.</param>
			/// <returns></returns>
			public override bool Equals(object obj)
			{
				return (obj is DisplayInfo && this == (DisplayInfo)obj);
			}

			/// <summary>
			/// Returns the hash code of the instance.
			/// </summary>
			/// <returns></returns>
			public override int GetHashCode()
			{
				return base.GetHashCode() ^
					(this.Rotation.HasValue ? this.Rotation.GetHashCode() : 0) ^
					(this.Width.HasValue && this.Height.HasValue ? (this.Width << 16 | this.Height).GetHashCode() : 0) ^
					(this.Navigation.HasValue ? this.Navigation.GetHashCode() : 0) ^
					(this.ScreenOn.HasValue ? this.ScreenOn.GetHashCode() : 0);
			}

			/// <summary>
			/// Implements the operator ==.
			/// </summary>
			/// <param name="x">The first object.</param>
			/// <param name="y">The second object.</param>
			public static bool operator ==(DisplayInfo x, DisplayInfo y)
			{
				return (((object)x == null && (object)y == null) || Object.ReferenceEquals(x, y) ||
					((object)x != null && (object)y != null) &&
					x.Rotation == y.Rotation &&
					x.Width == y.Width &&
					x.Height == y.Height &&
					x.Navigation == y.Navigation &&
					x.ScreenOn == y.ScreenOn);
			}

			/// <summary>
			/// Implements the operator !=.
			/// </summary>
			/// <param name="x">The first object.</param>
			/// <param name="y">The second object.</param>
			public static bool operator !=(DisplayInfo x, DisplayInfo y)
			{
				return !(x == y);
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".Rotation: " + (this.Rotation.HasValue ? this.Rotation.Value.GetStringValue() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Width: " + (this.Width.HasValue ? this.Width.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Height: " + (this.Height.HasValue ? this.Height.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Navigation: " + (this.Navigation.HasValue ? this.Navigation.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".ScreenOn: " + (this.ScreenOn.HasValue ? this.ScreenOn.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".AutoRotation: " + (this.AutoRotation.HasValue ? this.AutoRotation.Value.ToString() : "N/A"));

				return builder.ToString().TrimEnd();
			}

			#endregion Override
		}
	}

	#region Extensions

	/// <summary>
	/// Extensions.
	/// </summary>
	internal static partial class AFExtensions
	{
		internal static AFUserDevice.DisplayInfo.DisplayRotation Transform(this AirFusion.WebService.Adapters.DisplayRotation rotation)
		{
			switch (rotation)
			{
				case AirFusion.WebService.Adapters.DisplayRotation.Rotation0:
					return AFUserDevice.DisplayInfo.DisplayRotation.Rotation0;

				case AirFusion.WebService.Adapters.DisplayRotation.Rotation90:
					return AFUserDevice.DisplayInfo.DisplayRotation.Rotation90;

				case AirFusion.WebService.Adapters.DisplayRotation.Rotation180:
					return AFUserDevice.DisplayInfo.DisplayRotation.Rotation180;

				case AirFusion.WebService.Adapters.DisplayRotation.Rotation270:
					return AFUserDevice.DisplayInfo.DisplayRotation.Rotation270;
			}

			throw new ArgumentException("Inconvertible value '" + rotation.ToString() + "'.");
		}

		internal static AirFusion.WebService.Adapters.DisplayRotation Transform(this AFUserDevice.DisplayInfo.DisplayRotation rotation)
		{
			switch (rotation)
			{
				case AFUserDevice.DisplayInfo.DisplayRotation.Rotation0:
					return AirFusion.WebService.Adapters.DisplayRotation.Rotation0;

				case AFUserDevice.DisplayInfo.DisplayRotation.Rotation90:
					return AirFusion.WebService.Adapters.DisplayRotation.Rotation90;

				case AFUserDevice.DisplayInfo.DisplayRotation.Rotation180:
					return AirFusion.WebService.Adapters.DisplayRotation.Rotation180;

				case AFUserDevice.DisplayInfo.DisplayRotation.Rotation270:
					return AirFusion.WebService.Adapters.DisplayRotation.Rotation270;
			}

			throw new ArgumentException("Inconvertible value '" + rotation.ToString() + "'.");
		}
	}

	#endregion Extensions
}
