﻿using AirFusion.Common;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of pairing invitation (POST /pairing/invitation).
	/// </summary>
	internal sealed class PairingInvitationAdapter : WebServiceAdapter<PairingInvitationAdapter.PairingInvitationInfo, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/pairing/invitation";
		internal static HttpMethod WebServiceMethod = HttpMethod.Post;

		#endregion Definitions

		#region Classes

		/// <summary>
		/// Represents the invitation data of pairing invitation.
		/// </summary>
		public class PairingInvitationInfo
		{
			/// <summary>
			/// Gets the instance ID.
			/// </summary>
			[JsonProperty("InstanceID"), JsonConverter(typeof(AirFusion.Common.JsonConverters.GuidConverter))]
			public Guid? InstanceID { get; internal set; }

			/// <summary>
			/// Gets the device name.
			/// </summary>
			[JsonProperty("DeviceName")]
			public string DeviceName { get; internal set; }

			/// <summary>
			/// Gets the model name.
			/// </summary>
			[JsonProperty("ModelName")]
			public string ModelName { get; internal set; }

			/// <summary>
			/// Gets the user name.
			/// </summary>
			[JsonProperty("UserName")]
			public string UserName { get; internal set; }

			/// <summary>
			/// Gets the communication port.
			/// </summary>
			[JsonProperty("CommunicationPort")]
			public ushort? CommunicationPort { get; internal set; }

			/// <summary>
			/// Gets the session key.
			/// </summary>
			[JsonProperty("SessionKey")]
			public string SessionKey { get; internal set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="PairingInvitationInfo"/> class.
			/// </summary>
			internal PairingInvitationInfo()
			{
				this.InstanceID = null;
				this.DeviceName = null;
				this.ModelName = null;
				this.UserName = null;
				this.CommunicationPort = null;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".InstanceID: " + (this.InstanceID.HasValue ? this.InstanceID.Value.GetStringFormat() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".DeviceName: " + this.DeviceName);
				builder.AppendLine(this.GetType().GetFullName() + ".ModelName: " + this.ModelName);
				builder.AppendLine(this.GetType().GetFullName() + ".UserName: " + this.UserName);
				builder.AppendLine(this.GetType().GetFullName() + ".CommunicationPort: " + (this.CommunicationPort.HasValue ? this.CommunicationPort.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".SessionKey: " + this.SessionKey);

				return builder.ToString().TrimEnd();
			}
		}

		#endregion Classes

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="PairingInvitationAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public PairingInvitationAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager, WebServiceUri, WebServiceMethod)
		{
		}

		/// <summary>
		/// Sends a pairing invitation to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="invitationData">The invitation data.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> InviteAsync(IPAddress address, ushort port, PairingInvitationInfo invitationData)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}
			if (invitationData == null)
			{
				throw new ArgumentNullException("invitationData");
			}
			if (invitationData.InstanceID.HasValue == false)
			{
				throw new ArgumentNullException("invitationData.InstanceID", "The instance ID of the specified invitation data is unspecified.");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri).Uri,
				invitationData);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
