﻿using AirFusion.Common;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of display configuration (update) (PUT /configuration/display/{InstanceID}).
	/// </summary>
	internal sealed class DisplayConfigurationUpdateAdapter : WebServiceAdapter<DisplayConfigurationUpdateAdapter.DisplayConfiguration, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/configuration/display";
		internal static HttpMethod WebServiceMethod = HttpMethod.Put;

		#endregion Definitions

		#region Classes

		/// <summary>
		/// Represents the display configuration.
		/// </summary>
		internal sealed class DisplayConfiguration
		{
			/// <summary>
			/// Gets or sets the rotation of the display.
			/// </summary>
			[JsonProperty("Rotation"), JsonConverter(typeof(AirFusion.Common.JsonConverters.EnumConverter))]
			internal DisplayRotation? Rotation { get; set; }

			/// <summary>
			/// Gets or sets the width of the display.
			/// </summary>
			[JsonProperty("Width")]
			internal int? Width { get; set; }

			/// <summary>
			/// Gets or sets the height of the display.
			/// </summary>
			[JsonProperty("Height")]
			internal int? Height { get; set; }

			/// <summary>
			/// Gets or sets a value that indicates if the navigation bar is displayed virtually.
			/// </summary>
			[JsonProperty("Navigation")]
			internal bool? Navigation { get; set; }

			/// <summary>
			/// Gets or sets a value that indicates if the screen is ON.
			/// </summary>
			[JsonProperty("ScreenOn")]
			internal bool? ScreenOn { get; set; }

			/// <summary>
			/// Gets or sets a value that indicates if the AutoRotation is ON.
			/// </summary>
			[JsonProperty("AutoRotation")]
			internal bool? AutoRotation { get; set; }

			/// <summary>
			/// Initializes a new instance of the <see cref="DisplayConfiguration"/> class.
			/// </summary>
			internal DisplayConfiguration()
			{
				this.Rotation = null;
				this.Width = null;
				this.Height = null;
				this.Navigation = null;
				this.ScreenOn = null;
				this.AutoRotation = null;
			}

			/// <summary>
			/// Initializes a new instance of the <see cref="DisplayConfiguration"/> class.
			/// </summary>
			/// <param name="rotation">The rotation of the display.</param>
			/// <param name="width">The width of the display.</param>
			/// <param name="height">The height of the display.</param>
			/// <param name="navigation">A value that indicates if the navigation bar is displayed virtually.</param>
			/// <param name="screenOn">A value that indicates if the screen is ON.</param>
			/// <param name="screenOn">A value that indicates if the autoRotate is ON.</param>
			internal DisplayConfiguration(DisplayRotation? rotation, int? width, int? height, bool? navigation, bool? screenOn, bool? autoRotate)
			{
				this.Rotation = rotation;
				this.Width = width;
				this.Height = height;
				this.Navigation = navigation;
				this.ScreenOn = screenOn;
				this.AutoRotation = autoRotate;
			}

			/// <summary>
			/// Returns a <see cref="System.String" /> that represents this instance.
			/// </summary>
			/// <returns>
			/// A <see cref="System.String" /> that represents this instance.
			/// </returns>
			public override string ToString()
			{
				StringBuilder builder = new StringBuilder();
				builder.AppendLine(this.GetType().GetFullName() + ".Rotation: " + (this.Rotation.HasValue ? this.Rotation.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Width: " + (this.Width.HasValue ? this.Width.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Height: " + (this.Height.HasValue ? this.Height.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".Navigation: " + (this.Navigation.HasValue ? this.Navigation.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".ScreenOn: " + (this.ScreenOn.HasValue ? this.ScreenOn.Value.ToString() : "N/A"));
				builder.AppendLine(this.GetType().GetFullName() + ".AutoRotation: " + (this.AutoRotation.HasValue ? this.AutoRotation.Value.ToString() : "N/A"));

				return builder.ToString().TrimEnd();
			}
		}

		#endregion Classes

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="DisplayConfigurationUpdateAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public DisplayConfigurationUpdateAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager, WebServiceUri, WebServiceMethod)
		{
		}

		/// <summary>
		/// Sends a rotate request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="rotation">The rotation degree.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> RotateAsync(IPAddress address, ushort port, DisplayRotation rotation)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri).Uri,
				new DisplayConfiguration(rotation, null, null, null, null,null));

			return await this.SendRequestAsync(request, Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
