﻿using AirFusion.Common;
using AirFusion.Manager.DeviceManager;
using AirFusion.WebService.Adapters;
using System;
using System.Diagnostics;
using System.Net;
using System.Threading.Tasks;

namespace AirFusion.Manager.FeatureHelper
{
	/// <summary>
	/// The extensions of <see cref="AirFusion.Manager.DeviceManager.AFUserDevice"/>.
	/// </summary>
	public static partial class UserDeviceExtensions
	{
		#region Public Methods

		/// <summary>
		/// Gets the input daemon status of the target device asynchronously.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <returns>A value indicating the status of input daemon, true indicates the input daemon is alive.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NullReferenceException" />
		/// <exception cref="System.TimeoutException" />
		public static async Task<bool> GetInputDaemonStatusAsync(this AFUserDevice device)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("device: " + (device != null ? device.ToString() : "N/A"));

				// check parameters
				if (device == null)
				{
					throw new NullReferenceException();
				}

				// check prerequisite
				if (device.Address == null || device.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(device.Owner != null, "The 'device.Owner' is not specified.");
				Trace.Assert(InputControlHelper.DeviceManagerMapping != null, "The '" + typeof(InputControlHelper).Name + ".DeviceManagerMapping' is not specified.");
				Trace.Assert(InputControlHelper.DeviceManagerMapping.ContainsKey(device.Owner) && InputControlHelper.DeviceManagerMapping[device.Owner] != null, "No '" + typeof(InputControlHelper).Name + "' is associated for this instance.");

				// get input control service
				var inputControlService = InputControlHelper.DeviceManagerMapping[device.Owner].InputControlService;

				// debug assertion
				Trace.Assert(inputControlService != null, "The 'inputControlService' is not specified.");

				try
				{
					var address = device.Address;
					var port = device.WebServicePort;
					StaticLogger.LogTrace("Request input daemon status from '" + address + ":" + port + "'...");

					// send the HTTP request
					var response = await inputControlService.GetStatusAsync(address, port).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("InputControlService.GetStatusAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("InputControlService.GetStatusAsync() is done successfully.");
					}

					if (response.ResponseData.IsAlive.HasValue)
					{
						return response.ResponseData.IsAlive.Value;
					}
					else
					{
						throw new Exception("Server returned an invalid status.");
					}
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("InputControlService.GetStatusAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed getting status of input daemon from device '" + device + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Public Methods
	}
}
