﻿using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of the input daemon status (request) (GET /input/status).
	/// </summary>
	internal sealed class InputDaemonStatusAdapter : WebServiceAdapter<DBNull, InputDaemonStatusAdapter.InputDaemonStatus>
	{
		#region Definitions

		internal const string WebServiceUri = "/input/status";
		internal static HttpMethod WebServiceMethod = HttpMethod.Get;

		#endregion Definitions

		#region Classes

		/// <summary>
		/// Represents the status of input daemon.
		/// </summary>
		internal sealed class InputDaemonStatus
		{
			/// <summary>
			/// Gets a valud indicating if the input daemon is alive.
			/// </summary>
			[JsonProperty("Alive")]
			public bool? IsAlive { get; set; }
		}

		#endregion Classes

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="InputDaemonStatusAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public InputDaemonStatusAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager)
		{
		}

		/// <summary>
		/// Gets the input daemon status from target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> GetStatusAsync(IPAddress address, ushort port)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri).Uri);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
