﻿using System;
using System.Diagnostics;
using System.Net.Http;

namespace AirFusion.WebService
{
	internal sealed class VersionLocalRequestFilter : WebServiceLocalRequestFilter
	{
		#region Definitions

		private const string ApplicationVersionKey = "Application-Version";
		private const string MinimumTargetVersionKey = "Min-Target-Version";

		#endregion Definitions

		/// <summary>
		/// Gets or sets the version of application.
		/// </summary>
		public string ApplicationVersion { get; internal set; }

		/// <summary>
		/// Gets or sets the minimum target version.
		/// </summary>
		public string MinimumTargetVersion { get; internal set; }

		/// <summary>
		/// Initializes a new instance of the <see cref="VersionLocalRequestFilter"/> class.
		/// </summary>
		/// <param name="applicationVersion">The application version.</param>
		/// <param name="minimumTargetVersion">The minimum target version.</param>
		public VersionLocalRequestFilter(string applicationVersion, string minimumTargetVersion)
		{
			this.ApplicationVersion = applicationVersion;
			this.MinimumTargetVersion = minimumTargetVersion;
		}

		/// <summary>
		/// Filters a local request.
		/// </summary>
		/// <param name="requestMessage">The request message, any modification could be made to this instance.</param>
		/// <param name="responseMessage">The response message if any.</param>
		/// <returns>True if to continue next filters.</returns>
		/// <remarks>
		/// - The exception thrown by this method should be considered as an invalid request to be propagated.
		/// - The output 'responseMessage' is always null.
		/// </remarks>
		public override bool Filter(HttpRequestMessage requestMessage, out HttpResponseMessage responseMessage)
		{
			// check parameters
			if (requestMessage == null)
			{
				throw new ArgumentNullException("requestMessage");
			}

			// default output value
			responseMessage = null;

			Trace.Assert(requestMessage.Headers != null, "requestMessage.Headers = null");

			// fill Application-Version
			if (this.ApplicationVersion != null)
			{
				requestMessage.Headers.Add(ApplicationVersionKey, this.ApplicationVersion.ToString());
			}

			// fill Min-Target-Version
			if (this.MinimumTargetVersion != null)
			{
				requestMessage.Headers.Add(MinimumTargetVersionKey, this.MinimumTargetVersion.ToString());
			}

			return true;
		}
	}
}
