﻿using AirFusion.Common;
using AirFusion.Manager.DeviceManager;
using AirFusion.WebService.Adapters;
using System;
using System.Diagnostics;
using System.Net;
using System.Threading.Tasks;

namespace AirFusion.Manager.FeatureHelper
{
	/// <summary>
	/// The extensions of <see cref="AirFusion.Manager.DeviceManager.AFUserDevice"/>.
	/// </summary>
	public static partial class UserDeviceExtensions
	{
		#region Public Methods

		/// <summary>
		/// Sends screencast request to the device asynchronously.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NullReferenceException" />
		/// <exception cref="System.TimeoutException" />
		[Obsolete("This method is useless for now.")]
		public static async Task ScreencastRequestAsync(this AFUserDevice device)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("device: " + (device != null ? device.ToString() : "N/A"));

				// check parameters
				if (device == null)
				{
					throw new NullReferenceException();
				}

				// check prerequisite
				if (device.Address == null || device.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(device.Owner != null, "The 'device.Owner' is not specified.");
				Trace.Assert(device.Owner.LocalDevice != null, "The 'device.Owner.LocalDevice' is not specified.");
				Trace.Assert(device.Owner.LocalDevice.InstanceID.Equals(Guid.Empty) == false, "The 'device.Owner.LocalDevice.InstanceID' is Guid.Empty.");
				Trace.Assert(ScreencastHelper.DeviceManagerMapping != null, "The '" + typeof(ScreencastHelper).Name + ".DeviceManagerMapping' is not specified.");
				Trace.Assert(ScreencastHelper.DeviceManagerMapping.ContainsKey(device.Owner) && ScreencastHelper.DeviceManagerMapping[device.Owner] != null, "No '" + typeof(ScreencastHelper).Name + "' is associated for this instance.");

				// get screencast service
				var screencastService = ScreencastHelper.DeviceManagerMapping[device.Owner].ScreencastService;

				// debug assertion
				Trace.Assert(screencastService != null, "The 'screencastService' is not specified.");

				try
				{
					var address = device.Address;
					var port = device.WebServicePort;
					StaticLogger.LogTrace("Send screencast request to '" + address + ":" + port + "'...");

					// send the HTTP request
					var response = await screencastService.RequestAsync(address, port, device.Owner.LocalDevice.InstanceID).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("ScreencastService.RequestAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("ScreencastService.RequestAsync() is done successfully.");
					}
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("ScreencastService.RequestAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed sending screencast request to device '" + device + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends screencast control request to the device asynchronously.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <param name="operation">The screencast operation.</param>
		/// <param name="controlData">The control data.</param>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NullReferenceException" />
		/// <exception cref="System.TimeoutException" />
		public static async Task ControlScreencastAsync(this AFUserDevice device, ScreencastControlInfo.ScreencastOperation operation, ScreencastControlInfo controlData)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (device == null)
				{
					throw new NullReferenceException();
				}

				// log parameters
				StaticLogger.LogDebug("device: " + (device != null ? device.ToString() : "N/A"));
				StaticLogger.LogDebug("operation: " + operation.GetStringValue());
				StaticLogger.LogDebug("controlData: " + (controlData != null ? "\r\n" + controlData.ToString() : "N/A"));

				// check prerequisite
				if (device.Address == null || device.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(device.Owner != null, "The 'device.Owner' is not specified.");
				Trace.Assert(device.Owner.LocalDevice != null, "The 'device.Owner.LocalDevice' is not specified.");
				Trace.Assert(device.Owner.LocalDevice.InstanceID.Equals(Guid.Empty) == false, "The 'device.Owner.LocalDevice.InstanceID' is Guid.Empty.");
				Trace.Assert(ScreencastHelper.DeviceManagerMapping != null, "The '" + typeof(ScreencastHelper).Name + ".DeviceManagerMapping' is not specified.");
				Trace.Assert(ScreencastHelper.DeviceManagerMapping.ContainsKey(device.Owner) && ScreencastHelper.DeviceManagerMapping[device.Owner] != null, "No '" + typeof(ScreencastHelper).Name + "' is associated for this instance.");

				// get screencast service
				var screencastService = ScreencastHelper.DeviceManagerMapping[device.Owner].ScreencastService;

				// debug assertion
				Trace.Assert(screencastService != null, "The 'screencastService' is not specified.");

				// prepare control data
				{
					if (controlData == null)
					{
						controlData = new ScreencastControlInfo();
					}

					controlData.InstanceID = device.Owner.LocalDevice.InstanceID; // pass local device ID
				}

				try
				{
					var address = device.Address;
					var port = device.WebServicePort;
					StaticLogger.LogTrace("Send screencast control request to '" + address + ":" + port + "'...");

					// send the HTTP request
					var response = await screencastService.ControlAsync(address, port, operation, controlData).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("ScreencastService.ControlAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("ScreencastService.ControlAsync() is done successfully.");
					}
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					StaticLogger.LogWarn("ScreencastService.ControlAsync() returns: " + ex.Message);
					throw;
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed sending screencast control request to device '" + device + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Gets the screencast daemon status of the target device asynchronously.
		/// </summary>
		/// <param name="device">The device.</param>
		/// <returns>The status of screencast daemon.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.NullReferenceException" />
		/// <exception cref="System.TimeoutException" />
		public static async Task<DaemonStatus> GetScreencastDaemonStatusAsync(this AFUserDevice device)
		{
			try
			{
				StaticLogger.TrackEnter();

				// check parameters
				if (device == null)
				{
					throw new NullReferenceException();
				}

				// log parameters
				StaticLogger.LogDebug("device: " + (device != null ? device.ToString() : "N/A"));

				// check prerequisite
				if (device.Address == null || device.Address.Equals(IPAddress.None))
				{
					throw new InvalidOperationException("The address is not specified, might due to the device is not found yet.");
				}

				// debug assertion
				Trace.Assert(device.Owner != null, "The 'device.Owner' is not specified.");
				Trace.Assert(ScreencastHelper.DeviceManagerMapping != null, "The '" + typeof(ScreencastHelper).Name + ".DeviceManagerMapping' is not specified.");
				Trace.Assert(ScreencastHelper.DeviceManagerMapping.ContainsKey(device.Owner) && ScreencastHelper.DeviceManagerMapping[device.Owner] != null, "No '" + typeof(ScreencastHelper).Name + "' is associated for this instance.");

				// get screencast service
				var screencastService = ScreencastHelper.DeviceManagerMapping[device.Owner].ScreencastService;

				// debug assertion
				Trace.Assert(screencastService != null, "The 'screencastService' is not specified.");

				try
				{
					var address = device.Address;
					var port = device.WebServicePort;
					StaticLogger.LogTrace("Request screencast daemon status from '" + address + ":" + port + "'...");

					// send the HTTP request
					var response = await screencastService.GetStatusAsync(address, port).ConfigureAwait(continueOnCapturedContext: false);
					if (response.IsMessageResponse)
					{
						StaticLogger.LogInfo("ScreencastService.GetStatusAsync() is done successfully: " + response.Message);
					}
					else
					{
						StaticLogger.LogInfo("ScreencastService.GetStatusAsync() is done successfully.");
					}

					return response.ResponseData.CurrentStatus.HasValue ? response.ResponseData.CurrentStatus.Value : DaemonStatus.Unknown;
				}
				catch (ArgumentException ex)
				{
					Trace.Fail("The exception is an ArgumentException: " + ex.Message);
					throw;
				}
				catch (WebServiceException ex)
				{
					if (ex.StatusCode == HttpStatusCode.InternalServerError) // daemon not running
					{
						StaticLogger.LogWarn("Server returned " + (int)HttpStatusCode.InternalServerError + ", indicating the screencast daemon is not running.");
						return DaemonStatus.NotRunning;
					}
					else
					{
						StaticLogger.LogWarn("ScreencastService.GetStatusAsync() returns: " + ex.Message);
						throw;
					}
				}
				catch (Exception ex)
				{
					StaticLogger.LogError("Failed getting status of screencast daemon from device '" + device + "': " + ex.Message);
					throw;
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Public Methods
	}
}
