﻿using AirFusion.Common;
using AirFusion.Manager.FeatureHelper;
using System;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of wakeup (PUT /screencast/{play|stop|pause|resume}).
	/// </summary>
	internal sealed class ScreencastControlAdapter : WebServiceAdapter<ScreencastControlInfo, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/screencast";
		internal static HttpMethod WebServiceMethod = HttpMethod.Put;

		#endregion Definitions

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="ScreencastControlAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public ScreencastControlAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager)
		{
		}

		/// <summary>
		/// Sends a screencast control to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="operation">The screencast operation.</param>
		/// <param name="controlData">The control data.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> ControlAsync(IPAddress address, ushort port, ScreencastControlInfo.ScreencastOperation operation, ScreencastControlInfo controlData)
		{
			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri.TrimEnd('/') + "/" + operation.GetStringValue()).Uri,
				controlData);

			return await this.SendRequestAsync(request, Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
