﻿using AirFusion.Common;
using System;
using System.Diagnostics;
using System.Net;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Provides the web service manager of incoming call.
	/// </summary>
	internal sealed class IncomingCallService
	{
		#region Delegates

		/// <summary>
		/// Represents the callback function that will handle incoming call post request.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="postData">The request post data.</param>
		/// <returns>The response to the request.</returns>
		public delegate IncomingCallPostAdapter.WebServiceResponse IncomingCallNewPostCallback(IPAddress clientIP, IncomingCallPostAdapter.IncomingCallPostData postData);

		#endregion Delegates

		private IncomingCallPostAdapter _postAdapter;
		private IncomingCallReplyAdapter _replyAdapter;
		private IncomingCallNewPostCallback _incomingCallNewPostCallback;

		#region Methods

		/// <summary>
		/// Initializes the instance.
		/// </summary>
		/// <param name="webServiceManager"></param>
		public IncomingCallService(WebServicesManager webServiceManager)
		{
			// Incoming Call Handler
			this._postAdapter = new IncomingCallPostAdapter(webServiceManager);
			this._postAdapter.RegisterRequestCallback(this.OnPostAdapterRequestReceived);
			this._incomingCallNewPostCallback = null;

			// Incoming Call Reply Handler (Answer or Decline)
			this._replyAdapter = new IncomingCallReplyAdapter(webServiceManager);
		}

		#region Methods to register callbacks (Phone => PC)

		/// <summary>
		/// Registers the callback function of incoming call post request. (Phone => PC)
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterIncomingCallPostRequestCallback(IncomingCallNewPostCallback callback)
		{
			this._incomingCallNewPostCallback = callback;
		}

		#endregion Methods to register callbacks (Phone => PC)

		#region Active Methods (PC => Phone)

		/// <summary>
		/// Sends a answering phone call request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="replyData">The data of answering phone call request.</param>
		/// <returns></returns>
		public async Task<IncomingCallReplyAdapter.WebServiceResponse> AnswerPhoneCallAsync(IPAddress address, ushort port, IncomingCallReplyAdapter.IncomingCallReplyData replyData)
		{
			try
			{
				StaticLogger.TrackEnter();
				Debug.Assert(this._replyAdapter != null);
				return await this._replyAdapter.RequestAnswerPhoneAsync(address, port, replyData).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends a declining phone call request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="replyData">The data of declining phone call request.</param>
		/// <returns></returns>
		public async Task<IncomingCallReplyAdapter.WebServiceResponse> DeclinePhoneCallAsync(IPAddress address, ushort port, IncomingCallReplyAdapter.IncomingCallReplyData replyData)
		{
			try
			{
				StaticLogger.TrackEnter();
				Debug.Assert(this._replyAdapter != null);
				return await this._replyAdapter.RequestHangUpPhoneAsync(address, port, replyData).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Active Methods (PC => Phone)

		#region Callbacks

		/// <summary>
		/// Called when a incoming call post request is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private IncomingCallPostAdapter.WebServiceResponse OnPostAdapterRequestReceived(IPAddress address, ushort port, IncomingCallPostAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._incomingCallNewPostCallback == null)
				{
					return IncomingCallPostAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered event.");
				}
				if (remoteRequest == null)
				{
					return IncomingCallPostAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}
				if (remoteRequest.Request == null)
				{
					return IncomingCallPostAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "There's no request.");
				}

				// check the instance ID
				if (remoteRequest.Request.State == IncomingCallPostAdapter.IncomingCallPostData.NOTIFICATION_VALUE)
				{
					if (!remoteRequest.Request.DeviceInstanceID.HasValue || remoteRequest.Request.DeviceInstanceID.Value.CompareTo(Guid.Empty) == 0)
					{
						StaticLogger.LogWarn("Invalid instance ID.");
						return IncomingCallPostAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is unspecified or invalid.");
					}
				}

				// invoke callback function
				return this._incomingCallNewPostCallback(address, remoteRequest.Request);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks

		#endregion Methods
	}
}
