﻿using AirFusion.Common;
using AirFusion.Common.JsonConverters;
using Newtonsoft.Json;
using System;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Represents the web service of notification remove (PC => phone, Phone => PC)
	/// (DELETE /notification/[DeviceInstanceID]/[UserId]/[PackageName]/[Id]/[Tag])
	/// </summary>
	internal sealed class NotificationRemoveAdapter : WebServiceAdapter<DBNull, DBNull>
	{
		#region Definitions

		internal const string WebServiceUri = "/notification";
		internal static HttpMethod WebServiceMethod = HttpMethod.Delete;

		#endregion Definitions

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="NotificationRemoveAdapter"/> class.
		/// </summary>
		/// <param name="webServiceManager">The web service manager.</param>
		public NotificationRemoveAdapter(WebServicesManager webServiceManager)
			: base(webServiceManager, WebServiceUri, WebServiceMethod)
		{
		}

		/// <summary>
		/// Sends a notification removal request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="extendRemovalRequestUri">
		/// The required removal string to specify which notification to launch.
		/// The request string should follow the format
		/// /notification/[DeviceInstanceID]/[UserId]/[PackageName]/[Id]/[Tag]
		/// </param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<WebServiceResponse> RequestRemoveAsync(IPAddress address, ushort port, string extendRemovalRequestUri)
		{
			// Seoker comment: the request URI should be composed in this object, the responsibility doesn't belong to upper-layer.

			// check parameters
			if (address == null)
			{
				throw new ArgumentNullException("address");
			}

			var request = WebServiceRequest.CreateLocalRequest(
				WebServiceMethod,
				new UriBuilder(this.Scheme, address.ToString(), (int)port, WebServiceUri.TrimEnd('/') + extendRemovalRequestUri).Uri);

			return await this.SendRequestAsync(request, this.Timeout).ConfigureAwait(continueOnCapturedContext: false);
		}

		#endregion Methods
	}
}
