﻿using AirFusion.Common;
using System;
using System.Diagnostics;
using System.Net;
using System.Threading.Tasks;

namespace AirFusion.WebService.Adapters
{
	/// <summary>
	/// Provides the web service manager of notification
	/// </summary>
	internal sealed class NotificationService
	{
		#region Delegates

		/// <summary>
		/// Represents the callback function that will handle notification post request.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="postData">The request post data.</param>
		/// <returns>The response to the request.</returns>
		public delegate NotificationPostAdapter.WebServiceResponse NotificationNewPostCallback(IPAddress clientIP, NotificationPostAdapter.NotificationPostInfo postData);

		/// <summary>
		/// Represents the callback function that will handle notification removal request.
		/// </summary>
		/// <param name="clientIP">The IP address of the client.</param>
		/// <param name="requestUri">The request remove string.</param>
		/// <returns>The response to the request.</returns>
		public delegate NotificationRemoveAdapter.WebServiceResponse NotificationRemovalCallback(IPAddress clientIP, string requestUri);

		#endregion Delegates

		#region Private/Protected/Internal Variables

		internal NotificationPostAdapter _postAdapter;
		internal NotificationLaunchAdapter _launchAdapter;
		internal NotificationRemoveAdapter _removalAdapter;
		internal NotificationIconAdapter _iconAdapter;
		internal NotificationSyncAdapter _syncAdapter;
		internal NotificationIgnoreAdapter _ignoreAdapter;
		internal NotificationQuickActionAdapter _qaAdapter;

		internal int IgnoreTimeoutValue = 4000;

		#endregion Private/Protected/Internal Variables

		#region Events Definition

		private NotificationNewPostCallback _notificationNewPostCallback;
		private NotificationRemovalCallback _notificationRemovalCallback;

		#endregion Events Definition

		#region Methods

		/// <summary>
		/// Initializes the instance.
		/// </summary>
		/// <param name="webServiceManager"></param>
		public NotificationService(WebServicesManager webServiceManager)
		{
			// Post Handler
			this._postAdapter = new NotificationPostAdapter(webServiceManager);
			this._postAdapter.RegisterRequestCallback(this.OnPostAdapterRequestReceived);
			this._notificationNewPostCallback = null;

			// Remove Handler
			this._removalAdapter = new NotificationRemoveAdapter(webServiceManager);
			this._removalAdapter.RegisterRequestCallback(this.OnRemovalAdapterRequestReceived);
			this._notificationRemovalCallback = null;

			// Launch Handler
			this._launchAdapter = new NotificationLaunchAdapter(webServiceManager);

			// Icon Request Handler
			this._iconAdapter = new NotificationIconAdapter(webServiceManager);

			// Sync Handler
			this._syncAdapter = new NotificationSyncAdapter(webServiceManager);

			// Ignore Handler
			this._ignoreAdapter = new NotificationIgnoreAdapter(webServiceManager, TimeSpan.FromMilliseconds(this.IgnoreTimeoutValue));

			// Quick Action Handler
			this._qaAdapter = new NotificationQuickActionAdapter(webServiceManager);
		}

		#region Methods to register callbacks (Phone => PC)

		/// <summary>
		/// Registers the callback function of notification post request. (Phone => PC)
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterNotificationPostRequestCallback(NotificationNewPostCallback callback)
		{
			this._notificationNewPostCallback = callback;
		}

		/// <summary>
		/// Registers the callback function of notification removal request. (Phone => PC)
		/// </summary>
		/// <param name="callback">The callback function.</param>
		public void RegisterNotificationRemovalRequestCallback(NotificationRemovalCallback callback)
		{
			this._notificationRemovalCallback = callback;
		}

		#endregion Methods to register callbacks (Phone => PC)

		#region Active Methods (PC => Phone)

		/// <summary>
		/// Sends a notification launch request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="launchData">The data of the notification launch request.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<NotificationLaunchAdapter.WebServiceResponse> LaunchAsync(IPAddress address, ushort port, NotificationLaunchAdapter.NotificationLaunchInfo launchData)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._launchAdapter != null);
				return await this._launchAdapter.RequestLaunchAsync(address, port, launchData).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends a notification remove request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="removeRequestUri">
		/// The data of the notification remove request string.
		/// The remove string should follow the format
		/// /notification/[DeviceInstanceID]/[UserId]/[PackageName]/[Id]/[Tag]
		/// </param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<NotificationRemoveAdapter.WebServiceResponse> RemoveAsync(IPAddress address, ushort port, string removeRequestUri)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._removalAdapter != null);
				return await this._removalAdapter.RequestRemoveAsync(address, port, removeRequestUri).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends a notification image request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="iconId">The icon ID.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<NotificationIconAdapter.WebServiceResponse> ImageQueryAsync(IPAddress address, ushort port, string iconId)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._iconAdapter != null);
				Debug.Assert(!string.IsNullOrEmpty(iconId));

				string imageRequestUri = "/" + iconId + "/json"; // or /png
				return await this._iconAdapter.RequestImageAsync(address, port, imageRequestUri).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends a notification sync request to the target device asynchronously.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<NotificationSyncAdapter.WebServiceResponse> SyncNotificationAsync(IPAddress address, ushort port)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._syncAdapter != null);
				return await this._syncAdapter.RequestSyncAsync(address, port).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends a notification ignore request to the target device asynchronouslly.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="packageName">The package name.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<NotificationIgnoreAdapter.WebServiceResponse> ConfigureIgnoreAsync(IPAddress address, ushort port, string packageName)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._ignoreAdapter != null);
				return await this._ignoreAdapter.RequestIgnoreAsync(address, port, packageName).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Sends a notification quick action launch request to the target device asynchronouslly.
		/// </summary>
		/// <param name="address">The IP address of the target device.</param>
		/// <param name="port">The port number of the target device.</param>
		/// <param name="qaData">The request data of the notification quick action request.</param>
		/// <returns>The response.</returns>
		/// <exception cref="AirFusion.WebService.Adapters.WebServiceException" />
		/// <exception cref="System.ArgumentNullException" />
		/// <exception cref="System.Exception" />
		/// <exception cref="System.InvalidOperationException" />
		/// <exception cref="System.TimeoutException" />
		public async Task<NotificationQuickActionAdapter.WebServiceResponse> QuickActionLaunchAsync(IPAddress address, ushort port, NotificationQuickActionAdapter.NotificationQuickActionInfo qaData)
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._qaAdapter != null);
				return await this._qaAdapter.RequestQuickActionAsync(address, port, qaData).ConfigureAwait(continueOnCapturedContext: false);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Active Methods (PC => Phone)

		#endregion Methods

		#region Callbacks

		/// <summary>
		/// Called when a notification post request is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private NotificationPostAdapter.WebServiceResponse OnPostAdapterRequestReceived(IPAddress address, ushort port, NotificationPostAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._notificationNewPostCallback == null)
				{
					return NotificationPostAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered event.");
				}
				if (remoteRequest == null)
				{
					return NotificationPostAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}
				if (remoteRequest.Request == null)
				{
					return NotificationPostAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "There's no request.");
				}

				// check the instance ID
				if (!remoteRequest.Request.DeviceInstanceID.HasValue || remoteRequest.Request.DeviceInstanceID.Value.CompareTo(Guid.Empty) == 0)
				{
					StaticLogger.LogWarn("Invalid instance ID.");
					return NotificationPostAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "The instance ID is unspecified or invalid.");
				}

				// invoke callback function
				return this._notificationNewPostCallback(address, remoteRequest.Request);
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Called when a notification removal request is received.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">The source port.</param>
		/// <param name="remoteRequest">The remote request.</param>
		/// <returns>The response to the request.</returns>
		private NotificationRemoveAdapter.WebServiceResponse OnRemovalAdapterRequestReceived(IPAddress address, ushort port, NotificationRemoveAdapter.WebServiceRequest remoteRequest)
		{
			try
			{
				StaticLogger.TrackEnter();

				// log parameters
				StaticLogger.LogDebug("address: " + (address != null ? address.ToString() : "N/A"));
				StaticLogger.LogDebug("port: " + port);
				StaticLogger.LogDebug("remoteRequest: " + (remoteRequest != null ? "\r\n" + remoteRequest.ToString() : "N/A"));

				Trace.Assert(address != null, "address = null");
				Trace.Assert(remoteRequest != null, "remoteRequest = null");

				// check prerequisite
				if (this._notificationRemovalCallback == null)
				{
					return NotificationRemoveAdapter.WebServiceResponse.Create(HttpStatusCode.NotImplemented, "There's no registered event.");
				}
				if (remoteRequest == null)
				{
					return NotificationRemoveAdapter.WebServiceResponse.Create(HttpStatusCode.InternalServerError, "There's no remote request.");
				}
				if (remoteRequest.RequestUri == null)
				{
					return NotificationRemoveAdapter.WebServiceResponse.Create(HttpStatusCode.BadRequest, "There's no request.");
				}

				// invoke callback function
				return this._notificationRemovalCallback(address, remoteRequest.RequestUri.ToString());
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Callbacks
	}
}
