﻿using AirFusion.Common;
using System;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AirFusion.WebService
{
	// TODO:
	// Rewrite this with TcpClient pattern: Start(), AcceptRequest(), BeginAccepRequest() & EndAcceptRequest(), Stop()
	// And WebServiceManager leverage this object rather than inheriting from it.
	/// <summary>
	/// Represents the basic HTTP listener.
	/// </summary>
	public abstract class HttpListener
	{
		#region Properties

		/// <summary>
		/// Gets or sets the local address of the HTTP listener.
		/// </summary>
		public IPAddress Address { get; protected set; }

		/// <summary>
		/// Gets or sets the port number of the HTTP listener.
		/// </summary>
		public ushort Port { get; protected set; }

		#endregion Properties

		#region Private/Protected/Internal Variables

		internal TcpListener _tcpListener;
		internal Thread _listenThread;

		#endregion Private/Protected/Internal Variables

		#region Methods

		/// <summary>
		/// Initializes a new instance of the <see cref="HttpListener"/> class.
		/// </summary>
		public HttpListener()
		{
			this.Address = IPAddress.Any;
			this.Port = 0;
			this._listenThread = null;
		}

		/// <summary>
		/// Starts listening.
		/// </summary>
		/// <exception cref="System.Net.Sockets.SocketException" />
		public virtual void StartListen()
		{
			this.StartListen(IPAddress.Any, 0);
		}

		/// <summary>
		/// Starts listening.
		/// </summary>
		/// <param name="port">The port number to serve the HTTP listener.</param>
		/// <exception cref="System.Net.Sockets.SocketException" />
		public virtual void StartListen(ushort port)
		{
			this.StartListen(IPAddress.Any, port);
		}

		/// <summary>
		/// Starts listening.
		/// </summary>
		/// <param name="address">The address to serve the HTTP listener.</param>
		/// <param name="port">The port number to serve the HTTP listener.</param>
		/// <exception cref="System.Net.Sockets.SocketException" />
		public virtual void StartListen(IPAddress address, ushort port)
		{
			try
			{
				StaticLogger.TrackEnter();

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, "Listen")))
				{
					this.StopListen();

					this._tcpListener = new TcpListener(address, port);
					this._tcpListener.Start();
					this.Address = address;
					this.Port = (ushort)(this._tcpListener.LocalEndpoint as IPEndPoint).Port;

					this.BeginAcceptNextClient();
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Stops listening.
		/// </summary>
		public virtual void StopListen()
		{
			try
			{
				StaticLogger.TrackEnter();

				lock (CommonUtils.GetLocker(new Tuple<object, string>(this, "Listen")))
				{
					if (this._tcpListener != null)
					{
						try
						{
							this._tcpListener.Stop();
						}
						catch { }
					}
				}
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Returns a <see cref="System.String" /> that represents this instance.
		/// </summary>
		/// <returns>
		/// A <see cref="System.String" /> that represents this instance.
		/// </returns>
		public override string ToString()
		{
			StringBuilder builder = new StringBuilder();
			builder.AppendLine(this.GetType().GetFullName() + ".Address: " + (this.Address != null ? this.Address.ToString() : "N/A"));
			builder.AppendLine(this.GetType().GetFullName() + ".Port: " + this.Port);

			return builder.ToString().TrimEnd();
		}

		/// <summary>
		/// Reads the content asynchronously.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">Thr source port.</param>
		/// <param name="requestMessage">The request message.</param>
		/// <returns>The content read.</returns>
		/// <remarks>Please override this method, otherwise no content will be read.</remarks>
		protected virtual async Task<HttpContent> ReadContentAsync(IPAddress address, ushort port, HttpRemoteRequestMessage requestMessage)
		{
			try
			{
				StaticLogger.TrackEnter();

				// do nothing
				await Task.Yield();
				return null;
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Responses the client.
		/// </summary>
		/// <param name="address">The source address.</param>
		/// <param name="port">Thr source port.</param>
		/// <param name="requestMessage">The request message.</param>
		/// <returns>The response message.</returns>
		protected abstract HttpResponseMessage ResponseClient(IPAddress address, ushort port, HttpRemoteRequestMessage requestMessage);

		/// <summary>
		/// Starts the listening thread.
		/// </summary>
		/// <exception cref="System.Threading.ThreadStateException" />
		/// <exception cref="System.OutOfMemoryException" />
		protected virtual void BeginAcceptNextClient()
		{
			try
			{
				StaticLogger.TrackEnter();

				Debug.Assert(this._tcpListener != null);
				this._tcpListener.BeginAcceptTcpClient(delegate(IAsyncResult result)
				{
					var listener = result.AsyncState as TcpListener;
					Trace.Assert(listener != null, "listener is null.");
					try
					{
						var client = listener.EndAcceptTcpClient(result);
						Trace.Assert(client != null, "client is null.");
						if (client != null && client.Connected)
						{
							this.StartClientTask(client);
						}
					}
					catch (SocketException ex)
					{
						StaticLogger.LogWarn("Failed accepting the client:\r\nSocker error code: " + ex.SocketErrorCode + "\r\nNative error code: " + ex.NativeErrorCode + "\r\nError code: " + ex.ErrorCode + "\r\nMessage: " + ex.Message);
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed accepting the client, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					}
					finally
					{
						this.BeginAcceptNextClient();
					}
				}, this._tcpListener);
			}
			catch (SocketException ex)
			{
				StaticLogger.LogWarn("Failed beginning to accept next client:\r\nSocker error code: " + ex.SocketErrorCode + "\r\nNative error code: " + ex.NativeErrorCode + "\r\nError code: " + ex.ErrorCode + "\r\nMessage: " + ex.Message);
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed beginning to accept next client, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		/// <summary>
		/// Starts the client thread.
		/// </summary>
		/// <param name="client">The client.</param>
		protected virtual void StartClientTask(TcpClient client)
		{
			try
			{
				StaticLogger.TrackEnter();

				var endPoint = client.Client.RemoteEndPoint as IPEndPoint;
				StaticLogger.LogInfo("A client '" + (endPoint != null ? endPoint.ToString() : "(unknown endpoint)") + "' has been connected.");

				int workerThreads, completionPortThreads;
				ThreadPool.GetMaxThreads(out workerThreads, out completionPortThreads);
				StaticLogger.LogDebug("The maximum number of worker threads in the thread pool: " + workerThreads);
				StaticLogger.LogDebug("The maximum number of asynchronous I/O threads in the thread pool: " + completionPortThreads);
				ThreadPool.GetAvailableThreads(out workerThreads, out completionPortThreads);
				StaticLogger.LogDebug("The number of available worker threads: " + workerThreads);
				StaticLogger.LogDebug("The number of available asynchronous I/O threads: " + completionPortThreads);

				// create a threadpool thread to process the client
				ThreadPool.QueueUserWorkItem(async delegate(object state)
				{
					var threadName = Thread.CurrentThread.Name;
					try
					{
						StaticLogger.LogTrace("A thread has been created to process " + endPoint);
						//Thread.CurrentThread.Name = "HTTP client (" + endPoint + ")";

						var handler = new HttpRemoteRequestHandler(client);

						// read request
						{
							bool failedReadRequest = false;
							try
							{
								await handler.ReadRequestAsync();
							}
							catch (Exception ex)
							{
								StaticLogger.LogError("Failed reading request, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
								failedReadRequest = true;
							}
							if (failedReadRequest)
							{
								await handler.WriteResponseAsync(new HttpResponseMessage(HttpStatusCode.BadRequest) { Content = new StringContent("Failed reading request.") });
								return;
							}
						}

						// read content
						if (handler.RequestMessage.HasContent)
						{
							bool failedReadContent = false;
							try
							{
								await this.ReadContentAsync(endPoint.Address, (ushort)endPoint.Port, handler.RequestMessage);
							}
							catch (Exception ex)
							{
								StaticLogger.LogError("Failed reading content, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
								failedReadContent = true;
							}
							if (failedReadContent)
							{
								await handler.WriteResponseAsync(new HttpResponseMessage(HttpStatusCode.BadRequest) { Content = new StringContent("Failed reading content.") });
								return;
							}
						}

						// write response
						{
							bool failedWrite = false;
							try
							{
								var response = this.ResponseClient(endPoint.Address, (ushort)endPoint.Port, handler.RequestMessage);
								await handler.WriteResponseAsync(response);
							}
							catch (Exception ex)
							{
								StaticLogger.LogError("Failed responding the client, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
								failedWrite = true;
							}
							if (failedWrite)
							{
								await handler.WriteResponseAsync(new HttpResponseMessage(HttpStatusCode.InternalServerError) { Content = new StringContent("Failed writing response.") });
								return;
							}
						}
					}
					catch (Exception ex)
					{
						StaticLogger.LogError("Failed handling the client, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
					}
					finally
					{
						try
						{
							StaticLogger.LogInfo("Close the client '" + (endPoint != null ? endPoint.ToString() : "(Unknown endpoint)") + "'.");
							client.Close();
						}
						catch { }

						StaticLogger.LogTrace("The thread has finished processing '" + (endPoint != null ? endPoint.ToString() : "(Unknown endpoint)") + "'.");
						Thread.CurrentThread.Name = threadName;
					}
				});
			}
			catch (Exception ex)
			{
				StaticLogger.LogError("Failed handling the client, exception:\r\n" + CommonUtils.GetExceptionMessages(ex));
			}
			finally
			{
				StaticLogger.TrackLeave();
			}
		}

		#endregion Methods
	}
}
