﻿using AirFusion.Common;
using AirFusion.Manager.Notification;
using Microsoft.WindowsAPICodePack.Shell.PropertySystem;
using MS.WindowsAPICodePack.Internal;
using ShortcutHelpers;
using System;
using System.Drawing;
using System.IO;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using Windows.Data.Xml.Dom;
using Windows.UI.Notifications;

namespace AirFusion.Toast
{
    public partial class AFToastManager
    {
        #region event handler definition
        /// <summary>
        /// Represents the method that will be invoked when a Toast is activated
        /// </summary>
        /// <param name="sender">The sender</param>
        /// <param name="args">The event arguments</param>
        public delegate void AFToastActivatedEventHandler(object sender, AFToastActionActivateEventArgs args);

        /// <summary>
        /// Represents the method that will be invoked when a Toast is dismissed
        /// </summary>
        /// <param name="sender">The sender</param>
        /// <param name="args">The event arguments</param>
        public delegate void AFToastToastDismissedEventHandler(object sender, AFToastActionDismissEventArgs args);

        /// <summary>
        /// Represents the method that will be invoked when a Toast is failed
        /// </summary>
        /// <param name="sender">The sender</param>
        /// <param name="args">The event arguments</param>
        public delegate void AFToastToastFailedEventHandler(object sender, AFToastActionFailEventArgs args);

        /// <summary>
        /// Occurs when a toast is activated, ex: pressed by user
        /// </summary>
        public event AFToastActivatedEventHandler ToastActivated;

        /// <summary>
        /// Occurs when a toast is dismissed, ex: cancelled by user
        /// </summary>
        public event AFToastToastDismissedEventHandler ToastDismissed;

        /// <summary>
        /// Occurs when a toast is failed
        /// </summary>
        public event AFToastToastFailedEventHandler ToastFailed;

        #endregion

        #region Classes
        /// <summary>
        /// Represents the raw event arguments of action of Toast
        /// </summary>
        public class AFRawToastActionEventArgs : EventArgs
        {
            /// <summary>
            /// Gets the device ID
            /// </summary>
            public string DeviceInstanceId { get; internal set; }

            /// <summary>
            /// Gets the notification ID
            /// </summary>
            public string NotificationKey { get; internal set; }
        }

        /// <summary>
        /// Represents the event arguments of toast activated action
        /// </summary>
        public class AFToastActionActivateEventArgs : AFRawToastActionEventArgs
        {
            //TBD
        }
        /// <summary>
        /// Represents the event arguments of toast dismissed action
        /// </summary>
        public class AFToastActionDismissEventArgs : AFRawToastActionEventArgs
        {
            /// <summary>
            /// Gets the dismissed reason of toast
            /// </summary>
            public ToastDismissalReason DismissReason { get; internal set; }
        }

        /// <summary>
        /// Represents the event arguments of toast failed action
        /// </summary>
        public class AFToastActionFailEventArgs : AFRawToastActionEventArgs
        {
            /// <summary>
            /// Gets the error code that was generated in the process of raising a toast notification.
            /// </summary>
            public Exception ErrorCode { get; internal set; }
        }
        #endregion

        /// <summary>
        /// Initialize a new instance of the <see cref="AFToastManager"/> class.
        /// </summary>
        public AFToastManager()
        {
            TryCreateShortcut();
        }

        #region public methods
        /// <summary>
        /// Create the Toast XML structure and show the Toast
        /// </summary>
        /// <param name="notificationKey">A key to represent a specific Toast</param>
        /// <param name="toastItem">A ToastItem structure used to show the Toast</param>
        /// <exception cref="System.ArgumentNullException" />
        public void CreateAndShowToast(string notificationKey, ToastItem toastItem)
        {
            if (notificationKey == null)
            {
                throw new ArgumentNullException("notificationKey");
            }

            ToastTemplateType toastType = GetToastType(toastItem);

            XmlDocument toastXml = ToastXmlGenerator(notificationKey, toastItem, toastType);

            //Generate Toast
            //Create the toast and attach event listeners
            StaticLogger.LogInfo("[Toast][CreateAndShowToast()] : Generate Toast / Append event.");
            ToastNotification toast = new ToastNotification(toastXml);
            toast.Activated += AFToastActivated;
            toast.Dismissed += AFToastDismissed;
            toast.Failed += AFToastFailed;

            // Show the toast. Be sure to specify the AppUserModelId on your application's shortcut!
            StaticLogger.LogInfo("[Toast][CreateAndShowToast()] : CreateToastNotifier()");
            ToastNotificationManager.CreateToastNotifier(APP_ID).Show(toast);
        }
        #endregion

        /// <summary>
        /// Get the BitmapSource from the the image path
        /// </summary>
        /// <param name="iconFile">The image path</param>
        /// <returns>The BitmapSource-format bitmap</returns>
        /// <exception cref="System.ArgumentException" />
        public BitmapSource GreateBitmapSourceByFile(string iconFile)
        {
            if (string.IsNullOrEmpty(iconFile))
            {
                throw new ArgumentException("'iconFile' is not specified.");
            }

            Bitmap bitmapIcon = new Bitmap(System.Drawing.Image.FromFile(iconFile));
            System.Windows.Media.Imaging.BitmapSource bitmapSource = System.Windows.Interop.Imaging.CreateBitmapSourceFromHBitmap(
                        bitmapIcon.GetHbitmap(),
                        IntPtr.Zero,
                        System.Windows.Int32Rect.Empty,
                        System.Windows.Media.Imaging.BitmapSizeOptions.FromWidthAndHeight(bitmapIcon.Width, bitmapIcon.Height));

            bitmapSource.Freeze();

            return bitmapSource;
        }

        #region private methods
        /// <summary>
        /// Check the existence of the shortcut.
        /// A shortcut is required for a toast to jump
        /// </summary>
        /// <returns></returns>
        private bool TryCreateShortcut()
        {
            String shortcutPath = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData) + "\\Microsoft\\Windows\\Start Menu\\Programs\\Acer\\Acer Extend.lnk";
			if (!File.Exists(shortcutPath))
			{
                //For temporary use since to create a shortcut under common user location need priviledge.
				shortcutPath = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) + "\\Microsoft\\Windows\\Start Menu\\Programs\\Acer Extend.lnk";
				InstallShortcut(shortcutPath);
				return true;
			}
			return false;
        }

        /// <summary>
        /// Create a shortcut with the specified path
        /// A shortcut is required for a toast to jump
        /// </summary>
        /// <param name="shortcutPath">Shortcut path</param>
        private void InstallShortcut(String shortcutPath)
        {
            try
            {
                StaticLogger.TrackEnter();
                StaticLogger.LogInfo("[Toast][InstallShortcut()] : Start.");
                // Find the path to the current executable
                String exePath = System.Diagnostics.Process.GetCurrentProcess().MainModule.FileName;
                IShellLinkW newShortcut = (IShellLinkW)new CShellLink();

                // Create a shortcut to the exe
                ShortcutHelpers.ErrorHelper.VerifySucceeded(newShortcut.SetPath(exePath));
                ShortcutHelpers.ErrorHelper.VerifySucceeded(newShortcut.SetArguments(""));

                // Open the shortcut property store, set the AppUserModelId property
                IPropertyStore newShortcutProperties = (IPropertyStore)newShortcut;

                using (PropVariant appId = new PropVariant(APP_ID))
                {
                    ShortcutHelpers.ErrorHelper.VerifySucceeded(newShortcutProperties.SetValue(SystemProperties.System.AppUserModel.ID, appId));
                    ShortcutHelpers.ErrorHelper.VerifySucceeded(newShortcutProperties.Commit());
                }

                // Commit the shortcut to disk
                IPersistFile newShortcutSave = (IPersistFile)newShortcut;

                ShortcutHelpers.ErrorHelper.VerifySucceeded(newShortcutSave.Save(shortcutPath, true));

                StaticLogger.LogInfo("[Toast][InstallShortcut()] : End.");
            }
            finally
            {
                StaticLogger.TrackLeave();
            }
        }

        /// <summary>
        /// Get the default template of the Toast from Windows
        /// </summary>
        /// <param name="iconText">Used to check if the toast template should have a image</param>
        /// <param name="numOfText">Used to check how many text lines the toast template should support</param>
        /// <returns>Toast template type</returns>
        private ToastTemplateType GetToastType(ToastItem item)
        {
            try
            {
                StaticLogger.TrackEnter();

                ToastTemplateType toastType = ToastTemplateType.ToastImageAndText01;
                if (String.IsNullOrWhiteSpace(item.IconPath)) //only text, no image exist
                {
                    if (item.ToastContent == null || item.ToastContent.Length == 0)
                    {
                        toastType = ToastTemplateType.ToastText01;
                        StaticLogger.LogInfo("[Toast][GetToastType()] : Template = ToastText01.");
                    }
                    else
                    {
                        switch (item.ToastContent.Length)
                        {
                            case 1:
                                toastType = ToastTemplateType.ToastText02;
                                StaticLogger.LogInfo("[Toast][GetToastType()] : Template = ToastText02.");
                                break;
                            case 2:
                            default:
                                toastType = ToastTemplateType.ToastText04;
                                StaticLogger.LogInfo("[Toast][GetToastType()] : Template = ToastText04.");
                                break;
                        }
                    }          
                }
                else
                {
                    if (item.ToastContent == null || item.ToastContent.Length == 0)
                    {
                        toastType = ToastTemplateType.ToastImageAndText01;
                        StaticLogger.LogInfo("[Toast][GetToastType()] : Template = ToastImageAndText01.");
                    }
                    else
                    {
                        switch (item.ToastContent.Length)
                        {
                            case 1:
                                toastType = ToastTemplateType.ToastImageAndText02;
                                StaticLogger.LogInfo("[Toast][GetToastType()] : Template = ToastImageAndText02.");
                                break;
                            case 2:
                            default:
                                toastType = ToastTemplateType.ToastImageAndText04;
                                StaticLogger.LogInfo("[Toast][GetToastType()] : Template = ToastImageAndText04.");
                                break;
                        }
                    }     
                }

                return toastType;
            }
            finally
            {
                StaticLogger.TrackLeave();
            } 
        }

        /// <summary>
        /// Create a new toast icon image with the specified transparent edge.
        /// The target is to create a transparent rect B to include the original image A.
        ///  _____________
        /// |   _______   |
        /// |  |       |  | <--- B
        /// |  |   A   |  |  
        /// |  |_______|  |
        /// |_____________|
        /// 
        /// </summary>
        /// <param name="originIconPath">The original icon path in the local drive.</param>
        /// <param name="distanceToEdge">The width of the transparent edge.</param>
        /// <returns></returns>
        /// <exception cref="System.ArgumentException" />
        private string CreateToastTransparentEdgeIconImage(string originIconPath, int distanceToEdge)
        {
            if (string.IsNullOrEmpty(originIconPath))
            {
                throw new ArgumentException("'originIconPath' is not specified.");
            }

            //First, check if the image with transparent edge has already being created before.
            string outputImagePath = string.Empty;
            outputImagePath = originIconPath.Substring(0, originIconPath.LastIndexOf('.')) + "-1.png";
            if (File.Exists(outputImagePath))
            {
                return outputImagePath;
            }

            try
            {
                Bitmap bitmapIcon = (Bitmap)Bitmap.FromFile(originIconPath);
                int totalIncreasedSize = distanceToEdge * 2;
                int newWidth = bitmapIcon.Width + totalIncreasedSize;
                int newHeight = bitmapIcon.Height + totalIncreasedSize;

                //Create the transparent rectangle with new width & height.
                GeometryGroup myRect = new GeometryGroup();
                myRect.Children.Add(new RectangleGeometry(new System.Windows.Rect(new System.Windows.Size(newWidth, newHeight))));

                GeometryDrawing myGrometryDraw = new GeometryDrawing();
                uint color = 0x00ffffff;
                myGrometryDraw.Geometry = myRect;
                myGrometryDraw.Brush = new SolidColorBrush(System.Windows.Media.Color.FromArgb((byte)(color >> 24), (byte)color, (byte)(color >> 8), (byte)(color >> 16)));

                //Get the bitmapSource of the original bitmap
                System.Windows.Media.Imaging.BitmapSource bitmapSource = System.Windows.Interop.Imaging.CreateBitmapSourceFromHBitmap(
                            bitmapIcon.GetHbitmap(),
                            IntPtr.Zero,
                            System.Windows.Int32Rect.Empty,
                            System.Windows.Media.Imaging.BitmapSizeOptions.FromWidthAndHeight(bitmapIcon.Width, bitmapIcon.Height));
                bitmapSource.Freeze();

                //Integrate the transparent rectangle with the bitmapSource
                //The bitmapSource should put in the central of the rectangle.
                DrawingGroup DrawingGroup = new DrawingGroup();
                DrawingGroup.Children.Add(myGrometryDraw);
                DrawingGroup.Children.Add(new ImageDrawing(bitmapSource, new System.Windows.Rect(distanceToEdge, distanceToEdge, bitmapIcon.Width, bitmapIcon.Height)));

                DrawingImage drawImage = new DrawingImage(DrawingGroup);
                DrawingVisual vis = new DrawingVisual();
                DrawingContext cont = vis.RenderOpen();
                cont.DrawImage(drawImage, new System.Windows.Rect(new System.Windows.Size(newWidth, newHeight)));
                cont.Close();

                RenderTargetBitmap rtb = new RenderTargetBitmap(newWidth, newHeight, 96d, 96d, PixelFormats.Default);
                rtb.Render(vis);

                //Generate a png bitmap encoder to create and save the image.
                using (MemoryStream outStream = new MemoryStream())
                {
                    BitmapEncoder enc = new PngBitmapEncoder();
                    enc.Frames.Add(BitmapFrame.Create(rtb));
                    enc.Save(outStream);
                    System.Drawing.Bitmap bitmap = new System.Drawing.Bitmap(outStream);
                    bitmap.Save(outputImagePath);
                }
            }
            catch (Exception ex)
            {
                StaticLogger.LogError("[CreateToastTransparentEdgeIconImage] : Exception: " + ex.ToString());
            }
           
            return outputImagePath;
        }

        /// <summary>
        /// Generate the toast XML structure
        /// </summary>
        /// <param name="toastType">The toast template type</param>
        /// <param name="toastItem">The toast content structure</param>
        /// <returns>Toast XmlDocument</returns>
        private XmlDocument ToastXmlGenerator(string notificationKey, ToastItem toastItem, ToastTemplateType toastType)
        {
            try
            {
                StaticLogger.TrackEnter();
                StaticLogger.LogInfo("[Toast][ToastXmlGenerator()] : Start.");

                // Get a toast XML template
                XmlDocument toastXml = ToastNotificationManager.GetTemplateContent(toastType);

                // Fill in the text elements
                XmlNodeList stringElements = toastXml.GetElementsByTagName("text");
                stringElements[0].AppendChild(toastXml.CreateTextNode(toastItem.ToastTitle));
                for (int index = 1; index < stringElements.Length; ++index)
                {
                    stringElements[index].AppendChild(toastXml.CreateTextNode(toastItem.ToastContent[index - 1]));
                }

                string smallerImagePath = CreateToastTransparentEdgeIconImage(toastItem.IconPath, 24);

                // Specify the absolute path to an image   
                if (toastItem.IconPath != null)
                {
                    string imagePath = "file:///" + smallerImagePath; //Path.GetFullPath(toastItem.IconPath);
                    XmlNodeList imageElements = toastXml.GetElementsByTagName("image");
                    imageElements[0].Attributes.GetNamedItem("src").NodeValue = imagePath;
                }

                //Embed "airFusion" node 
                XmlNodeList toastNodes = toastXml.GetElementsByTagName("toast");
                toastNodes[0].AppendChild(toastXml.CreateElement(AFToastXmlTag));
                //Embed notificationKey in "airFusion" node
                XmlNodeList AFNotificationNodes = toastXml.GetElementsByTagName(AFToastXmlTag);
                AFNotificationNodes[0].AppendChild(toastXml.CreateTextNode(notificationKey));
                //Embed DeviceInstanceId in "airFusion" node
                AFNotificationNodes[0].AppendChild(toastXml.CreateTextNode(toastItem.DeviceInstanceId.Value.ToString()));

                //toast XML example
                /*
                var xml = @"<toast>
                    <visual>
                        <binding template=""ToastText02"" branding=""none"">
                            <text id=""1"">Sean</text>
                            <text id=""2"">Tsai</text>  
                        </binding>  
                    </visual>
                    <airFusion>8888888</airFusion>
                </toast>";
                var xmlDoc = new XmlDocument();
                xmlDoc.LoadXml(xml);
                */
                StaticLogger.LogInfo("[Toast][ToastXmlGenerator()] : End.");
                return toastXml;
            }
            finally
            {
                StaticLogger.TrackLeave();
            }
        }
        #endregion

        private string GetAFItemByToastXml(ToastNotification sender, int nodeIndex)
        {
            XmlNodeList AFNotificationNodes = sender.Content.GetElementsByTagName(AFToastXmlTag);
            string itemValue = String.Empty;
            if (AFNotificationNodes[0].ChildNodes[nodeIndex].NodeValue != null)
                itemValue = AFNotificationNodes[0].ChildNodes[nodeIndex].NodeValue.ToString();

            return itemValue;
        }

        #region event function
        /// <summary>
        /// Occurs when user activates a toast notification through a click or touch.
        /// </summary>
        /// <param name="sender">The sender</param>
        /// <param name="e">TBD?</param>
        private void AFToastActivated(ToastNotification sender, object e)
        {
            if (ToastActivated != null)
            {
                ToastActivated(this, new AFToastActionActivateEventArgs()
                {
                    DeviceInstanceId = GetAFItemByToastXml(sender, AFDeviceIDXmlIndex),
                    NotificationKey = GetAFItemByToastXml(sender, AFNotificationIDXmlIndex)
                });
            }
        }

        /// <summary>
        /// Occurs when a toast notification leaves the screen, either by expiring or
        //  being explicitly dismissed by the user.
        /// </summary>
        /// <param name="sender">The sender</param>
        /// <param name="e">The toast dismissed arguments</param>
        private void AFToastDismissed(ToastNotification sender, ToastDismissedEventArgs e)
        {
            if (ToastDismissed != null)
            {
                ToastDismissed(this, new AFToastActionDismissEventArgs()
                {
                    DeviceInstanceId = GetAFItemByToastXml(sender, AFDeviceIDXmlIndex),
                    NotificationKey = GetAFItemByToastXml(sender, AFNotificationIDXmlIndex),
                    DismissReason = e.Reason
                });
            }
        }

        /// <summary>
        /// Occurs when an error is caused when Windows attempts to raise a toast notification.
        /// </summary>
        /// <param name="sender">The sender</param>
        /// <param name="e">The toast failed arguments</param>
        private void AFToastFailed(ToastNotification sender, ToastFailedEventArgs e)
        {
            if (ToastFailed != null)
            {
                ToastFailed(this, new AFToastActionFailEventArgs()
                {
                    DeviceInstanceId = GetAFItemByToastXml(sender, AFDeviceIDXmlIndex),
                    NotificationKey = GetAFItemByToastXml(sender, AFNotificationIDXmlIndex),
                    ErrorCode = e.ErrorCode
                });
            }
        }
        #endregion
    }
}
