#include "stdafx.h"
#include "PullOutputPin.h"
#include "PullSrcFilter.h"

PullOutputPin::PullOutputPin(
	HRESULT * phr,
	PullSrcFilter *pFilter,
	CCritSec * pLock)
	: CBasePin(
	NAME("Async output pin"),
	pFilter,
	pLock,
	phr,
	L"Output",
	PINDIR_OUTPUT),
	m_pFilter(pFilter)
{
	m_ulPositionInPack  = 0;
	m_pTempPack = NULL;
}

PullOutputPin::~PullOutputPin()
{
	delete m_pTempPack;
}

STDMETHODIMP PullOutputPin::NonDelegatingQueryInterface(REFIID riid, void** ppv)
{
	CheckPointer(ppv,E_POINTER);

	if(riid == IID_IAsyncReader)
	{
		m_bQueriedForAsyncReader = TRUE;
		return GetInterface((IAsyncReader*) this, ppv);
	}
	else
	{
		return CBasePin::NonDelegatingQueryInterface(riid, ppv);
	}
}

HRESULT PullOutputPin::GetMediaType(int iPosition, CMediaType *pMediaType)
{
	if(iPosition < 0)
	{
		return E_INVALIDARG;
	}
	if(iPosition > 0)
	{
		return VFW_S_NO_MORE_ITEMS;
	}

	CheckPointer(pMediaType,E_POINTER);
	CheckPointer(m_pFilter,E_UNEXPECTED);

	*pMediaType = *m_pFilter->LoadType();

	return S_OK;
}

HRESULT PullOutputPin::CheckMediaType(const CMediaType* pType)
{
	CAutoLock lck(m_pLock);

	/*  We treat MEDIASUBTYPE_NULL subtype as a wild card */
	if((m_pFilter->LoadType()->majortype == pType->majortype) &&
		(m_pFilter->LoadType()->subtype == MEDIASUBTYPE_NULL   ||
		m_pFilter->LoadType()->subtype == pType->subtype))
	{
		return S_OK;
	}

	return S_FALSE;
}

HRESULT PullOutputPin::InitAllocator(IMemAllocator **ppAlloc)
{
	CheckPointer(ppAlloc,E_POINTER);

	HRESULT hr = NOERROR;
	CMemAllocator *pMemObject = NULL;

	*ppAlloc = NULL;

	/* Create a default memory allocator */
	pMemObject = new CMemAllocator(NAME("Base memory allocator"), NULL, &hr);
	if(pMemObject == NULL)
	{
		return E_OUTOFMEMORY;
	}
	if(FAILED(hr))
	{
		delete pMemObject;
		return hr;
	}

	/* Get a reference counted IID_IMemAllocator interface */
	hr = pMemObject->QueryInterface(IID_IMemAllocator,(void **)ppAlloc);
	if(FAILED(hr))
	{
		delete pMemObject;
		return E_NOINTERFACE;
	}

	ASSERT(*ppAlloc != NULL);
	return NOERROR;
}

// we need to return an addrefed allocator, even if it is the preferred
// one, since he doesn't know whether it is the preferred one or not.
STDMETHODIMP
	PullOutputPin::RequestAllocator(
	IMemAllocator* pPreferred,
	ALLOCATOR_PROPERTIES* pProps,
	IMemAllocator ** ppActual)
{
	CheckPointer(pPreferred,E_POINTER);
	CheckPointer(pProps,E_POINTER);
	CheckPointer(ppActual,E_POINTER);

	ALLOCATOR_PROPERTIES Actual;
	HRESULT hr;

	if(pPreferred)
	{
		hr = pPreferred->SetProperties(pProps, &Actual);

		if(SUCCEEDED(hr))
		{
			pPreferred->AddRef();
			*ppActual = pPreferred;
			return S_OK;
		}
	}

	// create our own allocator
	IMemAllocator* pAlloc;
	hr = InitAllocator(&pAlloc);
	if(FAILED(hr))
	{
		return hr;
	}

	//...and see if we can make it suitable
	hr = pAlloc->SetProperties(pProps, &Actual);
	if(SUCCEEDED(hr))
	{
		// we need to release our refcount on pAlloc, and addref
		// it to pass a refcount to the caller - this is a net nothing.
		*ppActual = pAlloc;
		return S_OK;
	}

	// failed to find a suitable allocator
	pAlloc->Release();

	// if we failed because of the IsAligned test, the error code will
	// not be failure
	if(SUCCEEDED(hr))
	{
		hr = VFW_E_BADALIGN;
	}
	return hr;
}

// queue an aligned read request. call WaitForNext to get
// completion.
STDMETHODIMP PullOutputPin::Request(
	IMediaSample* pSample,
	DWORD_PTR dwUser)           // user context
{
	return S_OK;
}

// sync-aligned request. just like a request/waitfornext pair.
STDMETHODIMP
	PullOutputPin::SyncReadAligned(
	IMediaSample* pSample)
{
	return S_OK;
}

//
// collect the next ready sample
STDMETHODIMP
	PullOutputPin::WaitForNext(
	DWORD dwTimeout,
	IMediaSample** ppSample,  // completed sample
	DWORD_PTR * pdwUser)        // user context
{
	return S_OK;
}

//
// synchronous read that need not be aligned.
STDMETHODIMP
	PullOutputPin::SyncRead(
	LONGLONG llPosition,    // absolute Io position
	LONG lLength,       // nr bytes required
	BYTE* pBuffer)      // write data here
{
	/*swprintf_s(SourceFilterLogManager::m_logBuffer, _T("Thread ID: %d, <SyncRead> llPos: %lld, lLength: %ld, Wait...\n"), GetCurrentThreadId(), llPosition, lLength);
	SourceFilterLogManager::OutputLog(SourceFilterLogManager::m_logBuffer);*/

	DWORD tick1 = 0, tick2 = 0;
	tick1 = GetTickCount();

	DWORD         dwHaveRead = 0;
	PMPEGTS_PACK   pPack  = NULL;

	int iWaitCount = 0;
	while (m_pFilter->m_idxRecvWriteRound*m_pFilter->RINGBUFFERSIZE + m_pFilter->m_idxRecvWritePos - llPosition < lLength)
	{
		//SourceFilterLogManager::OutputLog(_T("!!! Can't find any data from Ring Queue, Wait for data input and Send to Demuex to Decoder\n"));

		// avoid waiting too long, wait at most 30*iWaitCount ms
		Sleep(20);
		iWaitCount++;

		if(m_pFilter->m_bSendDummyinfo)
		{
			SourceFilterLogManager::OutputLog(_T("!!! Send Dummy data to LAVSplitter\n"));
			memcpy(pBuffer, m_pFilter->m_pRingBuffer, lLength);
			return S_OK;
		}

		//if (iWaitCount-- < 0)
		{
			/*swprintf_s(SourceFilterLogManager::m_logBuffer, _T("!!! ERROR, RecvWriteRound %d, RecvWritePos %d, ReadPos %d\n"), m_pFilter->m_idxRecvWriteRound, m_pFilter->m_idxRecvWritePos, llPosition);
			SourceFilterLogManager::OutputLog(SourceFilterLogManager::m_logBuffer);*/
			//return S_FALSE;
		}
	}

	if (iWaitCount > 5)
	{
		swprintf_s(SourceFilterLogManager::m_logBuffer, _T("Thread ID: %d, Can't find any data from Ring Queue, wast time %d ms...\n"), GetCurrentThreadId(), iWaitCount*20);
		SourceFilterLogManager::OutputLog(SourceFilterLogManager::m_logBuffer);
	}

	int iReadPos = llPosition%m_pFilter->RINGBUFFERSIZE;
	int diff = iReadPos + lLength - m_pFilter->RINGBUFFERSIZE;
	if (diff>0)
	{
		memcpy(pBuffer, m_pFilter->m_pRingBuffer + iReadPos, lLength-diff);
		memcpy(pBuffer + lLength - diff, m_pFilter->m_pRingBuffer, diff);
	}
	else
	{
		memcpy(pBuffer, m_pFilter->m_pRingBuffer + iReadPos, lLength);
	}

	/*tick2 = GetTickCount();
	if(tick2 != tick1)
	{
	swprintf_s(SourceFilterLogManager::m_logBuffer, _T("Send to next filter %d bytes, wait time %d, RecvWriteRound %d, RecvWritePos %d, ReadPos %d\n"), lLength, tick2 - tick1, m_pFilter->m_idxRecvWriteRound, m_pFilter->m_idxRecvWritePos, llPosition);
	SourceFilterLogManager::OutputLog(SourceFilterLogManager::m_logBuffer);
	}*/

	return S_OK;

	/*
	while ((LONG)dwHaveRead < lLength)
	{
	if (lLength - dwHaveRead >= MPEGTS_PACK_SIZE - m_ulPositionInPack)
	{
	if(m_pTempPack != NULL)
	{
	swprintf_s(SourceFilterLogManager::m_logBuffer, _T(", Put remnant bytes: %lud"), MPEGTS_PACK_SIZE - m_ulPositionInPack);
	SourceFilterLogManager::OutputLog(SourceFilterLogManager::m_logBuffer);

	CopyMemory((PVOID)(pBuffer + dwHaveRead), (PVOID)((PBYTE)(m_pTempPack) + m_ulPositionInPack), (SIZE_T)MPEGTS_PACK_SIZE - m_ulPositionInPack);

	m_pTempPack = NULL;

	dwHaveRead += MPEGTS_PACK_SIZE - m_ulPositionInPack;
	m_ulPositionInPack = 0;
	}
	else
	{
	// Just copy the whole pack data
	pPack = m_pFilter->QueuePop();
	if (pPack != NULL)
	{
	CopyMemory((PVOID)(pBuffer + dwHaveRead), (PVOID)((PBYTE)(pPack) + m_ulPositionInPack), (SIZE_T)MPEGTS_PACK_SIZE - m_ulPositionInPack);

	delete pPack;

	dwHaveRead += MPEGTS_PACK_SIZE - m_ulPositionInPack;
	m_ulPositionInPack = 0;
	}
	else
	{
	Sleep(10);
	}
	}
	}
	else
	{
	// Copy part of the pack data
	pPack = m_pFilter->QueuePop();
	if (pPack != NULL)
	{
	m_pTempPack = pPack;   // To save the packet which has remnant bytes

	SourceFilterLogManager::OutputLog(_T(", Get remnant packet"));

	m_ulPositionInPack = lLength - dwHaveRead;
	CopyMemory((PVOID)(pBuffer + dwHaveRead), (PVOID)(pPack), (SIZE_T)m_ulPositionInPack);
	dwHaveRead += m_ulPositionInPack;
	}
	else
	{
	Sleep(10);
	}
	}
	}

	SourceFilterLogManager::OutputLog(_T(", Get!!\n"));
	*/
	return S_OK;
}

// return the length of the file, and the length currently
// available locally. We only support locally accessible files,
// so they are always the same
STDMETHODIMP
	PullOutputPin::Length(
	LONGLONG* pTotal,
	LONGLONG* pAvailable)
{
	*pTotal = 22936;
	*pAvailable = 0;

	return S_OK;
}

STDMETHODIMP
	PullOutputPin::BeginFlush(void)
{
	return S_OK;
}

STDMETHODIMP
	PullOutputPin::EndFlush(void)
{
	return S_OK;
}

STDMETHODIMP
	PullOutputPin::Connect(
	IPin * pReceivePin,
	const AM_MEDIA_TYPE *pmt   // optional media type
	)
{
	return m_pFilter->Connect(pReceivePin, pmt);
}